from .flat_code import Block
from ..printer import auto_wrap, key_value
from ..model.wtypes import BASE, CONST, VAR


class FlattenerError(Exception):
    '''
    Represents an error detected when converting an AST into flat code.

     - message - the error message.
     - ast - the offending source code fragment.
    '''
    def __init__(self, message, ast):
        super(FlattenerError, self).__init__(message)
        self.ast = ast


def check_distinct_ids(ids, error_message):
    '''
    If ids contains any duplicate non-`None` Identifiers, raises a
    FlattenerError.

     - ids - an iterable of optional Identifiers.
     - error_message - a string containing a single '%s'.
       The error message will be constructed by replacing the '%s' with the
       duplicate identifier.
    '''
    seen = set()
    for ast in ids:
        if ast is not None:
            if ast.s in seen:
                raise FlattenerError(error_message % ast.s, ast)
            seen.add(ast.s)


def check_distinct_tags(tags, error_message):
    '''
    Like `check_distinct_ids()`, but for tags.
    '''
    seen = set()
    for ast in tags:
        if ast is not None:
            if ast.s in seen:
                raise FlattenerError(error_message % ast.s, ast)
            seen.add(ast.s)


def constness_from_keyword(ast):
    '''
    Returns the tag (a str) that represents the constness specified in an
    "array" or "ref" type literal.

     - ast - `None`, or a Keyword with tag 'const' or 'var'.

    Returns 'BASE', 'CONST', 'VAR'.
    '''
    # If this goes out of date wrt "wtypes.py", we will see TypeVMErrors.
    # It is therefore safe to hardwire the string constants here.
    if ast is None:
        return 'BASE'
    if ast.tag == 'const':
        return 'CONST'
    if ast.tag == 'var':
        return 'VAR'
    assert False, ast


class Names(object):
    '''
    Represents an ordered list of declared variables, some with names and some
    without. Variable names are represented as instances of Identifier. For
    those variables with names, records the Identifier used to declare them.
    Provides methods to:

     - Lookup variable names efficiently, and map them onto small integers.
       Variables are numbered sequentially from zero in the order they are
       declared.

     - Declare new named variables, possibly shadowing existing ones.
       Shadowed variables become anonymous.

     - Declare new anonymous variables.

     - Retrieve the all variables declared so far, in order, as a list of
       tokens, with `None` for the anonymous variables.
    '''
    def __init__(self, copy=None):
        ''' - copy - a Names object to copy, or `None`.'''
        self.used = copy.used if copy else 0
        # WONTFIX: Eliminate O(n) copy. In Python we probably can't beat it.
        self.by_name = dict(copy.by_name) if copy else {}

    def __contains__(self, ast):
        '''
        Tests whether a variable named `ast` has been declared.
         - ast - an Identifier.
        '''
        if ast is None:
            return False
        assert ast.tag == 'identifier'
        return ast.s in self.by_name

    def __len__(self):
        '''
        Returns the number of variables declared, including shadowed
        and anonymous ones.
        '''
        return self.used

    def __getitem__(self, ast):
        '''
        Returns a pair `index, first_ast` where `first_ast` is the Identifier
        used to declare the variable. It is not possible to look up anonymous
        variables.
         - ast - an Identifier.
        '''
        assert ast.tag == 'identifier'
        return self.by_name[ast.s]

    def __pretty__(self):
        return (
            'Names {',
            ', ',
            '}',
            [auto_wrap(from_id(ast)) for ast in self.as_list()],
        )

    def declare(self, ast):
        '''
        (Re)declares the variable named `ast`, and returns its index.
        If the new variable shadows an earlier one, the earlier one becomes
        anonymous.
         - ast - an Indentifier.
        '''
        assert ast.tag == 'identifier'
        index = self.declare_anonymous()
        self.by_name[ast.s] = (index, ast)
        return index

    def declare_anonymous(self):
        '''Declares an anonymous variable.'''
        index = self.used
        self.used += 1
        return index

    def ensure(self, ast):
        '''
        If `token` has not yet been declared, declares it.
        Either way, returns `index`.
         - token - an Indentifier.
        '''
        assert ast.tag == 'identifier'
        if ast in self:
            return self[ast][0]
        else:
            return self.declare(ast)

    def as_list(self):
        '''
        Returns a list with one element for each delcared identifier.
        Each item is either the Identifier used to declare the variable, or
        `None` if the variable is anonymous. Shadowed variables are anonymous.
        '''
        ans = [None] * self.used
        for index, ast in self.by_name.values():
            ans[index] = ast
        return ans


class TopLevel(object):
    '''
    Accumulates a list of identifiers mentioned but not declared in a block of
    code.

    Public fields:

     - captures - A Names represening the variables mentioned but not
       declared.
    '''
    def __init__(self):
        self.captures = Names()

    def __pretty__(self):
        return (
            '%s {' % self.__class__.__name__,
            ', ',
            '}',
            [auto_wrap(self.captures)],
        )

    def flatten_return(self, ast, scope):
        '''
        Appends to `scope` code for the statement "return <expr>", or for
        "return" if `expr` is `None`.

         - ast - a Return representing the AST of the return statement.
         - scope - a LexicalScope to which to append code.

        Subclasses must implement this abstract method as appropriate.
        '''
        raise NotImplementedError("abstract method")


class Repl(TopLevel):
    '''A TopLevel suitable for one compilation unit executed by the REPL.'''

    def flatten_return(self, ast, scope):
        raise FlattenerError(
            '"return" is not allowed outside a function', ast)


class Function(TopLevel):
    '''A TopLevel suitable for the body of a function definition.'''

    def flatten_return(self, ast, scope):
        assert ast.tag == 'return'
        if ast.expr is None:
            self.flatten_return_without_value(scope, ast.error_loc)
        else:
            self.flatten_return_with_value(ast.expr, scope, ast.error_loc)

    def flatten_return_with_value(self, expr, scope, loc):
        scope.flatten_expr(expr)
        scope.block.return_(loc)

    def flatten_return_without_value(self, scope, loc):
        scope.block.struct(0, loc)
        scope.block.return_(loc)


class FunctionWithReturnType(Function):
    '''
    A TopLevel suitable for the body of a function definition with an explicit
    return type.
    '''
    def __init__(self):
        super(FunctionWithReturnType, self).__init__()
        # Return type is implicit first "capture".
        self.captures.declare_anonymous()

    def flatten_return_with_value(self, expr, scope, loc):
        scope.flatten_expr(expr)
        scope.block.capture(0, loc) # Return type is implicit first capture.
        scope.block.cast(loc)
        scope.block.return_(loc)

    def flatten_return_without_value(self, scope, loc):
        scope.block.capture(0, loc) # Return type is implicit first capture.
        scope.block.return_void(loc)


class LexicalScope(object):
    '''
    Accumulates a sequence of flat code instructions corresponding to a
    high-level lexical block.

    Public fields:

     - top_level - The enclosing TopLevel.
     - locals - A Names representing the local variables currently in scope.
       Indexes must be distinct for locals that are simultaneously in scope,
       but can be reused otherwise.
     - block - A flat_code.Block into which to put instructions.

    The "first occurrence" is an 'identifier' token.
    '''
    def __init__(self, top_level):
        self.top_level = top_level
        self.locals_ = Names()
        self.block = Block()

    def nested_scope(self):
        '''
        Returns a LexicalScope with the same `top_level` as this one, a fresh
        Block, and a fresh copy of `locals`. New code and locals can be added
        to the new LexicalScope without affecting this LexicalScope.
        '''
        ans = LexicalScope(self.top_level)
        ans.locals_ = Names(self.locals_)
        return ans

    def __pretty__(self):
        return (
            'LexicalScope {',
            ', ',
            '}',
            [
                key_value('top_level', self.top_level),
                key_value('locals', self.locals_),
                key_value('block', self.block),
            ],
        )

    def flatten_stmt(self, ast, scope_noun):
        '''
         - ast - an AST representing a stmt.
         - scope - the LexicalScope to which to append instructions.
         - scope_noun - `None` if "break" and "continue" are legal, otherwise
           a noun phrase describing the scope, for use in error messages.
        '''
        # This test could be omitted but it can detect an error earlier.
        if not self.block.is_end_reachable:
            raise FlattenerError('Code is not reachable', ast)
        if ast.tag == 'assign':
            assert ast.op.tag == 'assign_op', repr(ast.op)
            # Note: evaluation order is RHS first. We think this is only
            # mildly surprising; in almost all cases the user won't notice.
            # It makes `flatten_assign()` much simpler, and avoids potentially
            # large stack growth.
            self.flatten_expr(ast.rhs)
            if ast.op.s == '=':
                # `lvalue` could be nested casts and/or structs.
                # Flattening the whole thing might not yield an `lvalue`.
                # Instead, descend it recursively.
                self.flatten_assign(ast.lhs, ast.error_loc)
            else:
                # In-place. `lvalue` must have an address in this case.
                # See "doc/compiler/assign_ops.txt" for rationale.
                # Note: evaluation order is RHS first.
                self.flatten_expr(ast.lhs)
                self.block.assign_op(
                    assign_op_to_bin_op(ast.op.s),
                    ast.error_loc,
                )
        elif ast.tag == 'break':
            if scope_noun:
                raise FlattenerError(
                    'cannot "break" out of %s' % scope_noun, ast)
            self.block.break_(ast.error_loc)
        elif ast.tag == 'continue':
            if scope_noun:
                raise FlattenerError(
                    'cannot "continue" in %s' % scope_noun, ast)
            self.block.continue_(ast.error_loc)
        elif ast.tag == 'decl':
            assert ast.constness.tag in ('const', 'var')
            self.flatten_expr(ast.rhs)
            self.flatten_decl(
                ast.constness.tag == 'var', ast.lhs, ast.error_loc)
        elif ast.tag == 'expr':
            self.flatten_expr(ast.expr)
            # If `ast` is "fn foo ..." etc. then bind identifier "foo" and
            # emit 'init' instead of 'expr'.
            identifier = expr_id(ast.expr)
            if identifier is None:
                self.block.expr(ast.error_loc)
            else:
                self.flatten_decl(False, identifier, identifier.error_loc)
        elif ast.tag == 'for':
            # We compile this as:
            # {
            #     const $next = sequence.iter()
            #     loop {
            #         switch $next() {
            #             case SOME(variable) {
            #                 for_body
            #             }
            #             case NONE {
            #                 else_body
            #                 break
            #             }
            #         }
            #     }
            # }
            # Make a scope to define the lifetime of `$next`.
            inline_scope = self.nested_scope()
            inline_scope.flatten_expr(ast.expr)
            inline_scope.block.field('iter', ast.expr.error_loc)
            inline_scope.block.call(0, ast.expr.error_loc)
            index = inline_scope.locals_.declare_anonymous()
            inline_scope.block.init(False, ast.expr.error_loc)
            # Make a scope for the loop body.
            loop_scope = inline_scope.nested_scope()
            # Make a scope for the 'SOME' case.
            for_scope = loop_scope.nested_scope()
            for_scope.locals_.declare(ast.identifier)
            for stmt in ast.body.stmts:
                for_scope.flatten_stmt(stmt, None)
            # Make a scope for the 'NONE' case.
            else_scope = loop_scope.nested_scope()
            if ast.else_ is not None:
                for stmt in ast.else_.stmts:
                    else_scope.flatten_stmt(stmt, scope_noun)
            break_scope = else_scope.nested_scope()
            break_scope.block.break_(ast.error_loc)
            else_scope.block.if_reachable(break_scope.block, ast.error_loc)
            # Wrap up the case scopes in a 'switch' instruction.
            loop_scope.block.local(index, ast.expr.error_loc)
            loop_scope.block.call(0, ast.expr.error_loc)
            loop_scope.block.switch([
                ('SOME', 1, for_scope.block, ast.error_loc),
                ('NONE', 0, else_scope.block, ast.error_loc),
            ], None, ast.expr.error_loc)
            # Wrap up the loop scope in a 'loop' instruction.
            inline_scope.block.loop(loop_scope.block, ast.error_loc)
            # Wrap up the whole thing in an 'inline' instruction.
            self.block.inline(inline_scope.block, ast.error_loc)
        elif ast.tag == 'fulfil':
            self.flatten_expr(ast.rhs)
            self.flatten_expr(ast.lhs)
            self.block.fulfil(ast.error_loc)
        elif ast.tag == 'if':
            self.flatten_expr(ast.condition)
            then_scope = self.nested_scope()
            for stmt in ast.then.stmts:
                then_scope.flatten_stmt(stmt, scope_noun)
            else_scope = self.nested_scope()
            if ast.else_ is not None:
                for stmt in ast.else_.stmts:
                    else_scope.flatten_stmt(stmt, scope_noun)
            self.block.if_(
                0, then_scope.block, else_scope.block, ast.error_loc)
        elif ast.tag == 'return':
            self.flatten_return(ast)
        elif ast.tag == 'switch':
            self.flatten_expr(ast.expr)
            check_distinct_tags(
                (case.name for case in ast.cases),
                "Case '%s' has already been handled"
            )
            case_quads = [
                self.flatten_case(case, scope_noun)
                for case in ast.cases
            ]
            if ast.else_ is None:
                else_pair = None
            else:
                else_scope = self.nested_scope()
                for stmt in ast.else_.body.stmts:
                    else_scope.flatten_stmt(stmt, scope_noun)
                else_pair = (else_scope.block, ast.else_.error_loc)
            self.block.switch(case_quads, else_pair, ast.error_loc)
        elif ast.tag == 'throw':
            raise NotImplementedError('throw')
        elif ast.tag == 'try':
            raise NotImplementedError('try')
        elif ast.tag == 'while':
            # We compile this as:
            # loop {
            #     if (c) {
            #         while_body
            #     } else {
            #         else_body
            #         break
            #     }
            # }
            loop_scope = self.nested_scope()
            loop_scope.flatten_expr(ast.condition)
            while_scope = loop_scope.nested_scope()
            for stmt in ast.body.stmts:
                while_scope.flatten_stmt(stmt, None)
            else_scope = loop_scope.nested_scope()
            if ast.else_ is not None:
                for stmt in ast.else_.stmts:
                    else_scope.flatten_stmt(stmt, scope_noun)
            break_scope = else_scope.nested_scope()
            break_scope.block.break_(ast.error_loc)
            else_scope.block.if_reachable(break_scope.block, ast.error_loc)
            # Wrap up and return.
            loop_scope.block.if_(
                0, while_scope.block, else_scope.block, ast.error_loc)
            self.block.loop(loop_scope.block, ast.error_loc)
        else:
            assert False

    def flatten_case(self, case, scope_noun):
        '''
        Returns a quad `(tag, num_params, body, loc)` in which:
         - tag - str.
         - num_params - int.
         - body - flat_code.Block.
         - loc - a location for error messages related to the case.

        Params:
         - case - a 'case' token.
         - scope_noun - passed on to `flatten_stmt()`.
        '''
        assert case.tag == 'case'
        check_distinct_ids(
            case.params,
            "Parameter '%s' has already been declared"
        )
        case_scope = self.nested_scope()
        for param_id in case.params:
            case_scope.locals_.declare(param_id)
        for stmt in case.body.stmts:
            case_scope.flatten_stmt(stmt, scope_noun)
        return (
            from_tag(case.name),
            len(case.params),
            case_scope.block,
            case.error_loc,
        )

    def flatten_return(self, ast):
        self.top_level.flatten_return(ast, self)

    def flatten_expr(self, ast):
        '''
        Appends to `scope` code to evaluate `ast` and leave the result on the
        data stack.
        '''
        if ast.tag == 'character':
            x = ord(ast.c)
            if not 0 <= x < 0x110000:
                raise FlattenerError('Character constant too large', ast)
            self.block.const_char(x, ast.error_loc)
        elif ast.tag == 'float':
            self.block.const_float(float(ast.s), ast.error_loc)
        elif ast.tag == 'hex_integer':
            x = int(ast.s, 16)
            if not (-1 << 63) <= x < (1 << 63):
                raise FlattenerError('Integer constant too large', ast)
            self.block.const_int(x, ast.error_loc)
        elif ast.tag == 'integer':
            x = int(ast.s)
            if not (-1 << 63) <= x < (1 << 63):
                raise FlattenerError('Integer constant too large', ast)
            self.block.const_int(x, ast.error_loc)
        elif ast.tag == 'string':
            self.block.const_str(ast.s, ast.error_loc)
        elif ast.tag == 'array':
            for expr in ast.exprs:
                self.flatten_expr(expr)
            self.block.array(len(ast.exprs), ast.error_loc)
        elif ast.tag == 'address_of':
            self.flatten_expr(ast.expr)
            self.block.address_of(ast.error_loc)
        elif ast.tag == 'binary':
            self.flatten_expr(ast.left)
            self.flatten_expr(ast.right)
            self.block.binary(ast.op.s, ast.error_loc)
        elif ast.tag == 'box':
            self.flatten_expr(ast.expr)
            self.block.box(ast.error_loc)
        elif ast.tag == 'call':
            self.flatten_expr(ast.expr)
            for arg in ast.args:
                self.flatten_expr(arg)
            self.block.call(len(ast.args), ast.error_loc)
        elif ast.tag == 'cast':
            self.flatten_expr(ast.expr)
            self.flatten_expr(ast.type_)
            self.block.cast(ast.type_.error_loc)
        elif ast.tag == 'defer':
            self.flatten_expr(ast.expr)
            self.block.defer(ast.error_loc)
        elif ast.tag == 'dereference':
            self.flatten_expr(ast.expr)
            self.block.dereference(ast.error_loc)
        elif ast.tag == 'field':
            self.flatten_expr(ast.expr)
            self.block.field(from_id(ast.name), ast.error_loc)
        elif ast.tag == 'function_def':
            self.flatten_function_def(ast)
        elif ast.tag == 'identifier':
            self.flatten_identifier(ast)
        elif ast.tag == 'new':
            self.flatten_expr(ast.expr)
            self.block.new(ast.error_loc)
        elif ast.tag == 'struct':
            for expr in ast.exprs:
                self.flatten_expr(expr)
            self.block.struct(len(ast.exprs), ast.error_loc)
        elif ast.tag == 'subscript':
            self.flatten_expr(ast.expr)
            self.flatten_expr(ast.index)
            self.block.subscript(ast.error_loc)
        elif ast.tag == 'ternary':
            self.flatten_expr(ast.condition)
            then_scope = self.nested_scope()
            then_scope.flatten_expr(ast.left)
            else_scope = self.nested_scope()
            else_scope.flatten_expr(ast.right)
            self.block.if_(
                1, then_scope.block, else_scope.block, ast.error_loc)
        elif ast.tag == 'type_array':
            self.block.union(
                constness_from_keyword(ast.constness), 0, ast.error_loc)
            self.flatten_expr(ast.type_)
            self.block.cast_to_type(
                'the element type', False, ast.type_.error_loc)
            self.block.union('ARRAY', 2, ast.error_loc)
            self.block.cast_to_type('SHOULD NOT HAPPEN', False, ast.error_loc)
        elif ast.tag == 'type_function':
            # TODO: Recursive function types.
            for param in ast.params:
                assert param.tag == 'param'
                self.flatten_expr(param.type_)
                self.block.cast_to_type(
                    'the parameter type', False, param.type_.error_loc)
            self.block.array(len(ast.params), ast.error_loc)
            self.flatten_expr(ast.return_type)
            self.block.cast_to_type(
                'the return type', False, ast.return_type.error_loc)
            self.block.union('FUNCTION', 2, ast.error_loc)
            self.block.cast_to_type('SHOULD NOT HAPPEN', False, None)
        elif ast.tag == 'type_ref':
            self.block.union(
                constness_from_keyword(ast.constness), 0, ast.error_loc)
            self.flatten_expr(ast.type_)
            self.block.cast_to_type(
                'the referee type', False, ast.type_.error_loc)
            self.block.union('REF', 2, ast.error_loc)
            self.block.cast_to_type('SHOULD NOT HAPPEN', False, ast.error_loc)
        elif ast.tag == 'type_struct':
            # TODO: Instead of testing whether the struct has a name, we
            # should test whether it refers to itself recursively.
            loc = ast.error_loc
            if ast.name is None:
                self.flatten_struct(ast.params, loc)
            else:
                # Make a private scope, and declare a deferred local.
                struct_scope = self.nested_scope()
                struct_scope.block.defer_type(loc)
                index = struct_scope.locals_.declare(ast.name)
                struct_scope.block.init(False, loc)
                # The struct can now refer to itself recursively.
                struct_scope.flatten_struct(ast.params, loc)
                # Fulfil the defer cell, and leave it on the stack.
                struct_scope.block.local(index, loc)
                struct_scope.block.fulfil(loc)
                struct_scope.block.local(index, loc)
                # Inline the private scope.
                self.block.inline(struct_scope.block, loc)
        elif ast.tag == 'type_union':
            # TODO: Instead of testing whether the union has a name, we
            # should test whether it refers to itself recursively.
            loc = ast.error_loc
            if ast.name is None:
                self.flatten_union(ast.cases, loc)
            else:
                # Make a private scope, and declare a deferred local.
                union_scope = self.nested_scope()
                union_scope.block.defer_type(loc)
                index = union_scope.locals_.declare(ast.name)
                union_scope.block.init(False, loc)
                # The union can now refer to itself recursively.
                union_scope.flatten_union(ast.cases, loc)
                # Fulfil the defer cell, and leave it on the stack.
                union_scope.block.local(index, loc)
                union_scope.block.fulfil(loc)
                union_scope.block.local(index, loc)
                # Inline the private scope.
                self.block.inline(union_scope.block, loc)
        elif ast.tag == 'typeof':
            self.flatten_expr(ast.expr)
            self.block.typeof(ast.error_loc)
        elif ast.tag == 'unary':
            self.flatten_expr(ast.expr)
            self.block.unary(ast.op.s, ast.error_loc)
        elif ast.tag == 'union':
            for expr in ast.exprs:
                self.flatten_expr(expr)
            self.block.union(
                from_tag(ast.name), len(ast.exprs), ast.error_loc)
        else:
            assert False, ast.tag

    def flatten_function_def(self, ast):
        # Make a scope for the function body.
        if ast.return_type is not None:
            top_level = FunctionWithReturnType()
        else:
            top_level = Function()
        body_scope = LexicalScope(top_level)
        # Loop through the parameter declarations.
        for param in ast.params:
            assert param.tag == 'param'
            # Declare the parameter as a local in the body.
            if param.name in body_scope.locals_:
                raise FlattenerError(
                    "Parameter '%s' has already been declared" % param.name.s,
                    param.name,
                )
            body_scope.locals_.declare(param.name)
            # Evaluate the parameter type in the outer scope.
            self.flatten_expr(param.type_)
            self.block.cast_to_type(
                'the parameter type', True, param.type_.error_loc)
        # Group the parameter types into an array.
        self.block.array(len(ast.params), ast.error_loc)
        # Body code.
        for stmt in ast.body.stmts:
            body_scope.flatten_stmt(stmt, 'a function body')
        # Add an implicit "return" statement if necessary.
        return_scope = body_scope.nested_scope()
        top_level.flatten_return_without_value(return_scope, ast.error_loc)
        body_scope.block.if_reachable(return_scope.block, ast.error_loc)
        # List the captured identifiers in order of first appearance.
        captures = top_level.captures.as_list()
        # Construct closure, using 'defer' if it is recursive.
        if ast.return_type is not None:
            # Evaluate return type in outer scope.
            self.flatten_expr(ast.return_type)
            self.block.cast_to_type(
                'the return type', True, ast.return_type.error_loc)
            if ast.name in top_level.captures:
                # Potentially recursive. Make a 'defer' cell.
                self.block.defer_function(ast.error_loc)
                # Evaluate all captures in outer scope.
                # `return_type` becomes the first (implicit) capture.
                # There is a special case for the recursive capture.
                index, _ = top_level.captures[ast.name]
                for identifier in captures[1:index]:
                    self.flatten_identifier(identifier)
                self.block.pick(index+1, ast.name.error_loc) # The defer cell.
                for identifier in captures[index+1:]:
                    self.flatten_identifier(identifier)
                # Construct the closure.
                self.block.fulfil_closure(
                    body_scope.block, len(captures), ast.error_loc)
            else:
                # Not recursive, so no 'defer' needed.
                # Evaluate all captures in outer scope.
                # `return_type` becomes the first (implicit) capture.
                for identifier in captures[1:]:
                    self.flatten_identifier(identifier)
                # Construct the closure.
                self.block.closure(
                    body_scope.block, len(captures), ast.error_loc)
        else:
            # Implicit return type. Must not be recursive.
            if ast.name in top_level.captures:
                raise FlattenerError(
                    "a function cannot refer to itself unless it "
                    "has an explicit return type",
                    top_level.captures[ast.name][1],
                )
            # Evaluate all captures in outer scope.
            for identifier in captures:
                self.flatten_identifier(identifier)
            # Construct the closure.
            self.block.closure(
                body_scope.block, len(captures), ast.error_loc)

    def flatten_struct(self, params, loc):
        '''
        Flattens a 'type_struct' expression.

         - params - a list of Params.
        '''
        check_distinct_ids(
            (param.name for param in params),
            "Field '%s' has already been declared"
        )
        for param in params:
            assert param.tag == 'param'
            if param.name is not None:
                raise NotImplementedError("Named field in struct type")
            self.flatten_expr(param.type_)
            self.block.cast_to_type(
                'the field type', False, param.type_.error_loc)
        self.block.array(len(params), loc)
        self.block.union('STRUCT', 1, loc)
        self.block.cast_to_type('SHOULD NOT HAPPEN', False, loc)

    def flatten_union(self, cases, loc):
        '''
        Flattens a 'type_union' expression.

         - cases - list of UnionCases.
        '''
        check_distinct_tags(
            (case.name for case in cases),
            "Case '%s' has already been declared"
        )
        for case in cases:
            self.block.const_str(from_tag(case.name), case.error_loc)
            for e in case.types:
                self.flatten_expr(e)
                self.block.cast_to_type(
                    'the case field type', False, e.error_loc)
            self.block.array(len(case.types), case.error_loc)
            self.block.struct(2, case.error_loc) # The union_case.
        self.block.array(len(cases), loc)
        self.block.union('UNION', 1, loc)
        self.block.cast_to_type('SHOULD NOT HAPPEN', False, loc)

    def flatten_identifier(self, ast):
        '''
        Finds the Variable named `ast` in `scope`, and generates a 'local'
        or 'capture' instruction as appropriate. In the latter case, also
        puts it in `self.top_level.captures` if it is not there already.
        '''
        assert ast.tag == 'identifier'
        loc = ast.error_loc
        if ast in self.locals_:
            self.block.local(self.locals_[ast][0], loc)
        else:
            self.block.capture(self.top_level.captures.ensure(ast), loc)

    def flatten_assign(self, ast, error_loc):
        '''
        For flattening "<lvalue> = <expr>". The `expr` has already been
        flattened and is on the data stack. Note assignment operator is just
        "=".

        Params:
         - ast - An AST representing the lvalue.
         - error_loc - A place to report errors related to the entire assign,
           e.g. the "=" symbol.
        '''
        if ast.tag in (
            'dereference', 'identifier', 'field', 'subscript', 'ternary'
        ):
            # These are terminal cases, i.e. `ast` must have an address.
            self.flatten_expr(ast)
            self.block.assign(error_loc)
        elif ast.tag == 'cast':
            self.flatten_expr(ast.type_)
            self.block.cast(ast.type_.error_loc)
            self.flatten_assign(ast.expr, error_loc)
        elif ast.tag == 'struct':
            self.block.unpack_struct(len(ast.exprs), error_loc)
            # Assign to the lvalues in left-to-right order.
            self.block.reverse(len(ast.exprs), error_loc)
            for lvalue_ast in ast.exprs:
                self.flatten_assign(lvalue_ast, error_loc)
        else:
            assert False

    def flatten_decl(self, is_var, ast, error_loc):
        '''
        For flattening "<decl> <lvalue> = <expr>". `expr` is on the data stack,
        i.e. already flattened. Note assignment operator is "=".

        Params:
         - is_var - `True` if `decl` is "var" or `False` for "const"
         - ast - The AST of the L-value.
         - error_loc - A place to report errors related to the entire decl,
           e.g. the "=" symbol.
        '''
        if ast.tag in ('dereference', 'field', 'subscript'):
            raise FlattenerError("expected a new variable name", ast)
        elif ast.tag == 'identifier':
            # Declare a new local variable.
            self.locals_.declare(ast)
            self.block.init(is_var, ast.error_loc)
        elif ast.tag == 'cast':
            self.flatten_expr(ast.type_)
            self.block.cast(ast.type_.error_loc)
            self.flatten_decl(is_var, ast.expr, error_loc)
        elif ast.tag == 'struct':
            self.block.unpack_struct(len(ast.exprs), error_loc)
            # Assign to the lvalues in left-to-right order.
            self.block.reverse(len(ast.exprs), error_loc)
            for lvalue_ast in ast.exprs:
                self.flatten_decl(is_var, lvalue_ast, error_loc)
        else:
            assert False, ast


# Utility methods for extracting information from AST tokens.

def from_id(ast):
    if ast is None:
        return None
    else:
        assert ast.tag == 'identifier'
        return ast.s


def from_tag(ast):
    assert ast.tag == 'tag'
    return ast.s


def assign_op_to_bin_op(op):
    '''
    Utility function.
    Given an assign_op symbol, return the corresponding bin_op symbol.
    '''
    # We should really use a switch statement or a lookup table or something.
    # Here we use string manipulation instead.
    assert type(op) is str, repr(op)
    assert op != '=', repr(op)
    assert op[-1] == '=', repr(op)
    return op[:-1]


def expr_id(ast):
    if ast.tag == 'function_def':
        return ast.name
    elif ast.tag == 'type_struct':
        return ast.name
    elif ast.tag == 'type_union':
        return ast.name
    else:
        return None
