'''
In Welly, given a value `x` of an appropriate type, the syntax `x.iter`
applies a "field" function to `x`.

The field function returns a "constructor" function, which takes no arguments.

The constructor function returns an "iterator" function, which takes no
arguments.

The iterator function returns `SOME(y)` for each element `y` of `x`, until
there are no more elements, then returns `NONE()`.
'''


from ..model import wtypes
from ..compiler import typed_code_static, operators

from ..model.memory import Array
from ..model.ops import Dynamic


def array_of_types(types):
    assert type(types) is list
    return Array.from_list(
        [t.read() for t in types],
        wtypes.sizeof(wtypes.type_),
    )


# We don't expect to report any errors in the typed code we're generating.
loc = None


# Iterators for integers.

# fn int_field(x: int) {
#     fn int_constructor() {
#         var counter = 0
#         fn int_next() {
#             if counter < x {
#                 const ret = counter
#                 counter += 1
#                 return SOME(ret)
#             } else {
#                 return NONE
#             }
#         }
#         return int_next
#     }
#     return int_constructor
# }


def _int_next_body(block):
    block.unpack_struct(0, loc)
    block.capture(0)
    block.capture(1)
    block.binary(operators.lookup('<', wtypes.int_, wtypes.int_), loc)
    def make_then(block):
        block.capture(0) # `counter`
        block.init(False) # `ret`
        block.capture(0) # `counter`
        block.pick(0)
        block.static(Dynamic.new(wtypes.int_, (1,)), 0) # 1
        block.binary(operators.lookup('+', wtypes.int_, wtypes.int_), loc)
        block.bury(1)
        block.assign(loc)
        block.local(0) # `ret`
        block.union('SOME', 1)
        block.return_()
    def make_else(block):
        result = typed_code_static.union('NONE', [])
        block.static(result, 0)
        block.return_()
    block.if_(0, make_then, make_else, loc)


def _int_constructor_body(block):
    block.unpack_struct(0, loc)
    block.static(Dynamic.new(wtypes.int_, (0,)), 0) # 0
    block.init(True) # `counter`
    block.local(0) # `counter`
    block.capture(0) # `x`
    block.closure(
        _int_next_body,
        array_of_types([]),
        2,
        loc,
    )
    block.return_()


def _int_field_body(block):
    block.unpack_struct(1, loc)
    block.closure(
        _int_constructor_body,
        array_of_types([]),
        1,
        loc,
    )
    block.return_()


# A Dynamic representing a Welly function that takes an int and returns a
# constructor of int iterators.
int_field = typed_code_static.closure(
    _int_field_body,
    array_of_types([wtypes.int_]),
    loc,
)


# Iterators for arrays.

# For each type `t`:
# fn array_field(a: array t) {
#     fn array_constructor() {
#         var counter = 0
#         fn array_next() {
#             if counter < a.length {
#                 const ret = a[counter]
#                 counter += 1
#                 return SOME(ret)
#             } else {
#                 return NONE
#             }
#         }
#         return array_next
#     }
#     return array_constructor
# }


def _array_next_body(block):
    block.unpack_struct(0, loc)
    block.capture(0) # `counter`
    block.capture(1) # `a`
    block.array_length()
    block.binary(operators.lookup('<', wtypes.int_, wtypes.int_), loc)
    def then_body(block):
        block.capture(1) # `a`
        block.capture(0) # `counter`
        block.subscript_array(loc)
        block.init(False) # `ret`
        block.capture(0) # `counter`
        block.pick(0)
        block.static(Dynamic.new(wtypes.int_, (1,)), 0)
        block.binary(operators.lookup('+', wtypes.int_, wtypes.int_), loc)
        block.bury(1)
        block.assign(loc)
        block.local(0) # `ret`
        block.union('SOME', 1)
        block.return_()
    def else_body(block):
        result = typed_code_static.union('NONE', [])
        block.static(result, 0)
        block.return_()
    block.if_(0, then_body, else_body, loc)


def _array_constructor_body(block):
    block.unpack_struct(0, loc)
    block.static(Dynamic.new(wtypes.int_, (0,)), 0) # 0
    block.init(True) # `counter`.
    block.local(0) # `counter`
    block.capture(0) # `a`
    block.closure(
        _array_next_body,
        array_of_types([]),
        2,
        loc,
    )
    block.return_()


def _array_field_body(block):
    block.unpack_struct(1, loc)
    block.closure(
        _array_constructor_body,
        array_of_types([]),
        1,
        loc,
    )
    block.return_()


def array_field(array_t):
    '''
    Returns a Dynamic representing a Welly function that takes an `array_t`
    and returns a constructor of `elem_t` iterators, where `elem_t` is the
    element type of `array_t`.

     - array_t - the Welly type of the array.
    '''
    return typed_code_static.closure(
        _array_field_body,
        array_of_types([array_t]),
        loc,
    )
