from ..compiler import operators, typed_code, typed_code_static, iterator
from ..model import defer, const, wtypes, subtypes

from ..printer import pretty, auto_wrap
from ..model.memory import Ref, Array
from ..model.errors import *
from ..model.ops import (
    Dynamic, is_int_type, is_array_type, array_constness, array_element_type,
    struct_length, struct_iter, union_cases)
from ..model.subtypes import is_subtype
from ..compiler.builtins import builtins


def raise_message(message, loc=None):
    raise StaticMessageVMError(message, loc=loc)


def _is_runtime(x):
    '''
    Returns `True` if `x` represents a value that will be on the stack at
    run time.

     - x - int or Dynamic. See the docstring for Stack.
    '''
    return type(x) is int


def _is_static_const(x):
    '''
    Tests whether `x` can be dereferenced at compile time.

     - x - int or Dynamic. See the docstring for Stack.
    '''
    return type(x) is Dynamic and x.is_const()


def _lookup_type(block, x):
    '''
    Looks up the type of stack item `x` in `block`.

     - block - a Block.
     - x - int or Dynamic. See the docstring for Stack.
    '''
    if _is_runtime(x):
        return block.stack_slots[x].type_
    else:
        return x.type_


def _lookup_constness(block, x):
    '''
    Looks up the constness of stack item `x` in `block`.

     - block - a Block.
     - x - int or Dynamic. See the docstring for Stack.
    '''
    if _is_runtime(x):
        return block.stack_slots[x].constness
    else:
        return x.constness


def _reify(block, *stack_items):
    '''
    Emits 'static' instructions for all of the items in `stack_items`
    which are not already on the runtime stack, i.e. for the items that
    are static (those represented by Dynamics).

     - stack_items - a list of int or Dynamic representing the top few
       items of the flat code stack. See the docstring for Stack.
    '''
    assert type(block) is typed_code.Block
    # WONTFIX: Quadratic runtime cost. REASON: Not worth the effort.
    for depth, item in enumerate(reversed(stack_items)):
        if type(item) is not int:
            assert type(item) is Dynamic
            block.static(item, depth)


def _check_type(
    block,
    value,
    required_type,
    loc,
    noun,
    message='%s must be %s',
):
    '''
    Checks that `value` is a static constant, and raises a StaticnessVMError
    if not. Then, checks that it is compatible with `required_type` and raises
    a StaticVMError if not. If all is well, returns `value.ref`, i.e. a Ref.

     - block - a Block.
     - value - an int or a Dynamic: the value to test.
     - required_type - a Ref to the type of value to return, or `None` if the
       type has already been checked.
     - noun - a noun phrase that can be used in error messages, e.g.
       'condition used for "if"' or 'type used for cast'.
     - message - an error message containing two "%s" markers. The first will
       be replaced with `noun`. The second will be replaced with the word
       "static" or the word "constant".
    '''
    constness = _lookup_constness(block, value)
    if const.switch(constness) != 'CONST':
        raise StaticnessVMError(message % (noun, "constant"), loc)
    if _is_runtime(value):
        raise StaticnessVMError(message % (noun, "static"), loc)
    assert type(value) is Dynamic and value.is_const()
    if required_type is not None:
        value.check_is_a(required_type, noun, loc)
    return value.ref


def _select_case_quad(tag, case_quads):
    '''
    Searches `case_quads` for `tag`. Returns the case triple, if found, or
    `None` otherwise.

    Raises UnreachableCaseVMError if more than one case triple has the same
    tag.

     - case_quads - a list of (tag, arity, flat code, loc) quads.
    '''
    assert type(case_quads) is list
    ret = None
    seen = set() # Set of tags.
    for case_quad in case_quads:
        tag2, _, _, case_loc = case_quad
        if tag2 in seen:
            raise UnreachableCaseVMError(tag2, case_loc)
        seen.add(tag2)
        if tag2 == tag:
            ret = case_quad
    return ret


def _macro_int_iter(block, stack, x, loc):
    _reify(block, x)
    block.static(iterator.int_field, 1)
    block.call_function(1, loc)
    stack.push_runtime()


def _macro_array_iter(block, stack, x, loc):
    _reify(block, x)
    x_t = _lookup_type(block, x)
    block.static(iterator.array_field(x_t), 1)
    block.call_function(1, loc)
    stack.push_runtime()


def _macro_array_slice(block, stack, x, loc):
    param_types = Array.from_list(
        [wtypes.int_.read(), wtypes.int_.read()],
        wtypes.sizeof(wtypes.type_),
    )
    if _is_static_const(x):
        # Do it at compile time.
        def make_body(block2):
            block2.unpack_struct(2, loc)
            block2.static(x, 2)
            block2.array_slice(loc)
            block2.return_()

        result = typed_code_static.closure(make_body, param_types, loc)
        stack.push_static(result)
    else:
        # Do it at run time.
        def make_body(block2):
            block2.unpack_struct(2, loc)
            block2.capture(0)
            block2.bury(2)
            block2.array_slice(loc)
            block2.return_()

        _reify(block, x)
        block.closure(make_body, param_types, 1, loc)
        stack.push_runtime()


# TODO: Move.
def _macro_str_length(block, stack, x, loc):
    if _is_static_const(x):
        # Do it at compile time.
        result = typed_code_static.str_length(x, loc)
        stack.push_static(result)
    else:
        # Do it at run time.
        _reify(block, x)
        block.str_length(loc)
        stack.push_runtime()


# TODO: Move.
def _macro_array0_length(block, stack, x, loc):
    if _is_static_const(x):
        # Do it at compile time.
        stack.push_static(_zero)
    else:
        # Do it at run time.
        # Happens e.g. in "var a = []; a.length"
        _reify(block, x)
        block.drop()
        block.static(_zero, 0)
        stack.push_runtime()


# TODO: Move.
def _macro_array_length(block, stack, x, loc):
    if _is_static_const(x):
        # Do it at compile time.
        result = typed_code_static.array_length(x)
        stack.push_static(result)
    else:
        # Do it at run time.
        _reify(block, x)
        block.array_length()
        stack.push_runtime()


def _field_for_type(type_, name):
    '''
    Works out how to do `x.name` given the type of `x`. Returns `None` or a
    "macro", i.e. a function which takes the following arguments:

     - block - a typed_code.Block to which to append instructions.
     - stack - a Stack to update as the instructions are generated.
     - x - int or Dynamic. See the docstring for Stack.
     - loc - an opaque source code location for use in error messages.
    '''
    t0, _ = wtypes.switch(type_)
    if t0 == 'INT':
        if name == 'iter':
            return _macro_int_iter
    elif t0 == 'STR':
        if name == 'length':
            return _macro_str_length
    elif t0 == 'ARRAY0':
        if name == 'iter':
            raise_message(
                '"x.iter" is not defined when x is an array0; '
                'cast x to the desired array type'
            )
        elif name == 'slice':
            raise_message(
                '"x.slice" is not defined when x is an array0; '
                'cast x to the desired array type'
            )
        elif name == 'length':
            return _macro_array0_length
    elif t0 == 'ARRAY':
        if name == 'iter':
            return _macro_array_iter
        elif name == 'slice':
            return _macro_array_slice
        elif name == 'length':
            return _macro_array_length
    elif t0 == 'UNION':
        raise NotImplementedError("Field of a union value")
    elif t0 == 'DYNAMIC':
        raise NotImplementedError("Field of a dynamic value")
    return None


# Useful static constants.

_empty_array = Dynamic.new(wtypes.array0, Array.malloc(0, 0).to_welly())
_empty_struct = Dynamic.new(wtypes.void, ())
_type_type = builtins['type']
_true = builtins['true']
_false = builtins['false']
_zero = Dynamic.new(wtypes.int_, (0,))

# Make a no-op implementation of `on_output`.

def _make_on_output_body(block):
    # The 0-length argument struct becomes the return value.
    block.return_()

_default_on_output = typed_code_static.closure(
    _make_on_output_body,
    Array.from_list(
        [wtypes.bool_.read(), wtypes.dynamic.read()],
        wtypes.sizeof(wtypes.type_),
    ),
    None,
)


class Stack(object):
    '''
    Represents the mapping from a list of flat code values to a list of typed
    code values. Stacks are used by the linker to represent the stack, the
    locals, and the captures. Each element of a Stack is one of:

     - a Dynamic, representing a static (compile-time-known) value.
     - an int, representing the position of the value in the corresponding
       typed code data structure.

    Public fields:
     - runtime_count - the number of non-compile-time-known values, i.e. the
       size of the corresponding typed code data structure.
    '''
    # TODO: Use `None` instead of integers.
    # Compute stack positions another way.
    def __init__(self):
        self._items = [] # Mixed list of int and Dynamic.
        self.runtime_count = 0

    def __len__(self):
        return len(self._items)

    def __getitem__(self, index):
        return self._items[index]

    def __pretty__(self):
        return (
            'Stack([',
            ', ',
            '])',
            [auto_wrap(x) for x in self._items],
        )

    def copy(self):
        result = Stack()
        result._items = self._items[:]
        result.runtime_count = self.runtime_count
        return result

    def push_static(self, x):
        '''
        Pushes onto the flat code stack (or locals or captures) a value that
        is known at compile time. There will not be a corresponding value on
        the typed code stack.

         - x - a Dynamic.
        '''
        assert type(x) is Dynamic
        self._items.append(x)

    def push_runtime(self):
        '''
        Pushes onto the flat code stack (or locals or captures) a value that
        is not known at compile time. There will be a corresponding value on
        the typed code stack.
        '''
        self._items.append(self.runtime_count)
        self.runtime_count += 1

    def push_runtime_or_static(self, x):
        '''
        Calls `push_runtime()` or `push_static(x)` as appropriate. This is
        useful for pushing a value back on the stack that was previously
        popped.

         - x - an int (representing the old position of `x` on the typed code
           stack) or a Dynamic (representing a compile-time-known value).
           Note that in the case where `x` is an int, the value of the int is
           ignored.
        '''
        if _is_runtime(x):
            self.push_runtime()
        else:
            self.push_static(x)

    def pop_many(self, n, error_message):
        '''
        Pops `n` items from the flat code stack (or locals or captures).
        Returns a mixed list of int and Dynamic.

        Ordering: If the stack is `..., x, y, z` then `pop_many(2)` returns
        `[y, z]`.

         - n - the number of items to pop.
         - error_message - used to construct an IllegalInstructionVMError or
           a StackUnderflowVMError if necessary.
        '''
        new_length = len(self._items) - n
        if n < 0:
            raise IllegalInstructionVMError(error_message)
        if new_length < 0:
            raise StackUnderflowVMError(error_message)
        result = self._items[new_length:]
        self._items[new_length:] = []
        for x in result:
            if _is_runtime(x):
                self.runtime_count -= 1
        return result


# Main entry points.
# See "doc/compiler/compiler-api.txt" for enumeration of cases.

# Cases 1 and 2.
def link_statement(flat_code, captures, on_output):
    '''
    Links a top-level statement and returns the resulting typed code and its
    type information.

    Params:
     - flat_code - a list of flat code instructions to run.
     - captures - a list of Dynamic, providing values for use by
       flat code 'capture' instructions. These are used to represent values
       mentioned in the code but defined outside it.
     - on_output - `None` or a Dynamic representing a Welly function to be
       used to print out expressions used as statements. It takes:
        - is_static - bool.
        - value - dynamic.
       and returns void.

    Returns `(typed code, cts)` where:
     - `typed_code` has the following interface:
        - On entry, the stack is empty, and there are no locals or captures.
        - On exit, the stack holds the values that `flat_code` leaves on the
          stack.
     - `cts` is a list of (Ref to constness, Ref to type) pairs, describing
       the values left on the stack.
    '''
    # Link and type-check the code.
    captures_stack = Stack()
    for x in captures:
        captures_stack.push_static(x)
    block = typed_code.Block()
    exit_ = link(block, flat_code, captures_stack, Stack(), on_output)
    if not subtypes.eq(block.get_return_type(None), wtypes.never):
        raise IllegalInstructionVMError(
            '"return" statement is not in a function')
    # Reify the newly defined values.
    for depth, item in enumerate(reversed(exit_.stack)):
        if type(item) is not int:
            assert type(item) is Dynamic
            block.static(item, depth)
    # Capture the constnesses and types.
    cts = [(slot.constness, slot.type_) for slot in block.stack_slots]
    assert len(exit_.stack) == len(cts)
    return block.get_code(is_break_allowed=False), cts


# Case 3.
def link_eval(flat_code, TODO):
    '''
    Links an expression passed to `eval()`.
     - flat_code - a list of flat code instructions to run.
    '''
    raise NotImplementedError("eval")


# Case 4.
def link_function_call(
    block, # From the type-checker.
    flat_code, # From 'closure' instruction.
    captures, # From the linker.
    on_output, # From the linker.
):
    '''
    Link a function. Substitutes in the static captures, performs
    constant-folding and type-checking, and populates `block` by generating
    typed code.

    Params:
     - block - a typed_code.Block to populate with the function body. 
     - flat_code - a list of flat code instructions representing the
       function body (as included in the 'closure' instruction).
     - captures - a Stack, providing values for use by flat code 'capture'
       instructions. These are used to represent values mentioned in the code
       but defined outside it. Elements of type int are indexes into
       `block.captures`.
     - on_output - `None` or a Dynamic representing a Welly function to be
       used to print out expressions used as statements. It takes:
        - is_static - bool.
        - value - dynamic.
       and returns void.

    On entry to the code in `block`:
     - The stack contains the argument struct (a single value). It is CONST,
       and its type is `struct(param_types)` where `param_types` come from the
       function declaration.
     - There are no locals.
     - The captures correspond to the int values in `captures`, i.e. to the
       run-time values mentioned in the flat code but defined outside it.
       Their types and constnesses come from the type-checker.
    The code in `block` must not fall through. It can exit by executing a
    'return' instruction.
    '''
    assert len(block.stack_slots) == 1
    assert len(block.locals_) == 0
    assert len(block.captures) == captures.runtime_count
    # Unpack the argument struct.
    n = struct_length(block.stack_slots[0].type_)
    block.unpack_struct(n, None)
    if n > 1:
        block.reverse(n)
    locals_ = Stack()
    for _ in range(n):
        block.init(False)
        locals_.push_runtime()
    # Link the body.
    link(block, flat_code, captures, locals_, on_output)


# Case 5.
def link_macro_call(flat_code, TODO):
    '''
    Call a macro.
     - flat_code - a list of flat code instructions representing the macro
       body.
    '''
    raise NotImplementedError("macro call")


class Exit(object):
    '''
    Represents all the ways in which a block of code can exit, and records
    information about the values it computes. This is the return type of
    `link()`.

    Public fields:
     - stack - a Stack representing the values left on the stack when
       the code exits normally.
    '''
    def __init__(self, stack):
        self.stack = stack

    def __pretty__(self):
        return (
            'Exit(',
            ', ',
            ')',
            [auto_wrap(self.stack)]
        )


# All cases call this.
def link(block, flat_code, captures, locals_, on_output):
    '''
    Links `flat_code`. Substitutes in the static captures, performs
    constant-folding and type-checking, and populates `block` by generating
    typed code. Returns an Exit representing all the ways in which the code
    can exit.

    Parameters:
     - block - a typed_code.Block to populate with the instructions.
     - flat_code - a list of flat code instructions.
     - captures - a Stack, providing values for use by flat code 'capture'
       instructions. These are used to represent values mentioned in
       the code but defined outside it.
     - locals_ - a Stack, representing the flat code local variables. These
       are used by 'local' instructions, and 'init' instructions can create
       new ones.
     - on_output - `None` or a Dynamic representing a Welly function to be
       used to print out expressions used as statements. It takes:
        - is_static - bool.
        - value - dynamic.
       and returns void.
    '''
    assert type(block) is typed_code.Block
    assert type(flat_code) is list
    assert type(captures) is Stack
    assert type(locals_) is Stack
    if on_output is None:
        on_output = _default_on_output
    assert type(on_output) is Dynamic
    stack = Stack()
    assert captures.runtime_count == len(block.captures)
    assert locals_.runtime_count == len(block.locals_)
    assert stack.runtime_count == len(block.stack_slots)
    for instr in flat_code:
        i0 = instr[0]
        if not block.is_end_reachable:
            if i0 == 'if_reachable':
                continue
            else:
                loc = instr[-1]
                raise UnreachableCodeVMError(loc)
        if i0 == 'address_of':
            _, loc = instr
            (x,) = stack.pop_many(1, 'address_of')
            if _is_runtime(x):
                # Do it at runtime.
                _reify(block, x) # No-op.
                block.address_of()
                stack.push_runtime()
            else:
                # Do it at compile time. Even if it is var.
                t = wtypes.ref(x.constness, x.type_)
                result = Dynamic.new(t, (x.ref,))
                stack.push_static(result)
        elif i0 == 'array':
            _, length, loc = instr
            items = stack.pop_many(length, 'array')
            if all(_is_static_const(x) for x in items):
                # Do it at compile time.
                if length == 0:
                    result = typed_code_static.array0()
                    stack.push_static(result)
                else:
                    elem_type = subtypes.join_many(
                        [x.type_ for x in items],
                        'the array element type',
                        loc,
                    )
                    result = typed_code_static.array(elem_type, items, loc)
                    stack.push_static(result)
            else:
                # Do it at run time.
                assert length > 0
                _reify(block, *items)
                elem_type = subtypes.join_many(
                    [x.type_ for x in block.stack_slots[-length:]],
                    'the array element type',
                    loc,
                )
                block.array(elem_type, length, loc)
                stack.push_runtime()
        elif i0 == 'assign':
            _, loc = instr
            rvalue, lvalue = stack.pop_many(2, 'assign')
            _reify(block, rvalue, lvalue)
            block.assign(loc)
        elif i0 == 'binary':
            _, op, loc = instr
            x, y = stack.pop_many(2, 'binary')
            x_t = _lookup_type(block, x)
            y_t = _lookup_type(block, y)
            # Dispatch.
            operator = operators.lookup(op, x_t, y_t)
            if operator is not None:
                if _is_static_const(x) and _is_static_const(y):
                    # Do it at compile time.
                    result = typed_code_static.binary(operator, x, y, loc)
                    stack.push_static(result)
                else:
                    # Do it at run time.
                    _reify(block, x, y)
                    block.binary(operator, loc)
                    stack.push_runtime()
            elif (
                op == '+' and
                is_array_type(x_t) and
                is_array_type(y_t)
            ):
                if wtypes.switch(x_t)[0] == 'ARRAY0':
                    raise_message(
                        '"x + y" is not defined when x is an array0; '
                        'cast x to the desired array type',
                        loc=loc,
                    )
                else:
                    # It's a non-empty array type.
                    if (_is_static_const(x) and
                        _is_static_const(y) and
                        const.switch(array_constness(x_t)) == 'CONST' and
                        const.switch(array_constness(y_t)) == 'CONST'
                    ):
                        # Do it at compile time.
                        result = typed_code_static.array_concat(x, y, loc)
                        stack.push_static(result)
                    else:
                        # Do it at run time.
                        _reify(block, x, y)
                        block.array_concat(loc)
                        stack.push_runtime()
            elif (
                op == '*' and
                is_array_type(x_t) and
                is_int_type(y_t)
            ):
                if wtypes.switch(x_t)[0] == 'ARRAY0':
                    if _is_static_const(x) and _is_static_const(y):
                        # Do it at compile time.
                        result = typed_code_static.array0_repeat(x, y, loc)
                        stack.push_static(result)
                    else:
                        # Do it at run time.
                        _reify(block, x, y)
                        block.array0_repeat(loc)
                        stack.push_runtime()
                else:
                    # It's a non-empty array type.
                    if (_is_static_const(x) and
                        _is_static_const(y) and
                        const.switch(array_constness(x_t)) == 'CONST'
                    ):
                        # Do it at compile time.
                        result = typed_code_static.array_repeat(x, y, loc)
                        stack.push_static(result)
                    else:
                        # Do it at run time.
                        _reify(block, x, y)
                        block.array_repeat(loc)
                        stack.push_runtime()
            elif (
                op == '*' and
                is_int_type(x_t) and
                is_array_type(y_t)
            ):
                if wtypes.switch(y_t)[0] == 'ARRAY0':
                    if _is_static_const(x) and _is_static_const(y):
                        # Do it at compile time.
                        result = typed_code_static.array0_repeat(y, x, loc)
                        stack.push_static(result)
                    else:
                        # Do it at run time.
                        _reify(block, x, y)
                        block.swap()
                        block.array0_repeat(loc)
                        stack.push_runtime()
                else:
                    # It's a non-empty array type.
                    if (_is_static_const(x) and
                        _is_static_const(y) and
                        const.switch(array_constness(y_t)) == 'CONST'
                    ):
                        # Do it at compile time.
                        result = typed_code_static.array_repeat(y, x, loc)
                        stack.push_static(result)
                    else:
                        # Do it at run time.
                        _reify(block, x, y)
                        block.swap()
                        block.array_repeat(loc)
                        stack.push_runtime()
            else:
                raise BinaryOperatorVMError(op, x_t, y_t, loc)
        elif i0 == 'box':
            _, loc = instr
            (x,) = stack.pop_many(1, 'box')
            if _is_runtime(x):
                # Do it at run time.
                block.box()
                stack.push_runtime()
            else: # Need not be const.
                # Do it at compile time.
                result = typed_code_static.box(x)
                stack.push_static(result)
        elif i0 == 'break':
            _, loc = instr
            # TODO: Nice error messages.
            block.break_()
        elif i0 == 'bury':
            _, n, loc = instr
            if n < 0:
                raise IllegalInstructionVMError('bury index is negative')
            if n >= len(stack):
                raise StackUnderflowVMError('bury')
            (x,) = stack.pop_many(1, 'bury')
            ys = stack.pop_many(n, 'bury')
            if _is_runtime(x):
                # Do it at run time.
                block.bury(sum(_is_runtime(y) for y in ys))
                stack.push_runtime()
            else: # Need not be const.
                # Do it at compile time.
                stack.push_static(x)
            for y in ys:
                stack.push_runtime_or_static(y)
        elif i0 == 'call':
            _, num_args, loc = instr
            args = stack.pop_many(num_args, 'call')
            (func,) = stack.pop_many(1, 'call')
            # Switch on the function type.
            func_t = _lookup_type(block, func)
            ft0, _ = wtypes.switch(func_t)
            if ft0 == 'FUNCTION':
                # Do it at run time.
                _reify(block, func, *args)
                block.call_function(num_args, loc)
                stack.push_runtime()
            elif ft0 == 'MACRO':
                raise NotImplementedError("macro call")
            else:
                raise CallTypeVMError(func_t, loc)
        elif i0 == 'capture':
            _, index, loc = instr
            if index < 0:
                raise IllegalInstructionVMError('capture index is negative')
            if index >= len(captures):
                raise IllegalInstructionVMError('capture index is too large')
            x = captures[index]
            if _is_runtime(x):
                # Do it at run time.
                block.capture(x)
                stack.push_runtime()
            else:
                # Do it at compile time.
                stack.push_static(x)
        elif i0 == 'cast':
            _, loc = instr
            x, t = stack.pop_many(2, 'cast')
            t = _check_type(
                block,
                t,
                wtypes.type_,
                loc,
                'the type used for the cast',
            )
            if _is_static_const(x):
                # Do it at compile time.
                result = typed_code_static.cast(t, x, loc)
                stack.push_static(result)
            else:
                # Do it at run time.
                _reify(block, x)
                block.cast(t, loc)
                stack.push_runtime()
        elif i0 == 'cast_to_type':
            _, noun, require_static, loc = instr
            (x,) = stack.pop_many(1, 'cast_to_type')
            x_t = _lookup_type(block, x)
            subtypes.check(x_t, wtypes.type_, noun, loc)
            if _is_static_const(x):
                # Do it at compile time.
                result = typed_code_static.cast(wtypes.type_, x, loc)
                stack.push_static(result)
            else:
                # Do it at run time.
                if require_static:
                    _check_type(block, x, None, loc, noun)
                _reify(block, x)
                block.cast(wtypes.type_, loc)
                stack.push_runtime()
        elif i0 == 'closure':
            _, body_code, n, loc = instr
            # Pop captured values.
            body_captures = Stack()
            new_n = 0
            for x in stack.pop_many(n, 'closure'):
                if _is_runtime(x):
                    body_captures.push_runtime()
                    new_n += 1
                else:
                    body_captures.push_static(x)
            # Pop parameter types.
            (param_types,) = stack.pop_many(1, 'closure')
            param_types = _check_type(
                block,
                param_types,
                wtypes.array_of_type,
                loc,
                'the parameter type array of the function',
            ).read()
            param_types = Array.from_welly(
                param_types,
                wtypes.sizeof(wtypes.type_),
            )
            # Swich on staticness of captures.
            if new_n != 0:
                # Some captures are non-static. Do it at run time.
                block.closure(
                    lambda block2: link_function_call(
                        block2,
                        body_code,
                        body_captures,
                        on_output,
                    ),
                    param_types,
                    new_n,
                    loc,
                )
                stack.push_runtime()
            else:
                # All captures are static. Do it at compile time.
                result = typed_code_static.closure(
                    lambda block2: link_function_call(
                        block2,
                        body_code,
                        body_captures,
                        on_output,
                    ),
                    param_types,
                    loc,
                )
                stack.push_static(result)
        elif i0 == 'const_char':
            _, x, loc = instr
            if type(x) is not int:
                raise IllegalInstructionVMError('const_char')
            stack.push_static(Dynamic.new(wtypes.int_, (x,)))
        elif i0 == 'const_float':
            _, x, loc = instr
            if type(x) is not float:
                raise IllegalInstructionVMError('const_float')
            stack.push_static(Dynamic.new(wtypes.float_, (x,)))
        elif i0 == 'const_int':
            _, x, loc = instr
            if type(x) is not int:
                raise IllegalInstructionVMError('const_int')
            stack.push_static(Dynamic.new(wtypes.int_, (x,)))
        elif i0 == 'const_str':
            _, x, loc = instr
            if type(x) is not str:
                raise IllegalInstructionVMError('const_str')
            stack.push_static(Dynamic.new(wtypes.str_, (x,)))
        elif i0 == 'continue':
            _, loc = instr
            # TODO: Nice error messages.
            block.continue_()
        elif i0 == 'defer':
            _, loc = instr
            (t,) = stack.pop_many(1, 'defer')
            t = _check_type(
                block,
                t,
                wtypes.type_,
                loc,
                'the type used for "defer"'
            )
            block.defer(t, loc)
            stack.push_runtime()
        elif i0 == 'defer_type':
            _, loc = instr
            block.defer(wtypes.type_, loc)
            stack.push_runtime()
        elif i0 == 'dereference':
            _, loc = instr
            (x,) = stack.pop_many(1, 'dereference')
            x_t = _lookup_type(block, x)
            if wtypes.switch(x_t)[0] != 'REF':
                raise DereferenceTypeVMError(x_t, loc)
            if _is_static_const(x):
                # Do it at compile time.
                result = typed_code_static.dereference(x)
                stack.push_static(result)
            else:
                # Do it at run time.
                _reify(block, x)
                block.dereference()
                stack.push_runtime()
        elif i0 == 'drop':
            _, loc = instr
            (x,) = stack.pop_many(1, 'drop')
            if _is_runtime(x):
                # Do it at run time.
                block.drop()
            else:
                # Do it at compile time.
                pass
        elif i0 == 'expr':
            _, loc = instr
            (x,) = stack.pop_many(1, 'expr')
            x_t = _lookup_type(block, x)
            if is_subtype(x_t, wtypes.void):
                # Special case: suppress values of type `void`.
                if _is_runtime(x):
                    block.drop()
            else:
                if _is_runtime(x):
                    block.box()
                    block.static(_false, 1)
                else:
                    result = typed_code_static.box(x)
                    block.static(result, 0)
                    block.static(_true, 1)
                block.static(on_output, 2)
                block.call_function(2, loc)
                block.drop()
        elif i0 == 'field':
            _, name, loc = instr
            if type(name) is not str:
                raise IllegalInstructionVMError('field')
            (x,) = stack.pop_many(1, 'field')
            x_t = _lookup_type(block, x)
            # Switch on type.
            t0, _ = wtypes.switch(x_t)
            if t0 == 'DYNAMIC':
                # TODO: Try to move this into `_field_for_type()`.
                if name == 'value':
                    dynamic = _check_type(
                        block,
                        x,
                        wtypes.dynamic,
                        loc,
                        'the dynamic value',
                        message="%s must be %s in order to access its value",
                    )
                    result = Dynamic.from_welly(dynamic)
                    stack.push_static(result)
                elif name == 'type':
                    if _is_static_const(x):
                        # Do it at compile time.
                        result = typed_code_static.dynamic_type(x)
                        stack.push_static(result)
                    else:
                        # Do it at run time.
                        _reify(block, x)
                        block.dynamic_type()
                        stack.push_runtime()
                elif name == 'constness':
                    if _is_static_const(x):
                        # Do it at compile time.
                        result = typed_code_static.dynamic_constness(x)
                        stack.push_static(result)
                    else:
                        # Do it at run time.
                        _reify(block, x)
                        block.dynamic_constness()
                        stack.push_runtime()
                else:
                    raise FieldAccessVMError(x_t, name, loc)
            else:
                # Normal case. Choose a macro and apply it to `x`.
                macro = _field_for_type(x_t, name)
                if macro is not None:
                    macro(block, stack, x, loc)
                else:
                    raise FieldAccessVMError(x_t, name, loc)
        elif i0 == 'fulfil':
            _, loc = instr
            rvalue, lvalue = stack.pop_many(2, 'fulfil')
            _reify(block, rvalue, lvalue)
            block.fulfil(loc)
        elif i0 == 'if':
            _, n, then_code, else_code, loc = instr
            # TODO: Refactor. Move the type check outside the "if" and check
            # staticness only once.
            (x,) = stack.pop_many(1, 'if')
            if _is_static_const(x):
                # Do it at compile time.
                (x,) = _check_type(
                    block,
                    x,
                    wtypes.bool_,
                    loc,
                    'the "if" condition',
                ).read()
                assert type(x) is bool
                exit_ = block.inline(lambda block2: link(
                    block2,
                    then_code if x else else_code,
                    captures,
                    locals_.copy(),
                    on_output,
                ))
                if len(exit_.stack) != n:
                    raise IllegalInstructionVMError(
                        '"if" has the wrong number of results')
                for x in exit_.stack:
                    stack.push_runtime_or_static(x)
            else:
                # Do it at run time.
                x_t = _lookup_type(block, x)
                subtypes.check(x_t, wtypes.bool_, 'the "if" condition', loc)
                _reify(block, x)

                def recurse(block2, flat_code):
                    exit_ = link(
                        block2,
                        flat_code,
                        captures,
                        locals_.copy(),
                        on_output,
                    )
                    if len(exit_.stack) != n:
                        raise IllegalInstructionVMError(
                            '"if" has the wrong number of results')
                    _reify(block2, *exit_.stack)
                    return exit_

                then_exit, else_exit = block.if_(
                    n,
                    lambda block2: recurse(block2, then_code),
                    lambda block2: recurse(block2, else_code),
                    loc,
                )
                for _ in range(n):
                    stack.push_runtime()
                recurse = None # Cannot "del recurse".
        elif i0 == 'if_reachable':
            _, inline_code, loc = instr
            exit_ = block.inline(lambda block2: link(
                block2,
                inline_code,
                captures,
                locals_.copy(),
                on_output,
            ))
            if len(exit_.stack) != 0:
                raise IllegalInstructionVMError(
                    '"If-Reachable" code creates values')
        elif i0 == 'init':
            _, is_var, loc = instr
            assert type(is_var) == bool
            (x,) = stack.pop_many(1, 'init')
            if _is_static_const(x) and not is_var:
                # Do it at compile time.
                locals_.push_static(x)
            else:
                # Do it at run time.
                _reify(block, x)
                block.init(is_var)
                locals_.push_runtime()
        elif i0 == 'inline':
            _, inline_code, loc = instr
            exit_ = block.inline(lambda block2: link(
                block2,
                inline_code,
                captures,
                locals_.copy(),
                on_output,
            ))
            for x in exit_.stack:
                stack.push_runtime_or_static(x)
        elif i0 == 'local':
            _, index, loc = instr
            if index < 0:
                raise IllegalInstructionVMError('"local" index is negative')
            if index >= len(locals_):
                raise IllegalInstructionVMError('"local" index is too large')
            x = locals_[index]
            if _is_runtime(x):
                # Do it at run time.
                block.local(x)
                stack.push_runtime()
            else:
                # Do it at compile time.
                stack.push_static(x)
        elif i0 == 'loop':
            _, loop_code, loc = instr
            exit_ = block.loop(lambda block2: link(
                block2,
                loop_code,
                captures,
                locals_.copy(),
                on_output,
            ))
            assert len(exit_.stack) == 0
        elif i0 == 'new':
            _, loc = instr
            (x,) = stack.pop_many(1, 'new')
            _reify(block, x)
            block.new()
            stack.push_runtime()
        elif i0 == 'pick':
            _, n, loc = instr
            if n < 0:
                raise IllegalInstructionVMError('pick index is negative')
            if n >= len(stack):
                raise StackUnderflowVMError('pick')
            x = stack[len(stack) - 1 - n]
            if _is_runtime(x):
                # Do it at run time.
                block.pick(len(block.stack_slots) - 1 - x)
                stack.push_runtime()
            else:
                # Do it at compile time.
                stack.push_static(x)
        elif i0 == 'return':
            _, loc = instr
            (x,) = stack.pop_many(1, 'return')
            if len(stack) != 0:
                raise IllegalInstructionVMError(
                    'stack is not empty after return')
            _reify(block, x)
            block.return_()
        elif i0 == 'return_void':
            _, loc = instr
            (return_type,) = stack.pop_many(1, 'return')
            if len(stack) != 0:
                raise IllegalInstructionVMError(
                    'stack is not empty after return')
            return_type = _check_type(
                block,
                return_type,
                wtypes.type_,
                loc,
                'the return type',
            )
            if not is_subtype(wtypes.void, return_type):
                # Cosmetic error message instead of TypeVMError.
                raise_message('missing return value', loc=loc)
            # Simulate "struct(0) swap cast return".
            result = typed_code_static.cast(return_type, _empty_struct, loc)
            block.static(result, 0)
            block.return_()
        elif i0 == 'reverse':
            _, n, loc = instr
            items = stack.pop_many(n, 'reverse')
            num_runtime = sum(_is_runtime(x) for x in items)
            if num_runtime >= 2:
                block.reverse(num_runtime)
            for x in reversed(items):
                stack.push_runtime_or_static(x)
        elif i0 == 'struct':
            _, n, loc = instr
            values = stack.pop_many(n, 'struct')
            if all(_is_static_const(x) for x in values):
                # Do it at compile time.
                result = typed_code_static.struct(values)
                stack.push_static(result)
            else:
                # Do it at run time.
                _reify(block, *values)
                block.struct(n)
                stack.push_runtime()
        elif i0 == 'subscript':
            _, loc = instr
            collection, index = stack.pop_many(2, 'subscript')
            collection_t = _lookup_type(block, collection)
            tag, _ = wtypes.switch(collection_t)
            if tag == 'ARRAY':
                if _is_static_const(collection) and _is_static_const(index):
                    # Do it at compile time.
                    result = typed_code_static.subscript_array(
                        collection, index, loc)
                    stack.push_static(result)
                else:
                    # Do it at run time.
                    _reify(block, collection, index)
                    block.subscript_array(loc)
                    stack.push_runtime()
            elif tag == 'STRUCT':
                index_ref = _check_type(
                    block,
                    index,
                    wtypes.int_,
                    loc,
                    'the struct index',
                )
                (index_value,) = index_ref.read()
                if _is_static_const(collection):
                    # Do it at compile time.
                    result = typed_code_static.subscript_struct(
                        index_value, collection, loc)
                    stack.push_static(result)
                else:
                    # Do it at run time.
                    _reify(block, collection)
                    block.subscript_struct(index_value, loc)
                    stack.push_runtime()
            else:
                raise SubscriptTypeVMError(collection_t, loc=loc)
        elif i0 == 'switch':
            _, case_quads, else_pair, loc = instr
            (x,) = stack.pop_many(1, 'switch')
            x_t = _lookup_type(block, x)
            if wtypes.switch(x_t)[0] != 'UNION':
                raise SwitchTypeVMError(x_t, loc=loc)
            unhandled_cases = union_cases(x_t)
            if _is_static_const(x):
                # Do it at compile time.
                tag, payload = defer.read(x.ref)
                assert type(tag) is str, x
                payload = Ref.block(payload)
                # Loop through case_quads looking for `tag`.
                case_quad = _select_case_quad(tag, case_quads)
                if case_quad is None:
                    # If the else clause does not exist, report an error.
                    if else_pair is None:
                        raise InexhaustiveSwitchVMError(tag, loc=loc)
                    else_code, else_loc = else_pair
                    # Inline the else clause.
                    exit_ = block.inline(lambda block2: link(
                        block2,
                        else_code,
                        captures,
                        locals_.copy(),
                        on_output,
                    ))
                else:
                    _, num_params, case_code, case_loc = case_quad
                    # Tag must be found, because `x` has been type-checked.
                    types = unhandled_cases[tag]
                    # Combine `payload` with the types from `x_t_case`.
                    if len(types) != num_params:
                        raise NumberOfFieldsVMError(
                            num_params, len(types), loc=case_loc)
                    payload_t = wtypes.struct(types)
                    payload_struct = Dynamic(payload_t, payload)
                    # Declare the pattern arguments as local variables.
                    locals2 = locals_.copy()
                    for field in struct_iter(payload_struct):
                        locals2.push_static(field)
                    # Inline `case_code`.
                    exit_ = block.inline(lambda block2: link(
                        block2,
                        case_code,
                        captures,
                        locals2,
                        on_output,
                    ))
            else:
                # Do it at run time.
                _reify(block, x)
                # Make callbacks to compile the case blocks.
                make_cases = []
                for tag, num_params, case_code, case_loc in case_quads:
                    try:
                        types = unhandled_cases.pop(tag)
                    except KeyError:
                        raise UnreachableCaseVMError(tag, case_loc)
                    if len(types) != num_params:
                        raise NumberOfFieldsVMError(
                            num_params, len(types), loc=case_loc)
                    locals2 = locals_.copy()
                    for _ in types:
                        locals2.push_runtime()
                    # `c` and `l` solve the "lambda in a loop" problem:
                    # http://stackoverflow.com/a/19837683
                    make_cases.append((
                        tag,
                        lambda block2, c=case_code, l=locals2:
                            link(
                                block2,
                                c,
                                captures,
                                l,
                                on_output,
                            )
                    ))
                # Make a callback to compile the else block.
                if else_pair is not None:
                    else_code, else_loc = else_pair
                    if not unhandled_cases:
                        raise UnreachableCaseVMError("else", else_loc)
                    make_else = lambda block2: link(
                        block2,
                        else_code,
                        captures,
                        locals_.copy(),
                        on_output,
                    )
                else:
                    for tag in unhandled_cases:
                        raise InexhaustiveSwitchVMError(tag, loc=loc)
                    make_else = None
                # Compile a 'switch' instruction.
                case_exits, else_exit = block.switch(
                    make_cases, make_else, loc)
                if else_pair is None:
                    assert else_exit is None
        elif i0 == 'typeof':
            _, loc = instr
            (x,) = stack.pop_many(1, 'typeof')
            x_t = _lookup_type(block, x)
            if _is_runtime(x):
                block.drop()
            result = Dynamic(wtypes.type_, x_t)
            stack.push_static(result)
        elif i0 == 'unary':
            _, op, loc = instr
            (x,) = stack.pop_many(1, 'unary')
            x_t = _lookup_type(block, x)
            operator = operators.lookup(op, x_t)
            if operator is not None:
                if _is_static_const(x):
                    # Do it at compile time.
                    result = typed_code_static.unary(operator, x, loc)
                    stack.push_static(result)
                else:
                    # Do it at run time.
                    _reify(block, x)
                    block.unary(operator, loc)
                    stack.push_runtime()
            else:
                raise UnaryOperatorVMError(op, x_t, loc)
        elif i0 == 'union':
            _, tag, n, loc = instr
            values = stack.pop_many(n, 'union')
            if all(_is_static_const(x) for x in values):
                # Do it at compile time.
                result = typed_code_static.union(tag, values)
                stack.push_static(result)
            else:
                # Do it at run time.
                _reify(block, *values)
                block.union(tag, n)
                stack.push_runtime()
        elif i0 == 'unpack_struct':
            _, n, loc = instr
            (x, ) = stack.pop_many(1, 'unpack_struct')
            x_t = _lookup_type(block, x)
            if wtypes.switch(x_t)[0] != 'STRUCT':
                raise UnpackTypeVMError(x_t, loc=loc)
            if _is_runtime(x):
                # Do it at runtime.
                block.unpack_struct(n, loc)
                for _ in range(n):
                    stack.push_runtime()
            else:
                # Do it at compile time.
                results = typed_code_static.unpack_struct(x, n, loc)
                for result in results:
                    stack.push_static(result)
        else:
            assert False, 'Unknown flat code instruction: %s' % i0
        assert captures.runtime_count == len(block.captures)
        assert locals_.runtime_count == len(block.locals_)
        assert stack.runtime_count == len(block.stack_slots)
    if not block.is_end_reachable:
        assert len(stack) == 0
    return Exit(stack)
