'''
Lists all the "regular" operators in Welly, defines their type signature, and
provides implementations in Python.

A few operators require special treatment, and are not listed here. For
example:

 - The dereference operator "*" has a variety of result types and weird
   staticness rules.
 - The "defer" operator has a variety of result types and is never static.
 - The address of operator "&" has a variety of input types and weird
   staticness rules.
 - The "new" operator has a variety of input types and is never static.
 - The "array" and "ref" operators have a constness, and apply extra syntax
   checks to their operand.
'''

from math import copysign

from ..model.errors import ArithmeticVMError
from ..model.wtypes import bool_, int_, float_, str_, sizeof
from ..model.subtypes import is_subtype


__all__ = [
    'Operator', 'all_operators', 'index', 'lookup_operator',
]


# Abbreviations for types, retaining just enough information for selecting
# Operators.
def abbreviate(t):
    if is_subtype(t, bool_):
        return 'b'
    elif is_subtype(t, float_):
        return 'f'
    elif is_subtype(t, int_):
        return 'i'
    elif is_subtype(t, str_):
        return 's'
    else:
        return '?'


def make_opcode(op, *in_types):
    return op + ''.join(abbreviate(t) for t in in_types)


class Operator(object):
    '''
    Represents a type specialisation of one of the built-in Welly operators.
    When Welly sees an expression, e.g. "3+4", it chooses an Operator based
    on the operator symbol (e.g. "+") and the types of the operands (e.g.
    `int`, `int`).

    Public fields:
     - op - the operator symbol, e.g. "+".
     - in_types - a tuple of Refs to the Welly types of the operands, e.g.
       `(int_, int_)`.
     - result_type - a Ref to the Welly type of the result, e.g. `int_`.
     - opcode - a short string that identifies this Operator, e.g. '+ii'.
       You can use `operators.index[opcode]` to retrieve this Operator.
     - implementation - a Python function that takes a word for each operand
       and returns a word.
    '''
    def __init__(self, op, result_type, implementation, *in_types):
        assert 'all_operators' not in globals(), 'already finished'
        assert all(sizeof(t) == 1 for t in in_types)
        assert sizeof(result_type) == 1
        self.op = op
        self.in_types = in_types
        self.result_type = result_type
        self.opcode = make_opcode(op, *in_types)
        self.implementation = implementation


class Unary(Operator):
    '''
    Represents a unary (1-operand) operator whose result type and operand
    type are the same.
    '''
    def __init__(self, op, t, implementation):
        super(Unary, self).__init__(op, t, implementation, t)


class Test(Operator):
    '''
    Represents a unary (1-operand) operator whose result type is `bool`.
    '''
    def __init__(self, op, t, implementation):
        super(Test, self).__init__(op, bool_, implementation, t)


class Binary(Operator):
    '''
    Represents a binary (2-operand) operator whose result type and operand
    type are the same.
    '''
    def __init__(self, op, t, implementation):
        super(Binary, self).__init__(op, t, implementation, t, t)


class Comparison(Operator):
    '''
    Represents a binary (2-operand) operator whose result type is `bool`.
    '''
    def __init__(self, op, t, implementation):
        super(Comparison, self).__init__(op, bool_, implementation, t, t)


def integer_divide(x, y):
    try:
        return x // y
    except ZeroDivisionError:
        raise ArithmeticVMError('cannot divide by zero')


def integer_modulo(x, y):
    try:
        return x % y
    except ZeroDivisionError:
        raise ArithmeticVMError('cannot divide by zero')


def integer_power(x, y):
    if y > 100:
        raise NotImplementedError('power exceeds 100')
    try:
        return x ** y
    except ZeroDivisionError:
        raise ArithmeticVMError(
            'cannot raise a zero to a negative power')


def float_divide(x, y):
    try:
        return x / y
    except ZeroDivisionError:
        return copysign(float('inf'), x) * copysign(1.0, y)


def float_modulo(x, y):
    try:
        return x % y
    except ZeroDivisionError:
        raise ArithmeticVMError('cannot divide by zero')


def float_power(x, y):
    try:
        return x ** y
    except ZeroDivisionError:
        raise ArithmeticVMError(
            'cannot raise a zero to a negative power')
    except ValueError:
        raise ArithmeticVMError(
            'cannot raise a negative number to a fractional power')


all_operators = [
    # Boolean operators.
    Test('!', bool_, lambda x: not x),
    Binary('&', bool_, lambda x, y: x & y),
    Binary('|', bool_, lambda x, y: x | y),
    Binary('^', bool_, lambda x, y: x ^ y),
    Binary('==', bool_, lambda x, y: x == y),
    Binary('!=', bool_, lambda x, y: x != y),
    # Integer operators.
    Unary('+', int_, lambda x: x),
    Unary('-', int_, lambda x: -x),
    Unary('~', int_, lambda x: ~x),
    Test('!', int_, lambda x: x == 0),
    Binary('+', int_, lambda x, y: x + y),
    Binary('-', int_, lambda x, y: x - y),
    Binary('*', int_, lambda x, y: x * y),
    Binary('/', int_, integer_divide),
    Binary('%', int_, integer_modulo),
    Binary('&', int_, lambda x, y: x & y),
    Binary('|', int_, lambda x, y: x | y),
    Binary('^', int_, lambda x, y: x ^ y),
    Binary('<<', int_, lambda x, y: x << (y & 63)),
    Binary('>>', int_, lambda x, y: x >> (y & 63)),
    Binary('>>>', int_, lambda x, y: (x & 0xFFFFFFFFFFFFFFFF) >> (y & 63)),
    Binary('**', int_, integer_power),
    Comparison('==', int_, lambda x, y: x == y),
    Comparison('!=', int_, lambda x, y: x != y),
    Comparison('<', int_, lambda x, y: x < y),
    Comparison('<=', int_, lambda x, y: x <= y),
    Comparison('>', int_, lambda x, y: x > y),
    Comparison('>=', int_, lambda x, y: x >= y),
    Comparison('<>', int_, lambda x, y: x < y or x > y),
    # Float operators.
    Unary('+', float_, lambda x: x),
    Unary('-', float_, lambda x: -x),
    Test('!', float_, lambda x: x == 0.0),
    Binary('+', float_, lambda x, y: x + y),
    Binary('-', float_, lambda x, y: x - y),
    Binary('*', float_, lambda x, y: x * y),
    Binary('/', float_, float_divide),
    Binary('%', float_, float_modulo),
    Operator('<<', float_, lambda x, y: x * pow(2.0, y), float_, int_),
    Operator('>>', float_, lambda x, y: x * pow(0.5, y), float_, int_),
    Operator('**', float_, float_power, float_, float_),
    Comparison('==', float_, lambda x, y: x == y),
    Comparison('!=', float_, lambda x, y: x != y),
    Comparison('<', float_, lambda x, y: x < y),
    Comparison('<=', float_, lambda x, y: x <= y),
    Comparison('>', float_, lambda x, y: x > y),
    Comparison('>=', float_, lambda x, y: x >= y),
    Comparison('<>', float_, lambda x, y: x < y or x > y),
    # String operators.
    Operator('+', str_, lambda x, y: x + y, str_, str_),
    Operator('*', str_, lambda x, y: x * y, str_, int_),
    Comparison('==', str_, lambda x, y: x == y),
    Comparison('!=', str_, lambda x, y: x != y),
    Comparison('<', str_, lambda x, y: x < y),
    Comparison('<=', str_, lambda x, y: x <= y),
    Comparison('>', str_, lambda x, y: x > y),
    Comparison('>=', str_, lambda x, y: x >= y),
    Comparison('<>', str_, lambda x, y: x < y or x > y),
]


index = dict((make_opcode(x.op, *x.in_types), x) for x in all_operators)


def lookup(op, *in_types):
    '''
    Returns an Operator that implements `op` for the given `in_types`, or
    `None` if no such Operator is defined.
    '''
    return index.get(make_opcode(op, *in_types))
