import re
import sys

from ..model import pack, const, wtypes, pretty_type, pretty_value

from ..model.errors import AssertionVMError
from ..model.memory import Ref, Array
from ..model.wtypes import (
    bool_, int_, float_, str_, dynamic, void, struct, sizeof)
from ..model.ops import Dynamic
from ..printer import pretty



__all__ = [
    'Syscall', 'all_syscalls', 'index', 'log_format', 'TEST_LOG',
]


class Syscall(object):
    '''
    Represents a Welly VM system call. System calls are useful for interacting
    with the host.

     - name - the (str) name of the syscall.
     - in_type - a Ref to the Welly type of the argument value. It's usually
       a struct type.
     - out_type - a Ref to the Welly type of the result value.
     - implementation - a Python function that takes a tuple of words
       representing the argument value and returns a tuple of words
       representing the result value.
    '''
    def __init__(self, name, in_type, out_type, implementation):
        assert 'all_syscalls' not in globals(), 'already finished'
        self.name = name
        self.in_type = in_type
        self.out_type = out_type
        self.implementation = implementation


def _assert(args):
    condition, message = args
    assert type(condition) is bool
    assert type(message) is str
    if not condition:
        raise AssertionVMError(message)
    return ()


def _f2i(args):
    (f,) = args
    assert type(f) is float
    try:
        i = int(f)
    except ValueError:
        raise AssertionVMError('cannot convert the float value to type int')
    return (i,)


def _f2s(args):
    (f,) = args
    assert type(f) is float
    try:
        s = repr(f)
    except ValueError:
        raise AssertionVMError('cannot convert the float value to type str')
    return (s,)


def _i2f(args):
    (i,) = args
    assert type(i) is int
    try:
        f = float(i)
    except ValueError:
        raise AssertionVMError('cannot convert the int value to type float')
    return (f,)


def _i2s(args):
    (i,) = args
    assert type(i) is int
    try:
        s = str(i)
    except ValueError:
        raise AssertionVMError('cannot convert the int value to type str')
    return (s,)


def _s2f(args):
    (s,) = args
    assert type(s) is str
    try:
        f = float(s)
    except ValueError:
        raise AssertionVMError('cannot convert the str value to type float')
    return (f,)


def _s2i(args):
    (s,) = args
    assert type(s) is str
    m = re.match(r'-?\d+\Z', s)
    if m:
        i = int(m.group())
    else:
        raise AssertionVMError('cannot convert the str value to type int')
    return (i,)


def _print(args):
    '''Prints out a string.'''
    (line,) = args
    assert type(line) is str
    sys.stdout.write(line)
    return ()


def _on_output(args):
    '''
    Pretty prints a value and its type. `args` contains:
     - is_static - bool - ignored.
     - value - dynamic - the value to print.
    '''
    is_static, value = pack.unpack(Ref.new(args), 'w1r5')
    assert type(is_static) is bool
    value = Dynamic.from_welly(value)
    print(pretty(lambda: ('', ': ', '', [
        pretty_value.wrap(value),
        pretty_type.wrap(value.type_),
    ])))
    return ()


# Destination for the syscall 'log_output': a Python list of strings.
TEST_LOG = []


# Format used by the syscall 'log_output'.
def log_format(is_static, constness, type_, words):
    return "%s; %s; %s; %r" % (
        is_static,
        const.switch(constness),
        pretty(pretty_type.wrap(type_), height=1, max_depth=5),
        words
    )


def _log_output(args):
    '''
    Appends information about a value to `TEST_LOG`. `args` contains:
     - is_static - bool - ignored.
     - value - dynamic - the value to log.
    '''
    is_static, value = pack.unpack(Ref.new(args), 'w1r5')
    assert type(is_static) is bool
    value = Dynamic.from_welly(value)
    TEST_LOG.append(log_format(
        is_static,
        value.constness,
        value.type_,
        value.ref.read(),
    ))
    return ()


all_syscalls = [
    Syscall('assert', struct([bool_, str_]), void, _assert),
    Syscall('f2i', struct([float_]), int_, _f2i),
    Syscall('f2s', struct([float_]), str_, _f2s),
    Syscall('i2f', struct([int_]), float_, _i2f),
    Syscall('i2s', struct([int_]), str_, _i2s),
    Syscall('s2i', struct([str_]), int_, _s2i),
    Syscall('s2f', struct([str_]), float_, _s2f),
    Syscall('print', struct([str_]), void, _print),
    Syscall('on_output', struct([bool_, dynamic]), void, _on_output),
    Syscall('log_output', struct([bool_, dynamic]), void, _log_output),
]


index = dict((x.name, x) for x in all_syscalls)
