from .operators import Operator
from .syscalls import Syscall
from ..model import defer, const, wtypes, subtypes, ops

from ..model.memory import Array
from ..model.pack import pack, tag
from ..model.const import BASE, CONST, VAR
from ..model.wtypes import sizeof
from ..model.subtypes import is_subtype
from ..model.errors import *


'''
Data structures and utilities for generating type-correct low-level code. This
representation is similar to flat code. It uses a stack and a growing list of
locals, the same nested control-flow constructs, and mostly the same opcodes.
The differences allow it to be type-checked without executing any code. The
differences are as follows:

 1a. Every computation that produces a static result is replaced by a 'static'
     instruction that includes the type and constness of the result and a
     pointer to it.

 1b. Instructions obviated by 1a are removed. They include 'const_int',
     'const_str', 'const_char', and perhaps some more.

 2. Instructions that *require* static const operands include them as part
    of the instruction. For example, 'defer' includes its type. Such
    operands are never placed on the stack.

 3. Ad-hoc polymorphic instructions are type-specialised. For example,
    `('binary', '*')` is replaced by e.g. `('binary', '*ii')`,
    `('binary', '*ff')`, `(binary, '*si')` depending on the type of its
    operands.

For further motivation see "doc/compiler/compiler-levels.txt" and
"doc/compiler/il-levels.txt".

Like flat code, typed code is constructed incrementally by appending
instructions to Blocks, and completed Blocks can be nested inside control flow
instructions. Also like flat code, instructions notionally pass values to one
another by reference.

Pointers to temporary values are held on a stack, which is private to a Block.
Pointers to local variables are created by 'init' and retrieved by 'local'.
Pointers to captured variables are retrieved by 'capture'. Locals and captures
are numbered from zero.

Unlike flat code, on entry to a function, the arguments are passed as a
struct, which is placed on the stack.

Pointers to external variables are retrieved using 'static', i.e. they are no
different from other static values. 'capture' is forbidden at the outermost
level.
'''

# TODO: The representation of typed code is currently a Python data structure
# (a list of tuples), not a Welly data structure (a list of union values).
# Convert it, inlining types by value instead of by reference.


class TypedCodeError(Exception):
    pass


def sizesof(types):
    '''
    Returns an Array containing the sizes of a sequence of types.
    '''
    return Array.from_list(
        [(sizeof(t),) for t in types],
        sizeof(wtypes.int_),
    )


class Slot(object):
    '''
    Records the constness and type of a stack item.

     - constness - a Ref to a constness.
     - type_ - a Ref to a Welly type.
    '''
    def __init__(self, constness, type_):
        const.check_valid(constness)
        wtypes.check_valid(type_)
        self.constness = constness
        self.type_ = type_


class Block(object):
    '''
    Accumulates a sequence of instructions corresponding to a high-level
    lexical block.

    Public fields:

     - instructions - A list of typed code instructions.
     - is_end_reachable - Is there a path to the end of the Block?
     - is_break_reachable - Is there a path to a 'break' instruction?
     - is_continue_reachable - Is there a path to a 'continue' instruction?
     - stack_slots - A list of Slots describing the stack items at the end
       of the block.
     - locals_ - a list of Slots describing the local variables at the end
       of the block. Function parameters, if any, form the first few local
       variables.
     - captures - a list of Slots describing the captured variables.
     - _return_types - a Python list of Refs to the types of expressions
       appearing in 'return' instructions in this Block (recursively).

    Instructions are tuples, with the first element being the op-code (a str)
    and subsequent elements being opcode-dependent. Non-static operands are
    always popped from the stack, and results are pushed to the stack, so
    these do not appear in the tuple.

    All the public methods of this class append instructions. Instructions
    should only be appended using the methods, as they do bookkeeping and
    type-checking.
    '''

# There are various places where types are included in typed code. For now, we
# use Refs to types. In the longer term, we should inline them, and/or
# represent typed code instructions as union values.

    def __init__(self, stack_slots=None, locals_=None, captures=None):
        self.instructions = []
        self.is_end_reachable = True
        self.is_break_reachable = False
        self.is_continue_reachable = False
        self.stack_slots = [] if stack_slots is None else stack_slots
        self.locals_ = [] if locals_ is None else locals_
        self.captures = [] if captures is None else captures
        self._return_types = []

    def _append(self, *instruction):
        '''
        Appends `instruction` to this Block. It is an error to append an
        instruction unless `is_end_reachable`. The caller must also update
        `stack_slots` using `_pop()` and `_push()`.
        '''
        if not self.is_end_reachable:
            raise TypedCodeError('Code is not reachable')
        self.instructions.append(instruction)

    def _pop_many(self, n):
        '''
        Removes the topmost `n` stack items from `stack_slots`. Returns a list
        of Slots.
        '''
        new_len = len(self.stack_slots) - n
        if new_len < 0:
            raise TypedCodeError('Stack underflow')
        ans = self.stack_slots[new_len:]
        self.stack_slots[new_len:] = []
        return ans

    def _pop(self):
        '''
        Removes the topmost stack item from `stack_slots`. Returns a Slot.
        '''
        return self._pop_many(1)[0]

    def _push(self, slot):
        '''
        Appends a stack item, given its constness and type in the form of
        a Slot.
        '''
        self.stack_slots.append(slot)

    def _check_empty(self):
        '''Raises an exception if the stack is not empty.'''
        if self.stack_slots:
            raise TypedCodeError(
                'Stack is not empty after %r' % self.instructions)

    def _update_exit_analysis(self, n, exit_blocks, loc):
        '''
        After appending a control flow instruction, call this to update the
        following fields:
         - stack_slots
         - is_end_reachable
         - is_break_reachable
         - is_continue_reachable
         - return_types

        Params:
         - n - the number of stack items left by each of `exit_blocks`.
         - exit_blocks - a list of the Blocks that might be invoked by the
           control flow instruction. Control flow must pass through exactly
           one of these Blocks.
         - loc - An opaque source code location for use in error messages.
        '''
        if not all(len(b.stack_slots) == n for b in exit_blocks):
            raise TypedCodeError('Conditional branches are unbalanced')
        # Compute stack types and constnesses.
        if len(exit_blocks) == 0:
            # A control flow construct with no branches is possible.
            # E.g. fn(x: never) { switch x {} }
            # The stack effect is undefined, so we make a special case.
            if n != 0:
                raise TypedCodeError('Zero branches cannot create values')
        else:
            # For each stack slot, compute the join over the branches.
            for slots in zip(*[b.stack_slots for b in exit_blocks]):
                constness, type_ = subtypes.join_many_with_constnesses(
                    [slot.constness for slot in slots],
                    [slot.type_ for slot in slots],
                    'the conditional expression type',
                    loc,
                )
                self._push(Slot(constness, type_))
        # Update flags.
        if any(b.is_break_reachable for b in exit_blocks):
            self.is_break_reachable = True
        if any(b.is_continue_reachable for b in exit_blocks):
            self.is_continue_reachable = True
        if not any(b.is_end_reachable for b in exit_blocks):
            self.is_end_reachable = False
        # Collect return types.
        for b in exit_blocks:
            self._return_types.extend(b._return_types)

    def get_return_type(self, loc):
        '''
        Computes the join of the types of all values returned by this Block
        and its nested Blocks. If there are no such values, returns
        `wtypes.never`.

         - loc - An opaque source code location for use in error messages.
        '''
        return subtypes.join_many(self._return_types, 'the return type', loc)

    def get_code(self, is_break_allowed):
        '''
        Returns a fresh copy of self.instructions.

         - is_break_allowed - `True` if "break" and "continue" are allowed.
           If "break" is not allowed, but is reachable in `self`, then this
           method raises a TypedCodeError.
        '''
        if self.is_break_reachable and not is_break_allowed:
            raise TypedCodeError('Illegal "break" instruction')
        if self.is_continue_reachable and not is_break_allowed:
            raise TypedCodeError('Illegal "continue" instruction')
        return list(self.instructions)

    # Instruction set begins here.

    def address_of(self):
        slot = self._pop()
        t = wtypes.ref(slot.constness, slot.type_)
        self._push(Slot(CONST, t))
        self._append('address_of')

    def array(self, elem_type, num_items, loc):
        '''..., item1, ... -> ..., array_value'''
        if num_items == 0:
            raise TypedCodeError("Use 'static' to construct the empty array")
        for slot in self._pop_many(num_items):
            subtypes.check(slot.type_, elem_type, None, loc)
        t = wtypes.array(CONST, elem_type)
        self._push(Slot(CONST, t))
        self._append('array', sizeof(elem_type), num_items)

    def array_concat(self, loc):
        '''
        ..., left_array, right_array -> ..., array

        The result has the same constness and element type as `left_array`,
        which must not be of type `array0`. `right_array` can be of type
        `array0`. Otherwise, it can have any constness and its element type
        must be a subtype of the result element type.
        '''
        y = self._pop()
        x = self._pop()
        # Unpack `x.type_`.
        x_tag, x_payload = wtypes.switch(x.type_)
        if x_tag == 'ARRAY0':
            raise TypedCodeError('left operand is an array0')
        if x_tag != 'ARRAY':
            raise TypedCodeError('left operand is not an array')
        x_constness, x_elem_t = x_payload
        elem_size = sizeof(x_elem_t)
        # Unpack `y.type_`.
        y_tag, y_payload = wtypes.switch(y.type_)
        if y_tag == 'ARRAY0':
            pass
        elif y_tag == 'ARRAY':
            _, y_elem_t = y_payload
            subtypes.check(y_elem_t, x_elem_t, 'the element type', loc)
        else:
            raise TypedCodeError('right operand is not an array')
        # Generate the instruction.
        self._push(Slot(CONST, x.type_))
        self._append('array_concat', elem_size)

    def array_length(self):
        '''..., array -> ..., length'''
        array_slot = self._pop()
        t0, payload = wtypes.switch(array_slot.type_)
        if t0 != 'ARRAY':
            raise TypedCodeError('value is not an array')
        constness, elem_type = payload
        elem_size = sizeof(elem_type)
        self._push(Slot(CONST, wtypes.int_))
        self._append('array_length', elem_size)

    def array_repeat(self, loc):
        '''
        ..., array, count -> ..., array

        The result has the same constness and element type as `array`,
        which must not be of type `array0`. `count` is an integer.
        '''
        y = self._pop()
        x = self._pop()
        # Unpack `x.type_`.
        x_tag, x_payload = wtypes.switch(x.type_)
        if x_tag == 'ARRAY0':
            raise TypedCodeError('left operand is an array0')
        if x_tag != 'ARRAY':
            raise TypedCodeError('left operand is not an array')
        x_constness, x_elem_t = x_payload
        elem_size = sizeof(x_elem_t)
        # Unpack `y`.
        subtypes.check(y.type_, wtypes.int_, 'the repeat count', loc)
        # Generate the instruction.
        self._push(Slot(CONST, x.type_))
        self._append('array_repeat', elem_size)

    def array_slice(self, loc):
        '''
        ..., array, start, end -> ..., subarray

        The result has the same constness and element type as `array` and
        shares storage with it.
        '''
        y = self._pop()
        x = self._pop()
        a = self._pop()
        # Unpack `a`.
        a_tag, a_payload = wtypes.switch(a.type_)
        if a_tag == 'ARRAY0':
            raise TypedCodeError('operand is an array0')
        if a_tag != 'ARRAY':
            raise TypedCodeError('operand is not an array')
        constness, elem_t = a_payload
        # Unpack `x` and `y`.
        subtypes.check(x.type_, wtypes.int_, 'the start index', loc)
        subtypes.check(y.type_, wtypes.int_, 'the end index', loc)
        # Generate the instruction.
        self._push(Slot(CONST, a.type_))
        self._append('array_slice', sizeof(elem_t))

    def array0_repeat(self, loc):
        '''
        ..., array0, count -> ..., array0

        Emits the same instruction as `array_repeat()`, but applies different
        type checking. `array0` is an `array0`. `count` is an integer.
        '''
        y = self._pop()
        x = self._pop()
        # Unpack `x.type_`.
        x_tag, _ = wtypes.switch(x.type_)
        if x_tag != 'ARRAY0':
            raise TypedCodeError('left operand is not an array0')
        # Unpack `y.type_`.
        subtypes.check(y.type_, wtypes.int_, 'the repeat count', loc)
        # Generate the instruction.
        self._push(Slot(CONST, x.type_))
        self._append('array_repeat', 0)

    def assign(self, loc):
        '''..., rvalue, lvalue -> ...'''
        l_slot = self._pop()
        r_slot = self._pop()
        if const.switch(l_slot.constness) != 'VAR':
            raise ConstnessVMError(loc=loc)
        subtypes.check(r_slot.type_, l_slot.type_, None, loc)
        self._append('assign', sizeof(l_slot.type_))

    def binary(self, op, loc):
        '''
        ..., value, value -> ..., value
        '''
        assert isinstance(op, Operator)
        x_t, y_t = op.in_types
        y = self._pop()
        x = self._pop()
        subtypes.check(x.type_, x_t, None, loc)
        subtypes.check(y.type_, y_t, None, loc)
        self._push(Slot(CONST, op.result_type))
        self._append('binary', op.opcode)

    def box(self):
        '''..., value -> ..., dynamic_value'''
        slot = self._pop()
        self._push(Slot(CONST, wtypes.dynamic))
        self._append('box', slot.constness, slot.type_)

    def break_(self):
        '''Stack must be empty.'''
        self._check_empty()
        self.is_break_reachable = self.is_end_reachable
        self._append('break')
        self.is_end_reachable = False

    def bury(self, n):
        '''
        ..., <n items>, value -> ..., value, <n items>
        bury(0) is a no-op.
        bury(1) is also called 'swap' and 'zy'.
        bury(2) is also called 'zxy'.
        '''
        if n < 0:
            raise TypedCodeError('Negative index for "bury"')
        x = self._pop()
        ys = self._pop_many(n)
        self._push(x)
        for y in ys:
            self._push(y)
        if n == 0:
            # `bury(0)` is a no-op.
            return
        self._append('bury', n)

    def call_function(self, n, loc):
        '''..., func, <n args> -> ..., result'''
        arg_slots = self._pop_many(n)
        func_slot = self._pop()
        t0, payload = wtypes.switch(func_slot.type_)
        if t0 != 'FUNCTION':
            raise TypedCodeError('Called value is not a function')
        param_types, return_type = payload
        if len(param_types) != n:
            raise NumberOfArgumentsVMError(len(param_types), n, loc)
        sizes_ref, sizes_len = sizesof(param_types).to_welly()
        for arg_slot, param_type in zip(arg_slots, param_types):
            subtypes.check(arg_slot.type_, param_type, None, loc)
        self._push(Slot(CONST, return_type))
        self._append('call_function', sizes_ref, sizes_len)

    def capture(self, index):
        '''... -> ..., value'''
        if index < 0 or index >= len(self.captures):
            raise TypedCodeError("'capture' index out of range")
        slot = self.captures[index]
        self._push(slot)
        self._append('capture', index)

    def cast(self, new_type, loc):
        '''..., value -> ..., value'''
        slot = self._pop()
        subtypes.check(slot.type_, new_type, None, loc)
        self._append('cast', sizeof(new_type))
        self._push(Slot(CONST, new_type))

    def closure(self, make_body, param_types, n, loc):
        '''
        ..., <n captures> -> ..., function

         - make_body - a callback that generates the body of the function.
           The callback is given a Block and populates it by calling its
           methods. On entry to the Block:
            - The stack contains the argument struct (a single value).
            - There are no locals.
            - The captures are the `n` values popped by this instruction.
           The Block must not fall through. It can exit by executing a
           'return' instruction.
         - param_types - an Array of Welly types.
         - n - the number of stack slots holding captured values.
         - loc - an opaque source code location for use in error messages.
        '''
        # Check arguments.
        assert type(param_types) is Array
        param_types = param_types.to_list()
        for t in param_types:
            wtypes.check_valid(t)
        # Pop the captures.
        capture_slots = self._pop_many(n)
        # Compile the body.
        arg_type = wtypes.struct(param_types)
        block = Block(
            stack_slots=[Slot(CONST, arg_type)],
            locals_=[],
            captures=capture_slots,
        )
        make_body(block)
        if block.is_end_reachable:
            raise TypedCodeError('Missing "return" instruction')
        # Push the result.
        return_type = block.get_return_type(loc)
        if return_type is None:
            return_type = wtypes.never
        func_type = wtypes.function(param_types, return_type)
        self._push(Slot(CONST, func_type))
        # Generate the instruction.
        code = block.get_code(is_break_allowed=False)
        self._append('closure', code, n)

    def continue_(self):
        '''Stack must be empty.'''
        self._check_empty()
        self.is_continue_reachable = self.is_end_reachable
        self._append('continue')
        self.is_end_reachable = False

    def defer(self, type_, loc):
        '''... -> ..., value'''
        if not wtypes.is_deferrable(type_):
            raise DeferTypeVMError(type_, loc)
        self._push(Slot(CONST, type_))
        self._append('defer')

    def dereference(self):
        '''..., ref -> ..., value'''
        slot = self._pop()
        t0, payload = wtypes.switch(slot.type_)
        if t0 != 'REF':
            raise TypedCodeError('Value is not a reference')
        constness, type_ = payload
        self._push(Slot(constness, type_))
        self._append('dereference', sizeof(type_))

    def drop(self):
        '''..., z -> ...'''
        _ = self._pop()
        self._append('drop')

    def dup(self):
        '''
        ..., z -> ..., z, z
        This is actually implemented as: pick(0).
        '''
        self.pick(0)

    def dynamic_constness(self):
        '''
        ..., dynamic -> ..., constness
        '''
        slot = self._pop()
        t0, _ = wtypes.switch(slot.type_)
        if t0 != 'DYNAMIC':
            raise TypedCodeError('Value is not a dynamic')
        self._push(Slot(CONST, wtypes.constness))
        self._append('dynamic_constness')

    def dynamic_type(self):
        '''
        ..., dynamic -> ..., type
        '''
        slot = self._pop()
        t0, _ = wtypes.switch(slot.type_)
        if t0 != 'DYNAMIC':
            raise TypedCodeError('Value is not a dynamic')
        self._push(Slot(CONST, wtypes.type_))
        self._append('dynamic_type')

    def fulfil(self, loc):
        '''..., rvalue, lvalue -> ...'''
        l_slot = self._pop()
        r_slot = self._pop()
        if not wtypes.is_deferrable(l_slot.type_):
            raise DeferTypeVMError(l_slot.type_, loc)
        subtypes.check(r_slot.type_, l_slot.type_, None, loc)
        self._append('fulfil')

    def if_(self, n, make_then, make_else, loc):
        '''
        ..., condition -> ..., <n items>

         - make_then - A callback that takes a Block and generates the "then"
           branch. It can return any value.
         - make_else - A callback that takes a Block and generates the "else"
           branch. It can return any value.

        The `n` items left on the stack by `make_then()` or `make_else()` are
        appended to this Block's stack.

        This method returns a pair `(then_result, else_result)` where
        `then_result` is the return value of `make_then` and `else_result` is
        the return value of `make_else`.
        '''
        subtypes.check(
            self._pop().type_,
            wtypes.bool_,
            'the "if" condition',
            loc,
        )
        then_block = Block(locals_=list(self.locals_), captures=self.captures)
        then_result = make_then(then_block)
        else_block = Block(locals_=list(self.locals_), captures=self.captures)
        else_result = make_else(else_block)
        then_code = then_block.get_code(is_break_allowed=True)
        else_code = else_block.get_code(is_break_allowed=True)
        self._append('if', n, then_code, else_code)
        self._update_exit_analysis(n, [then_block, else_block], loc)
        return (then_result, else_result)

    def init(self, is_var):
        '''..., value -> ...'''
        constness = VAR if is_var else CONST
        slot = self._pop()
        self.locals_.append(Slot(constness, slot.type_))
        self._append('init', sizeof(slot.type_))

    def inline(self, make):
        '''
        ... -> ..., ?

         - make - A callback that takes a Block and populates it with
           instructions. It can return any value.

        Returns the return value of `make`.

        This instruction can be used e.g. to implement "if" when the condition
        is a static constant.

        The Block constructed by `make` has access to all the captures and
        locals but not to the stack of this Block. If the Block declares new
        locals, they are private, and this Block will not see them. Values
        left on the stack by the Block are appended to this Block's stack. The
        Block may 'break', 'continue', 'return' and 'throw' if this Block may.
        '''
        block = Block(locals_=list(self.locals_), captures=self.captures)
        result = make(block)
        self._append('inline', block.get_code(is_break_allowed=True))
        self._update_exit_analysis(len(block.stack_slots), [block], None)
        return result

    def local(self, index):
        '''... -> ..., value'''
        if index < 0 or index >= len(self.locals_):
            raise TypedCodeError("'local' index out of range")
        slot = self.locals_[index]
        self._push(slot)
        self._append('local', index)

    def loop(self, make_body):
        '''
        Stack must be empty.

         - make_body - A callback that takes a Block and generates the body
           of the loop. It can return any value.

        Returns the value that `make_body` returns.
        '''
        self._check_empty()
        block = Block(locals_=list(self.locals_), captures=self.captures)
        result = make_body(block)
        block._check_empty()
        self._append('loop', block.get_code(is_break_allowed=True))
        if not block.is_break_reachable:
            self.is_end_reachable = False
        # N.B. 'continue' does not break out of the 'loop', so does not make
        # a path to the end of the Block containing the 'loop' instruction.
        self._return_types.extend(block._return_types)
        return result

    def new(self):
        '''..., value -> ..., ptr'''
        slot = self._pop()
        t = wtypes.ref(VAR, slot.type_)
        self._push(Slot(CONST, t))
        self._append('new', sizeof(slot.type_))

    def pick(self, n):
        '''
        x, <n items> -> x, <n items>, x
        pick(0) does the same as 'zz', also called 'dup'.
        pick(1) does the same as 'yzy'.
        pick(2) does the same as 'xyzx'.
        '''
        if n < 0:
            raise TypedCodeError('Negative index for "pick"')
        index = len(self.stack_slots) - 1 - n
        if index < 0:
            raise TypedCodeError("Stack underflow")
        slot = self.stack_slots[index]
        self._push(slot)
        self._append('pick', n)

    def return_(self):
        '''
        value ->

        Stack must become empty.
        '''
        slot = self._pop()
        self._append('return', sizeof(slot.type_))
        self._check_empty()
        self._return_types.append(slot.type_)
        self.is_end_reachable = False

    def reverse(self, n):
        '''..., item_1, ..., item_n -> ..., item_n, ..., item_1'''
        slots = self._pop_many(n)
        for x in reversed(slots):
            self._push(x)
        self._append('reverse', n)

    def static(self, value, n):
        '''
        ..., <n items> -> ..., value, <n items>

         - value - a Dynamic representing the value to push onto the stack.
           You can use this instruction to push values of any type and
           constness, including L-values.
         - n - an int representing the depth in the stack at which to
           insert the value.
        '''
        items_above = self._pop_many(n)
        self._push(Slot(value.constness, value.type_))
        for x in items_above:
            self._push(x)
        self._append('static', value.ref, n)

    def str_length(self, loc):
        '''..., str -> ..., length'''
        str_slot = self._pop()
        subtypes.check(str_slot.type_, wtypes.str_, None, loc)
        self._push(Slot(CONST, wtypes.int_))
        self._append('str_length')

    def struct(self, n):
        '''
        ..., <n items> -> ..., struct_value
        '''
        types = [slot.type_ for slot in self._pop_many(n)]
        struct_type = wtypes.struct(types)
        sizes_ref, sizes_len = sizesof(types).to_welly()
        assert sizes_len == n
        self._push(Slot(CONST, struct_type))
        self._append('struct', sizes_ref, sizes_len)

    def subscript_array(self, loc):
        '''
        ..., array, index -> ..., value
        '''
        index_slot = self._pop()
        array_slot = self._pop()
        subtypes.check(index_slot.type_, wtypes.int_, "the index", loc)
        t0, payload = wtypes.switch(array_slot.type_)
        if t0 != 'ARRAY':
            raise TypedCodeError('Value is not an array')
        constness, elem_type = payload
        self._push(Slot(constness, elem_type))
        self._append('subscript_array', sizeof(elem_type))

    def subscript_struct(self, index, loc):
        '''
        ..., struct -> ..., value

         - index - int. Unlike e.g. `subscript_array()`, the index must be
           a compile-time-known constant.
         - loc - a location for reporting error messages.
        '''
        assert type(index) is int
        slot = self._pop()
        if wtypes.switch(slot.type_)[0] != 'STRUCT':
            raise TypedCodeError('Value is not a struct')
        for i, (t, offset) in enumerate(ops.struct_offsets(slot.type_)):
            if i == index:
                break
        else:
            raise StructIndexVMError(slot.type_, index, loc)
        self._push(Slot(slot.constness, t))
        self._append('subscript_struct', offset, sizeof(t))

    def swap(self):
        '''
        ..., y, z -> ..., z, y
        This is actually implemented as: bury(1).
        '''
        self.bury(1)

    def switch(self, make_cases, make_else, loc):
        '''
        ..., condition -> ...

        The 'switch' instruction examines a union value and executes one
        of several "case" branches, depending on its tag. If no case branch
        is provided that matches the tag then the "else" branch is executed
        (if present).

         - make_cases - A list of (tag, callback) pairs, one for each case.
           Each callback takes a Block and generates the code for that case.
           It can return any value.
         - make_else - Either `None` or a callback that takes a Block and
           generates the "else" branch. It can return any value.

        The code generated by each callback will be entered with the fields of
        the payload bound to fresh local variables. The code must not leave
        any items on the stack.

        This method returns a pair `(case_results, else_result)` where
        `case_results` is a list of the return values of the "case" callbacks
        and `else_result` is the return value of the "else" callback, or
        `None`.
        '''
        slot = self._pop()
        if wtypes.switch(slot.type_)[0] != 'UNION':
            raise TypedCodeError('Value is not a union')
        unhandled_cases = ops.union_cases(slot.type_)
        # Compile all the "case" Blocks.
        case_blocks = []
        case_results = []
        case_triples = [] # (tag, Welly array of sizes, typed code).
        for tag, make_case in make_cases:
            # Declare fresh locals for payload field.
            locals2 = list(self.locals_)
            try:
                types = unhandled_cases.pop(tag)
            except KeyError:
                raise TypedCodeError("Unreachable case '%s'" % tag)
            locals2.extend(Slot(CONST, t) for t in types)
            # Compile the block.
            case_block = Block(locals_=locals2, captures=self.captures)
            case_blocks.append(case_block)
            case_results.append(make_case(case_block))
            case_code = case_block.get_code(is_break_allowed=True)
            case_triples.append(pack(tag, sizesof(types), case_code))
        # Compile the "else" Block.
        if make_else is None:
            for tag in unhandled_cases:
                raise TypedCodeError("Case '%s' is not handled" % (tag,))
            else_block = None
            else_result = None
            else_code = None
        else:
            if not unhandled_cases:
                raise TypedCodeError('Unreachable "else" block')
            locals2 = list(self.locals_)
            else_block = Block(locals_=locals2, captures=self.captures)
            else_result = make_else(else_block)
            else_code = else_block.get_code(is_break_allowed=True)
        # Make the 'switch' instruction.
        self._append('switch', case_triples, else_code)
        # Update reachability and other type information.
        exit_blocks = list(case_blocks)
        if else_block is not None:
            exit_blocks.append(else_block)
        self._update_exit_analysis(0, exit_blocks, loc)
        # Return the results of the callbacks.
        return (case_results, else_result)

    def syscall(self, syscall, loc):
        '''
        ..., arg -> ..., result

         - syscall - a syscalls.Syscall.
        '''
        assert isinstance(syscall, Syscall)
        x = self._pop()
        subtypes.check(x.type_, syscall.in_type, None, loc)
        self._push(Slot(CONST, syscall.out_type))
        self._append('syscall', syscall.name)

    def unary(self, op, loc):
        '''
        ..., value -> ..., value

         - op - a welly.compiler.Operator with one operand.
        '''
        assert isinstance(op, Operator)
        (x_t,) = op.in_types
        x = self._pop()
        subtypes.check(x.type_, x_t, None, loc)
        self._push(Slot(CONST, op.result_type))
        self._append('unary', op.opcode)

    def union(self, tag, n):
        '''
        ..., <n items> -> ..., union_value

         - tag - a str.
        '''
        types = [slot.type_ for slot in self._pop_many(n)]
        union_type = wtypes.union([(tag, types)])
        sizes_ref, sizes_len = sizesof(types).to_welly()
        self._push(Slot(CONST, union_type))
        self._append('union', tag, sizes_ref, sizes_len)

    def unpack_struct(self, n, loc):
        '''
        ..., struct_value -> ..., field_1, ..., field_n

        Checks that `struct_value` is a struct with `n` fields. Leaves the
        field values on the stack.
        '''
        slot = self._pop()
        if wtypes.switch(slot.type_)[0] != 'STRUCT':
            raise TypedCodeError('Value is not a struct')
        if ops.struct_length(slot.type_) != n:
            raise NumberOfFieldsVMError(
                n, ops.struct_length(slot.type_), loc=loc)
        field_types = ops.struct_field_types(slot.type_)
        sizes_ref, sizes_len = sizesof(field_types).to_welly()
        for t in field_types:
            self._push(Slot(slot.constness, t))
        self._append('unpack_struct', sizes_ref, sizes_len)

    # Permutations of the top of the stack.
    # For each stack depth in {0, 1, 2}, we provide:
    #  - an instruction for copying a value up from that depth.
    #  - an instruction for copying a value down to that depth.
    #  - an instruction for moving a value down to that depth.
    # Except:
    #  - for depth 0, moving is a no-op, so not provided.
    #  - for depth 0, copying up and copying down are equivalent, and
    #    called 'dup'.
    #  - for depth 1, moving down is called 'swap'.
    # For all depths:
    #  - copying up is implemented using 'pick'.
    #  - moving down is implemented using 'bury'.
    #  - copying down is implemented is 'bury' followed by 'pick'.
    # 'pick' and 'bury' can be used directly to access arbitrary depths in
    # the stack.

    def xyzx(self):
        '''
        ..., x, y, z -> ..., x, y, z, x
        This is actually implemented as: pick(2).
        '''
        self.pick(2)

    def yzy(self):
        '''
        ..., y, z -> ..., y, z, y
        This is actually implemented as: pick(1).
        '''
        self.pick(1)

    def zxy(self):
        '''
        ..., x, y, z -> ..., z, x, y
        This is actually implemented as: bury(2).
        '''
        self.bury(2)

    def zxyz(self):
        '''
        ..., x, y, z -> ..., z, x, y, z
        This is actually implemented as: bury(2) pick(2).
        '''
        self.bury(2)
        self.pick(2)

    zy = swap

    def zyz(self):
        '''
        ..., y, z -> ..., z, y, z
        This is actually implemented as: bury(1) pick(1).
        '''
        self.bury(1)
        self.pick(1)

    zz = dup
