def _str_type(type_):
    '''
    Given a ref to a Welly type, return a str representing it. This is useful
    in the `__str__()` methods of type-related errors.

    This method does a poor job. It's just about good enough for testing. If
    you care about how an error is printed, don't use its `__str__()` method.
    Generate your own error message instead, e.g. using `pretty_type`. We
    cannot use `pretty_type` here because that would be a circular dependency.

     - type_ - a Ref to a Welly type.
    '''
    return type_.read()[0]


class VMError(Exception):
    '''
    Superclass of all errors detected by the VM.
    '''
    def __init__(self, loc=None):
        '''
         - loc - Optionally, an opaque object representing the source code
           location of the error.
        '''
        super(VMError, self).__init__()
        assert type(loc) is not str
        self.loc = loc


class StaticVMError(VMError):
    '''
    Superclass of all errors which are reported at compile time.
    '''
    def __init__(self, _='omit this', loc=None):
        assert _ == 'omit this' # Force caller to use keywords.
        super(StaticVMError, self).__init__(loc=loc)


class StaticMessageVMError(StaticVMError):
    '''
    Superclass of StaticVMErrors with a constant message.
    '''
    def __init__(self, message, loc=None):
        assert type(message) is str
        super(StaticMessageVMError, self).__init__(loc=loc)
        self.message = message

    def __str__(self):
        return self.message


class UndefinedNameVMError(StaticVMError):
    def __init__(self, name, loc):
        assert type(name) is str
        super(UndefinedNameVMError, self).__init__(loc=loc)
        self.name = name

    def __str__(self):
        return (
            "name '%s' has not been defined; you can declare it using "
            "\"const\" or \"var\""
        ) % (self.name,)


class UnreachableCodeVMError(StaticMessageVMError):
    '''
    Raised when the user writes code that control flow cannot reach under
    an abstract interpretation.
    '''
    def __init__(self, loc):
        super(UnreachableCodeVMError, self).__init__(
            'Unreachable code', loc=loc)


class ConstnessVMError(StaticMessageVMError):
    '''
    Raised when a value is assigned to a variable which is not VAR.
    '''
    def __init__(self, loc):
        super(ConstnessVMError, self).__init__(
            'can only assign to "var" locations', loc=loc)


class TypeVMError(StaticVMError):
    '''
    Raised when a value is assigned to a variable which is the wrong type to
    hold it.
    '''
    def __init__(self, rvalue_type, lvalue_type, noun, loc):
        '''
         - rvalue_type - a Ref to the type of the value.
         - lvalue_type - a Ref to the type of the variable.
         - noun - optionally, a noun phrase that describes the value, e.g.
           "return type of function". `None` indicates that a noun phrase
           should be constructed based on `rvalue_type`.
        '''
        assert noun is None or type(noun) is str
        super(TypeVMError, self).__init__(loc=loc)
        self.rvalue_type = rvalue_type
        self.lvalue_type = lvalue_type
        self.noun = noun

    def __str__(self):
        rvalue_type = _str_type(self.rvalue_type)
        lvalue_type = _str_type(self.lvalue_type)
        if self.noun is None:
            return "value of type %s is not of type %s" % (
                rvalue_type, lvalue_type)
        else:
            return "%s is not of type %s" % (
                self.noun, lvalue_type)


class TypeInferenceVMError(StaticVMError):
    '''
    Raised when a type inference calculation has no answer.
    '''
    def __init__(self, type1, type2, problem_type, noun, loc):
        '''
         - problem_type - 'supertype' if the problem is that `type1` and
           `type2` have no common supertype, or 'subtype' if the problem is
           that they have no common subtype.
         - noun - a noun phrase describing the expression whose type is being
           inferred.
        '''
        assert problem_type in ('supertype', 'subtype')
        assert type(noun) is str
        super(TypeInferenceVMError, self).__init__(loc=loc)
        self.type1 = type1
        self.type2 = type2
        self.problem_type = problem_type
        self.noun = noun

    def __str__(self):
        return "cannot compute %s; %s and %s have no common %s" % (
            self.noun,
            _str_type(self.type1),
            _str_type(self.type2),
            self.problem_type,
        )


class CallTypeVMError(StaticVMError):
    '''
    Raised when an attempt is made to call a value whose type is not callable.
    '''
    def __init__(self, type_, loc):
        super(CallTypeVMError, self).__init__(loc=loc)
        self.type_ = type_

    def __str__(self):
        return "cannot call values of type %s" % (_str_type(self.type_),)


class NumberOfArgumentsVMError(StaticVMError):
    '''
    Raised when a function is called passing the wrong number of arguments.
    '''
    def __init__(self, num_params, num_args, loc):
        super(NumberOfArgumentsVMError, self).__init__(loc=loc)
        self.num_params = num_params
        self.num_args = num_args

    def __str__(self):
        return "wrong number of arguments - expected %d but got %d" % (
            self.num_params, self.num_args)


class NumberOfFieldsVMError(StaticVMError):
    '''
    Raised when unpacking a struct with the wrong number of fields.
    '''
    def __init__(self, num_lvalues, num_rvalues, loc):
        super(NumberOfFieldsVMError, self).__init__(loc=loc)
        self.num_lvalues = num_lvalues
        self.num_rvalues = num_rvalues

    def __str__(self):
        return "wrong number of fields - expected %d but got %d" % (
            self.num_lvalues, self.num_rvalues)


class DeferTypeVMError(StaticVMError):
    '''
    Raised when an attempt is made to defer or fulfil a value whose type has
    no deferred representation.
    '''
    def __init__(self, type_, loc):
        super(DeferTypeVMError, self).__init__(loc=loc)
        self.type_ = type_

    def __str__(self):
        return 'cannot "defer" or "fulfil" values of type %s' % (
            _str_type(self.type_),)


class DereferenceTypeVMError(StaticVMError):
    '''
    Raised when an attempt is made to dereference a value whose type is not
    a reference.
    '''
    def __init__(self, type_, loc):
        super(DereferenceTypeVMError, self).__init__(loc=loc)
        self.type_ = type_

    def __str__(self):
        return 'cannot dereference values of type %s' % (
            _str_type(self.type_),)


class SubscriptTypeVMError(StaticVMError):
    '''
    Raised when an attempt is made to subscript a value whose type does not
    support subscripting.
    '''
    def __init__(self, type_, loc):
        super(SubscriptTypeVMError, self).__init__(loc=loc)
        self.type_ = type_

    def __str__(self):
        return 'cannot subscript values of type %s' % (
            _str_type(self.type_),)


class SwitchTypeVMError(StaticVMError):
    '''
    Raised when an attempt is made to switch on a value whose type is not
    a union.
    '''
    def __init__(self, type_, loc):
        super(SwitchTypeVMError, self).__init__(loc=loc)
        self.type_ = type_

    def __str__(self):
        return 'cannot switch on values of type %s' % (
            _str_type(self.type_),)


class UnpackTypeVMError(StaticVMError):
    '''
    Raised when an attempt is made to unpack a value that is not a struct.
    '''
    def __init__(self, type_, loc):
        super(UnpackTypeVMError, self).__init__(loc=loc)
        self.type_ = type_

    def __str__(self):
        return 'cannot unpack values of type %s' % (
            _str_type(self.type_),)


class StructIndexVMError(StaticVMError):
    def __init__(self, struct_type, index, loc):
        super(StructIndexVMError, self).__init__(loc=loc)
        self.struct_type = struct_type
        self.index = index

    def __str__(self):
        return 'index %r is out of range for type %s' % (
            self.index, _str_type(self.struct_type))


class StaticnessVMError(StaticVMError):
    '''
    Raised when a non-static or non-constant value is used where a static
    constant value is required.
    '''
    def __init__(self, message, loc):
        assert type(message) is str
        super(StaticnessVMError, self).__init__(loc=loc)
        self.message = message

    def __str__(self):
        return self.message


class FieldAccessVMError(StaticVMError):
    '''
    Raised when accessing a field that does not exist.
    '''
    def __init__(self, type_, name, loc):
        super(FieldAccessVMError, self).__init__(loc=loc)
        self.type_ = type_
        self.name = name

    def __str__(self):
        return "type %s has no field '%s'" % (
            _str_type(self.type_), self.name)


class InexhaustiveSwitchVMError(StaticVMError):
    '''
    Raised when a switch statement does not cover all cases that could occur.
    '''
    def __init__(self, tag, loc):
        super(InexhaustiveSwitchVMError, self).__init__(loc=loc)
        self.tag = tag

    def __str__(self):
        return "case '%s' is not handled" % (self.tag,)


class UnreachableCaseVMError(StaticVMError):
    '''
    Raised when a switch statement handles a case that is not possible
    according to the switch condition type, or when it handles a case twice,
    or when it handles all cases and also provides an "else" clause.
    '''
    def __init__(self, tag, loc):
        '''
         - tag - a tag, or the word "else".
        '''
        super(UnreachableCaseVMError, self).__init__(loc=loc)
        self.tag = tag

    def __str__(self):
        return "case '%s' is not reachable" % (self.tag,)


class BinaryOperatorVMError(StaticVMError):
    '''
    Raised when a binary operator is applied to the wrong types.
    '''
    def __init__(self, op, type1, type2, loc):
        super(BinaryOperatorVMError, self).__init__(loc=loc)
        self.op = op
        self.type1 = type1
        self.type2 = type2

    def __str__(self):
        return 'binary operator "%s" is not defined for (%s, %s)' % (
            self.op, _str_type(self.type1), _str_type(self.type2))


class UnaryOperatorVMError(StaticVMError):
    '''
    Raised when a unary operator is applied to the wrong types.
    '''
    def __init__(self, op, type_, loc):
        super(UnaryOperatorVMError, self).__init__(loc=loc)
        self.op = op
        self.type_ = type_

    def __str__(self):
        return 'binary operator "%s" is not defined for %s' % (
            self.op, _str_type(self.type_))


###


class EncodingVMError(StaticVMError):
    '''
    Superclass of errors in the instruction encoding. These can only be
    caused by generating malformed low-level code; code compiled from
    source should never cause these errors.
    '''


# TODO: Test.
class IllegalInstructionVMError(EncodingVMError):
    '''
    Raised on encountering malformed flat code.
    '''
    def __init__(self, message, loc=None):
        assert type(message) is str
        super(IllegalInstructionVMError, self).__init__(loc=loc)
        self.message = message

    def __str__(self):
        return self.message


# TODO: Test.
class StackUnderflowVMError(EncodingVMError):
    '''
    Raised if flat code attempts to access a stack item before the first.
    '''
    def __init__(self, message, loc=None):
        assert type(message) is str
        super(StackUnderflowVMError, self).__init__(loc=loc)
        self.message = message

    def __str__(self):
        return self.message


###


class RuntimeVMError(VMError):
    '''
    Superclass of all errors which are reported at run time.
    '''
    def __init__(self, loc=None):
        super(RuntimeVMError, self).__init__(loc=loc)


class AssertionVMError(RuntimeVMError):
    '''
    Raised on `assert(false, message)`.
    '''
    def __init__(self, message):
        assert type(message) is str
        super(AssertionVMError, self).__init__()
        self.message = message

    def __str__(self):
        return self.message


class ArithmeticVMError(RuntimeVMError):
    '''
    Raised e.g. on division by zero.
    '''
    def __init__(self, message):
        assert type(message) is str
        super(ArithmeticVMError, self).__init__()
        self.message = message

    def __str__(self):
        return self.message


class FulfilledValueVMError(RuntimeVMError):
    '''
    Raised when an attempt is made to fulfil a deferred value that has
    already been fulfilled.
    '''
    def __init__(self):
        super(FulfilledValueVMError, self).__init__()

    def __str__(self):
        return 'the value has already been fulfilled'


class UnfulfilledValueVMError(RuntimeVMError):
    '''
    Raised when an attempt is made to dereference a deferred value that has
    not been fulfilled.
    '''
    def __init__(self, noun):
        super(UnfulfilledValueVMError, self).__init__()
        assert type(noun) is str
        self.noun = noun

    def __str__(self):
        return '%s has not been fulfilled' % (self.noun,)


class ArrayVMError(RuntimeVMError):
    pass


class ArrayIndexVMError(ArrayVMError):
    '''
    Raised when a 'subscript_array' instruction tries to access an array
    element outside the array bounds.
    '''
    def __init__(self, index, length):
        assert type(index) is int
        assert type(length) is int
        super(ArrayIndexVMError, self).__init__()
        self.index = index
        self.length = length

    def __str__(self):
        return "array index %d is out of range for an array of length %d" % (
            self.index, self.length)


class ArrayRepeatVMError(ArrayVMError):
    '''
    Raised when an 'array_repeat' instruction tries to use a negative repeat
    count.
    '''
    def __init__(self, count):
        assert type(count) is int
        super(ArrayRepeatVMError, self).__init__()
        self.count = count

    def __str__(self):
        return "invalid array repeat count %d" % (self.count,)


class ArraySliceVMError(ArrayVMError):
    '''
    Raised when an 'array_slice' instruction tries to create a slice outside
    the array bounds or of negative length.
    '''
    def __init__(self, start_index, end_index, length):
        assert type(start_index) is int
        assert type(end_index) is int
        assert type(length) is int
        super(ArraySliceVMError, self).__init__()
        self.start_index = start_index
        self.end_index = end_index
        self.length = length

    def __str__(self):
        return (
            "array slice from %d to %d is out of range for an array of "
            "length %d"
        ) % (self.start_index, self.end_index, self.length)
