'''
Simulates the Welly heap using Python data structures.
'''


from ..printer import pretty, auto_wrap


__all__ = [
    'Ref', 'Array',
]


class SliceError(IndexError):
    '''
    Raised when trying to create a slice of an Array if the slice has a
    negative length or is not contained in the array.
    '''


class Ref(object):
    '''
    Represents a Welly reference. Also, for debugging purposes only, records
    the size of the value referred to.

    Public fields:
     - block - A Python list, representing the Welly heap block containing
       the value referred to by this Ref.
     - offset - An index into `block` indicating where the value referred to
       by this Ref begins.
     - size - The number of words occupied by the value referred to by this
       Ref.
    '''
    def __init__(self, block, offset, size):
        assert type(block) is list, block
        assert type(offset) is int, offset
        assert type(size) is int, size
        assert 0 <= offset <= offset+size <= len(block), (
            block, offset, size)
        self.block = block
        self.offset = offset
        self.size = size

    def __eq__(self, other):
        if type(other) is not Ref:
            return False
        elif self.size != other.size:
            # In Welly-in-Python we could be comparing a Ref to a struct
            # with a Ref to its first field. These two Refs would have the
            # same `block` and `offset`.
            # In real Welly, this is impossible due to type-checking; the two
            # Refs have different types.
            # In order to give a sensible answer in this Python-only case, we
            # distinguish the two Refs using their `size`.
            return False
        else:
            return self.block is other.block and self.offset == other.offset

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self.block) + self.offset

    def __repr__(self):
        return pretty(self.__pretty__, height=1)

    def __pretty__(self):
        '''
        Returns a 4-tuple as required by `welly.printer.pretty()`.
        '''
        return (
            'Ref(',
            ', ',
            ')',
            [
                auto_wrap(self.block),
                auto_wrap(self.offset),
                auto_wrap(self.size),
            ],
        )

    def read(self):
        '''
        Returns a tuple of words representing the value referred to by this
        Ref.
        '''
        return tuple(self.block[self.offset:self.offset+self.size])

    def write(self, value):
        '''
        Overwrites the value referred to by this Ref with the specified tuple
        of words.
        '''
        assert len(value) == self.size
        self.block[self.offset:self.offset+self.size] = value

    def field(self, offset, size):
        '''
        Returns a Ref to a field within the value referrred to by this Ref.

         - offset - The index within the value at which the field starts.
         - size - The size in words of the field.
        '''
        assert 0 <= offset <= offset+size <= self.size
        return Ref(self.block, self.offset+offset, size)

    def copy(self):
        '''
        Returns a Ref to a fresh copy of the value referred to by this Ref.
        '''
        return Ref.new(self.read())

    @staticmethod
    def block(block):
        '''
        Constructs a reference to an entire heap block.

         - block - a Python list representing a heap block.
        '''
        return Ref(block, 0, len(block))

    @staticmethod
    def malloc(size):
        '''
        Returns a fresh block of length `size`.
        '''
        return Ref.block([None]*(size))

    @staticmethod
    def new(value):
        '''
        Returns a reference to a fresh block containing a copy of `value`.

         - value - a tuple of words.
        '''
        return Ref.block(list(value))


class Array(object):
    '''
    Represents a reference from Python to a Welly array. Such a reference
    always points to en entire heap block.

    This class is provided for the convenience of Python programs, and should
    not be confused with the Welly array types. Instances can easily be
    converted to and from Welly arrays.

    This class defines `__len__()`, `__getitem()__` methods so that it
    behaves like an immutable Python sequence of Refs to the array elements.

    Public fields:
     - ref - a Ref to a contiguous sequence of array elements. The `size`
       field of the Ref must include the whole sequence.
     - length - the number of array elements in the contiguous sequence.
     - elem_size - the size in words of an array element.
    '''
    def __init__(self, ref, length, elem_size):
        assert type(ref) is Ref
        assert type(length) is int
        assert type(elem_size) is int
        assert ref.size == length * elem_size
        self.ref = ref
        self.length = length
        self.elem_size = elem_size

    def __repr__(self):
        return 'Array(%r, %r, %r)' % (self.ref, self.length, self.elem_size)

    def __len__(self):
        return self.length

    def __getitem__(self, index):
        if not 0 <= index < self.length:
            raise IndexError(
                "Array index %d out of range %d" % (index, self.length))
        return self.ref.field(index * self.elem_size, self.elem_size)

    def read(self):
        '''
        Yields tuples of words representing the array elements.
        '''
        return (ref.read() for ref in self)

    def slice(self, start, end):
        if 0 > start or start > end or end > self.length:
            raise SliceError(
                "Slice from %d to %d out of range for array of length %d" %
                (start, end, self.length)
            )
        length = end - start
        return Array(
            self.ref.field(start * self.elem_size, length * self.elem_size),
            length,
            self.elem_size,
        )

    def to_welly(self):
        '''
        Converts this Array into a Welly array: a `(Ref, length)` pair.
        Note that this discards `elem_size`; the caller should know it.
        '''
        return self.ref, self.length

    @staticmethod
    def from_welly(ref_and_length, elem_size):
        '''
        Equivalent to `Array(ref, length, elem_size)` where `ref_and_length`
        is the pair `(ref, length)`.
        '''
        ref, length = ref_and_length
        return Array(ref, length, elem_size)

    @staticmethod
    def malloc(length, elem_size):
        return Array(Ref.malloc(elem_size*length), length, elem_size)

    def to_list(self):
        '''
        Converts this Array into a Python list. Equivalent to `list(self)`.
        '''
        return list(self)

    @staticmethod
    def from_list(tuples, elem_size):
        '''
        Construct a Welly array from a Python list.

         - tuples - a Python list of Welly values, each represented by a tuple
           of words.
         - elem_size - the size in words of each array element.
        '''
        assert all(type(t) is tuple for t in tuples)
        assert all(len(t) == elem_size for t in tuples)
        array = Array.malloc(len(tuples), elem_size)
        for ref, t in zip(array, tuples):
            ref.write(t)
        return array
