from . import defer, const, wtypes, subtypes, pretty_type

from ..printer import pretty, auto_wrap
from .memory import Ref, Array
from .pack import pack, unpack
from .wtypes import BASE, CONST, VAR, sizeof


__all__ = [
    'Dynamic',
    'struct_length', 'struct_field_types', 'struct_offsets', 'struct_iter',
    'assign',
]


class Dynamic(object):
    '''
    A Python representation of a Welly dynamic value. Groups the constness
    and type of a Welly value with a Ref to it.

    This class is provided for the convenience of Python programs, and should
    not be confused with the Welly dynamic type. Instances can easily be
    converted to and from Welly dynamic values.

    Public fields:
     - type_ - a Ref to the Welly type of the value.
     - ref - a Ref to the value.
     - constness - a Ref to the constness of the value: BASE, CONST or VAR.
    '''
    def __init__(self, type_, ref, constness=CONST):
        const.check_valid(constness)
        assert type(ref) is Ref
        assert sizeof(type_) == ref.size
        self.type_ = type_
        self.ref = ref
        self.constness = constness

    def __call__(self):
        '''
        Returns a 4-tuple as required by `welly.printer.pretty()`.
        '''
        return (
            "Dynamic(",
            ", ",
            ")",
            [
                pretty_type.wrap(self.type_),
                self.ref.__pretty__,
                auto_wrap(const.switch(self.constness)),
            ],
        )

    def __repr__(self):
        return pretty(self, height=1)

    def is_var(self):
        return const.switch(self.constness) == 'VAR'

    def is_const(self):
        return const.switch(self.constness) == 'CONST'

    def read(self):
        '''Equivalent to `self.ref.read()`.'''
        return self.ref.read()

    def write(self, value):
        '''Equivalent to `self.ref.write(value)`.'''
        return self.ref.write(value)

    def is_a(self, type_):
        '''
        Tests whether `self.type_` is a subtype of `type_`.

         - type_ - a Ref to a type.
        '''
        return subtypes.is_subtype(self.type_, type_)

    def check_is_a(self, type_, noun, loc):
        '''
        Tests whether `self.type_` is a subtype of `type_`.

         - type_ - a Ref to a Welly type.
         - noun - a noun phrase that describes `self`, e.g. "return type of
           function", or `None` to use a default noun phrase.
         - loc - an opaque source code location for use in error messages.
        '''
        subtypes.check(self.type_, type_, noun, loc)

    def copy(self, loc, type_=None, constness=CONST):
        '''
        Returns a Dynamic referring to a fresh copy of this Dynamic.

         - loc - an opaque source code location for use in error messages.
         - type_ - a Ref to the desired type of the copy, or None to use the
           type of `x`.
         - constness - a Ref to the desired constness of the copy.
        '''
        if type_ is None:
            type_ = self.type_
        else:
            subtypes.check(self.type_, type_, None, loc)
        return Dynamic(type_, self.ref.copy(), constness=constness)

    def to_welly(self):
        '''
        Converts this Dynamic into a Welly dynamic value.
        '''
        return pack(self.constness.read(), self.type_.read(), self.ref)

    @staticmethod
    def from_welly(dynamic_value):
        '''
        Constructs a Dynamic from `dynamic_value`.

         - dynamic_value - a Ref to a Welly dynamic value.
        '''
        constness, type_, ref = unpack(dynamic_value, 'r2r2w1')
        return Dynamic(type_, ref, constness=constness)

    @staticmethod
    def malloc(type_, constness=CONST):
        '''
        Constructs a Dynamic that refers to uninitialised memory.

         - type_ - a Ref to the desired type of the value.
         - constness - a Ref to the desired constness of the value.
        '''
        return Dynamic(
            type_,
            Ref.malloc(sizeof(type_)),
            constness=constness,
        )

    @staticmethod
    def new(type_, value, constness=CONST):
        '''
        Constructs a Dynamic that refers to a freshly allocated value.

         - type_ - a Ref to the desired type of the value.
         - value - the words of the value: a tuple.
         - constness - a Ref to the desired constness of the value.
        '''
        return Dynamic(type_, Ref.new(value), constness=constness)


# Destructors for primitive types.


def is_int_type(type_):
    t0, _ = wtypes.switch(type_)
    return t0 == 'INT'


# Constructors and destructors for "array" types.


def is_array_type(type_):
    '''
    Tests whether `type_` is `array0` or an array type.
    '''
    # Equivalent to `is_subtype(array0, type_)`.
    t0, _ = wtypes.switch(type_)
    return t0 in ('ARRAY0', 'ARRAY')


def array_constness(array_type, z=CONST):
    '''
    Given an array type, returns a Ref to the constness of the elements.

     - array_type - a Ref to an 'ARRAY0' or 'ARRAY' type.
     - z - the result to use for 'ARRAY0' (which has no elements).
    '''
    t0, payload = wtypes.switch(array_type)
    if t0 == 'ARRAY0':
        return z
    elif t0 == 'ARRAY':
        constness, _ = payload
        return constness
    assert False, array_type


def array_element_type(array_type):
    '''
    Given an array type, returns a Ref to the type of the elements.

     - array_type - a Ref to an 'ARRAY' type.
    '''
    t0, payload = wtypes.switch(array_type)
    if t0 == 'ARRAY':
        _, elem_t = payload
        return elem_t
    assert False, array_type


# Constructors and destructors for "struct" types.


def struct_length(struct_type):
    '''
    Returns the number of fields in a value of type `struct_type`.
     - struct_type - a Ref to a type.
    '''
    t0, payload = wtypes.switch(struct_type)
    assert t0 == 'STRUCT', struct_type
    (fields,) = payload
    return len(fields)


def struct_field_types(struct_type):
    '''
    Returns an Array of the types of the fields of `struct_type`.

    Params:
     - struct_type - a Ref to a struct type.
    '''
    t0, payload = wtypes.switch(struct_type)
    assert t0 == 'STRUCT', struct_type
    (field_types,) = payload
    return field_types


def struct_offsets(struct_type):
    '''
    Iterates over a struct and yield (type, offset), where:
     - type is a Ref to the type of a field of the struct, and
     - offset is its offset in words from the beginning of the struct.

    Params:
     - struct_type - a Ref to the type of the struct.
    '''
    offset = 0
    for type_ in struct_field_types(struct_type):
        yield type_, offset
        offset += sizeof(type_)


def struct_iter(struct):
    '''
    Iterates over `struct` and yields a Dynamic representing each field.

     - struct - a Dynamic.
    '''
    offset = 0
    for type_ in struct_field_types(struct.type_):
        size = sizeof(type_)
        ref = struct.ref.field(offset, size)
        yield Dynamic(type_, ref, constness=struct.constness)
        offset += size
    assert offset == struct.ref.size


# Constructors and destructors for "union" types.


def union_cases(union_type):
    '''
    Extracts the union_cases of `union_type` and passes them to
    `wtypes.union_index()`. Returns the resulting OrderedDict, which maps
    case name (str) to Python list of Ref to Welly type.

    Params:
     - union_type - a Ref to a union type.
    '''
    t0, payload = wtypes.switch(union_type)
    assert t0 == 'UNION', union_type
    (cases,) = payload
    return wtypes.union_index(cases)


# "Arithmetic" operations. Arguments and results are passed as Dynamics.


def assign(lvalue, rvalue, loc):
    '''
    Calls `lvalue.write(rvalue.read())` after checking the types.
    '''
    rvalue.check_is_a(lvalue.type_, None, loc)
    lvalue.write(rvalue.read())
