from . import defer

from .memory import Ref, Array


__all__ = [
    'pack', 'unpack', 'tag', 'switch'
]


def pack(*values):
    '''
    A helper function to help Python programs to pack Welly structs. Returns
    a tuple of words formed by applying the following conversions to each of
    `values` and concatenating the results:
     - A tuple is interpreted as a sequence of words.
     - An Array is turned into a pair of words by calling `to_welly()`.
     - Any other value is interpreted as a single word.
    '''
    result = []
    for value in values:
        if type(value) is tuple:
            result.extend(value)
        elif type(value) is Array:
            result.extend(value.to_welly())
        else:
            result.append(value)
    return tuple(result)


def unpack(struct, fmt):
    '''
    A helper function to help Python programs to unpack Welly structs.

    Returns a tuple of Python values representing the fields of the value
    referred to by this Ref. `fmt` specifies the sizes of the fields and
    the Welly-to-Python conversion you would like to be applied.

     - struct - a Ref to a Welly struct value.
     - fmt - A str containing alternating letters and digits. Each
       (letter, digit) pair describes one field of the struct, as follows:
        - 'a' - The field is an array. Return it as an Array. The digit
          specifies the size in words of the array elements (the size of the
          field itself is `2`).
        - 'r' - Return a Ref to the field. The digit specifies the size in
          words of the field.
        - 'v' - Return the value of the field as a tuple of words. The digit
          specifies the size in words of the field.
        - 'w' - Return the value of the field as a single word. The digit must
          be `1` (meaning a single word).
       The total size of the fields must match the size of the struct.
       Example: The string 'w1v0r5a3' describes a struct with four fields.
       The first field is returned as a word. The second is returned as an
       empty tuple. The third is returned as a Ref to a value occupying five
       words. The fourth is returned as an Array with an element size
       of `3`. The total size is eight words (1+0+5+2).
    '''
    result = []
    offset = 0
    for pos in range(0, len(fmt), 2):
        letter = fmt[pos]
        size = int(fmt[pos+1])
        if letter == 'a':
            value = Array.from_welly(struct.field(offset, 2).read(), size)
            size = 2
        elif letter == 'r':
            value = struct.field(offset, size)
        elif letter == 'v':
            value = struct.field(offset, size).read()
        elif letter == 'w':
            assert size == 1
            (value,) = struct.field(offset, size).read()
        else:
            assert False, 'Unknown format code: %s%d' % (letter, size)
        result.append(value)
        offset += size
    assert offset == struct.size
    return tuple(result)


def tag(tag, *values):
    '''
    Constructs a union value with tag `tag` and payload `values`. The payload
    is converted from Python types using `pack()`. Returns a Ref.
    '''
    assert type(tag) is str
    payload = list(pack(*values))
    return Ref.new((tag, payload))


def switch(ref, fmts):
    '''
    Simulates a Welly "switch" statement. Takes a Ref to a value of some union
    type, and returns its tag and its fields in a convenient form for use in
    Python.

     - ref - a Ref to a union value.
     - fmts - a dict from tag (str) to fmt (str).
    Returns `(tag, fields)` where `tag` is a str and `fields` is the tuple
    returned by `unpack(payload, fmts[tag])`.
    '''
    assert type(ref) is Ref, ref
    v0, payload = defer.read(ref)
    assert type(v0) is str, ref
    fields = unpack(Ref.block(payload), fmts[v0])
    return v0, fields
