'''
Interfaces the pretty printer to the Welly type system.
'''


from . import defer, const, wtypes, subtypes

from .errors import UnfulfilledValueVMError
from .memory import Ref, Array


__all__ = ['wrap']


def _constant(value):
    assert type(value) is str
    return lambda: (value, '', '', [])


_primitive_types = {
    'BOOL': 'bool',
    'FLOAT': 'float',
    'INT': 'int',
    'STR': 'str',
    'ARRAY0': 'array0',
    'DYNAMIC': 'dynamic',
}


def _constness_syntax(constness):
    '''
    Returns '', 'const ' or 'var '. This is used to construct strings such as
    "ref const int" or "array var str".
    '''
    c0 = const.switch(constness)
    if c0 == 'BASE':
        return ''
    elif c0 == 'CONST':
        return 'const '
    elif c0 == 'VAR':
        return 'var '
    assert False


def _bracketed(contents, tag=''):
    '''
    Returns a printable of the form "tag(contents[0], contents[1], ...)"

     - contents - an iterable returning Refs to Welly types.
     - tag - (optional) a word to put before the open bracket.
    '''
    assert type(tag) is str
    return lambda: ('%s(' % tag, ', ', ')', [wrap(x) for x in contents])


def wrap(type_):
    '''
    Given a Welly type, returns a pretty-print-aware wrapper around it.

     - type_ - a Ref to a Welly type.
    '''
    try:
        tag, payload = wtypes.switch(type_)
    except UnfulfilledValueVMError:
        return _constant('unfulfilled')
    if tag in _primitive_types:
        return _constant(_primitive_types[tag])
    elif tag == 'REF':
        constness, referee_type = payload
        return lambda: (
            'ref ' + _constness_syntax(constness),
            '',
            '',
            [wrap(referee_type)]
        )
    elif tag == 'ARRAY':
        constness, elem_type = payload
        return lambda: (
            'array ' + _constness_syntax(constness),
            '',
            '',
            [wrap(elem_type)]
        )
    elif tag == 'STRUCT':
        (field_type_array,) = payload
        if subtypes.eq(type_, wtypes.void):
            return _constant('void')
        else:
            return _bracketed(field_type_array, tag='struct')
    elif tag == 'UNION':
        (case_array,) = payload
        if subtypes.eq(type_, wtypes.type_):
            return _constant('type')
        return lambda: (
            'union {',
            '; ',
            '}',
            [
                _bracketed(types, tag=tag)
                for (tag, types) in wtypes.union_index(case_array).items()
            ]
        )
    elif tag == 'FUNCTION':
        param_type_array, return_type = payload
        return lambda: (
            'fn', ': ', '',
            [_bracketed(param_type_array), _bracketed([return_type])]
        )
    assert False, t0
