import re

from welly.parser import (
    grammar, lexer, brackets, atom, unary_expr, expr, expr_list, stmt,
    block, astbuilder
)

from welly.parser.token import ASCII, NonASCII
from welly.parser.astbuilder import ASTError


__all__ = ['WellyParser', 'ASTError']


class WellyParser(object):
    '''
    Represents the state of a parser for the Welly language. Feed it
    characters (source code) using the `feed()` method. Call `flush()` to
    force all remaining input to be processed, e.g. at EOF. Complete
    statements and error tokens will be appended to `self.buffer`. Call
    `do_pending_stmts()` to process the contents of the buffer. Each
    statement in turn will be checked for errors, converted to an abstract
    syntax tree, and passed to `do_stmt()`.

    Public fields:
     - buffer - a grammar.Buffer for collecting complete statements and
       errors.
     - parser - a grammar.Parser object.
     - pos - the number of characters received so far.
    '''
    def __init__(self):
        '''
        Constructs a Welly parser, given its `do_stmt()` callback.
        '''

        def make_stmt_parser(p):
            p = grammar.Parser(block.block, p)
            p = grammar.Parser(stmt.stmt, p)
            p = grammar.Parser(expr.expr, p)
            p = grammar.Parser(unary_expr.unary_expr, p)
            p = grammar.Parser(atom.atom, p)
            return p

        def make_expr_parser(p):
            p = grammar.Parser(expr_list.expr_list, p)
            p = grammar.Parser(expr.expr, p)
            p = grammar.Parser(unary_expr.unary_expr, p)
            p = grammar.Parser(atom.atom, p)
            return p

        self.buffer = grammar.Buffer()
        p = make_stmt_parser(self.buffer)
        p = brackets.BracketMatcher(p, {
            '(': make_expr_parser,
            '[': make_expr_parser,
            '{': make_stmt_parser,
        })
        p = grammar.Parser(lexer.high_level_lexer, p)
        p = grammar.Parser(lexer.low_level_lexer, p)
        self.parser = p
        self.pos = 0

    def feed(self, str_):
        '''
        Parses `str_`. Complete statements and error tokens will be added to
        `self.buffer`. This method does not raise exceptions.

         - str_ - Source code, as a Python `str`.
        '''
        assert type(str_) is str, type(str_)
        for c in str_:
            old_pos = self.pos
            self.pos += 1
            if ord(c) < 0x80:
                t = ASCII(old_pos, c)
            else:
                t = NonASCII(old_pos, self.pos, c)
            self.parser.feed(t)

    def flush(self):
        '''
        Processes any buffered characters and restores this WellyParser to its
        home state. Note that if the buffer contains incomplete data, it will
        be truncated, which may cause error tokens.

        This method does not raise exceptions.
        '''
        self.parser.flush()

    def is_flushed(self):
        return self.parser.is_flushed()

    def do_pending_stmts(self):
        '''
        Removes all stmts in `self.buffer`. For each one in turn, checks it
        for errors, and raises ASTError if necessary, otherwise constructs an
        abstract syntax tree and passes it to `self.do_stmt()`. Exceptions
        (including ASTError) cause the rest of the buffer to be discarded.
        '''
        to_run = self.buffer.contents[:]
        self.buffer.contents[:] = []
        for statement in to_run:
            ast = astbuilder.visit_stmt(statement)
            self.do_stmt(ast)

    def do_stmt(self, ast):
        '''
        Called by `do_pending_stmts()` for every error-free statement in
        `self.buffer`. Subclasses must implement this method.
        '''
        raise NotImplementedError('abstract method')
