from .token import Token


def make_loc(token):
    return (token.start, token.end)


class AST(object):
    '''
    Superclass of all classes that represent Abstract Syntax Trees of Welly
    source code.

     - tag - a str that can be tested to discriminate syntactic elements.
     - loc - a (start, end) pair, representing the range of byte offsets in
       the source code corresponding to this AST.
    '''
    def __init__(self, tag, token, error_token, children):
        '''
         - token - a Token that spans the syntactic element. Used to compute
           `start` and `end`.
         - children - a list of the ASTs that are syntactically inside this
           one, in left-to-right order.
         - error_token - an optional Token indicating a good place to report
           errors. If `None`, `token` will be used.
        '''
        assert isinstance(token, Token)
        assert isinstance(error_token, Token)
        assert type(children) is list
        assert token.start <= error_token.start
        assert error_token.end <= token.end
        pos = token.start
        for child in children:
            if child is not None:
                assert isinstance(child, AST)
                (start, end) = child.loc
                assert pos <= start <= end
                pos = end
        assert pos <= token.end
        self.tag = tag
        self.loc = make_loc(token)
        self.error_loc = make_loc(error_token)
        super(AST, self).__init__()

    def matches(self, tag, *pattern):
        '''
        Returns `True` if `self.tag` equals `tag` and if this AST matches
        `pattern` (in some sense to be defined by the subclass).
        '''
        raise NotImplementedError("abstract method")


def matches(ast, pattern):
    '''
    If `ast` is an AST, returns `ast.matches(*pattern)`. If it's a list,
    recursively matches the contents. Otherwise returns `ast == pattern`.

     - ast - a list or an AST or a value that can be tested using "==".
    '''
    if type(ast) is list:
        return (
            type(pattern) is list and
            len(ast) == len(pattern) and
            all(matches(a, p) for a, p in zip(ast, pattern))
        )
    elif isinstance(ast, AST):
        return type(pattern) is tuple and ast.matches(*pattern)
    else:
        return ast == pattern


class ExprAST(AST):
    '''
    Superclass of all classes that represent Abstract Syntax Trees of Welly
    expressions.

     - is_lvalue - `True` if this expression can appear on the left-hand side
       of an assignment.
    '''

    def __init__(self, tag, token, error_token, children, is_lvalue=False):
        super(ExprAST, self).__init__(tag, token, error_token, children)
        self.is_lvalue = is_lvalue

    def matches(self, tag, is_lvalue, *pattern):
        '''
        Returns `True` if `self.tag` equals `tag` and `self.is_lvalue` equals
        `is_lvalue` and if this AST matches `pattern` (in some sense to be
        defined by the subclass).
        '''
        raise NotImplementedError("abstract method")


class Keyword(AST):
    '''
    An AST that represents a keyword. Its tag matches the keyword.
    '''
    def __init__(self, tag, token):
        assert tag == token.tag
        super(Keyword, self).__init__(tag, token, token, [])

    def matches(self, tag):
        return self.tag == tag


EXPR_TAGS = {
    'character', 'identifier', 'integer', 'hex_integer', 'float', 'string',
    'box', 'defer', 'new', 'typeof', 'dereference', 'address_of',
    'field', 'call', 'subscript', 'cast', 'unary', 'binary', 'ternary',
    'type_ref',
    'array', 'type_array',
    'struct', 'type_struct',
    'function_def', 'type_function',
    'union', 'type_union',
}
STMT_TAGS = {
    'assign', 'break', 'continue', 'decl', 'expr', 'for', 'fulfil', 'if',
    'return', 'switch', 'throw', 'try', 'while',
}
CONSTNESS_TAGS = {
    'const', 'var',
}


def is_expr(ast):
    return ast.tag in EXPR_TAGS


def is_lvalue(expr):
    assert is_expr(expr)
    return expr.is_lvalue


class AddressOf(ExprAST):
    def __init__(self, token, error_token, op, expr):
        assert op.tag == 'un_op'
        assert op.s == '&'
        assert is_expr(expr)
        super(AddressOf, self).__init__(
            'address_of', token, error_token, [expr])
        self.op = op
        self.expr = expr

    def matches(self, tag, is_lvalue, expr):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.expr, expr)
        )


class Array(ExprAST):
    def __init__(self, token, error_token, exprs):
        assert type(exprs) is list
        assert all(is_expr(e) for e in exprs), exprs
        super(Array, self).__init__('array', token, error_token, exprs)
        self.exprs = exprs

    def matches(self, tag, is_lvalue, exprs):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.exprs, exprs)
        )


class Assign(AST):
    def __init__(self, token, error_token, lhs, op, rhs):
        assert is_expr(lhs) and is_lvalue(lhs)
        assert op.tag == 'assign_op'
        assert is_expr(rhs)
        super(Assign, self).__init__(
            'assign', token, error_token, [lhs, op, rhs])
        self.lhs = lhs
        self.op = op
        self.rhs = rhs

    def matches(self, tag, lhs, op, rhs):
        return (
            self.tag == tag and
            matches(self.lhs, lhs) and
            matches(self.op, op) and
            matches(self.rhs, rhs)
        )


# TODO: Maybe delete this class now that Assign has an `error_token`?
class AssignOp(AST):
    def __init__(self, token, s):
        assert token.tag == 'assign_op'
        assert type(s) is str
        super(AssignOp, self).__init__('assign_op', token, token, [])
        self.s = s

    def matches(self, tag, s):
        return (
            self.tag == tag and
            self.s == s
        )


# TODO: Maybe delete this class now that Binary has an `error_token`?
class BinOp(AST):
    def __init__(self, token, s):
        assert token.tag in ('bin_op', 'un_bin_op')
        assert type(s) is str
        super(BinOp, self).__init__('bin_op', token, token, [])
        self.s = s

    def matches(self, tag, s):
        return (
            self.tag == tag and
            self.s == s
        )


class Binary(ExprAST):
    def __init__(self, token, error_token, left, op, right):
        assert is_expr(left)
        assert op.tag == 'bin_op'
        assert is_expr(right)
        super(Binary, self).__init__(
            'binary', token, error_token, [left, op, right])
        self.left = left
        self.op = op
        self.right = right

    def matches(self, tag, is_lvalue, left, op, right):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.left, left) and
            matches(self.op, op) and
            matches(self.right, right)
        )


class Block(AST):
    def __init__(self, token, error_token, stmts):
        assert all(s.tag in STMT_TAGS for s in stmts)
        super(Block, self).__init__('block', token, error_token, stmts)
        self.stmts = stmts

    def matches(self, tag, stmts):
        return (
            self.tag == tag and
            matches(self.stmts, stmts)
        )


class Box(ExprAST):
    def __init__(self, token, error_token, expr):
        assert is_expr(expr)
        super(Box, self).__init__('box', token, error_token, [expr])
        self.expr = expr

    def matches(self, tag, is_lvalue, expr):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.expr, expr)
        )


class Call(ExprAST):
    def __init__(self, token, error_token, expr, args):
        assert is_expr(expr)
        assert type(args) is list
        assert all(is_expr(a) for a in args)
        super(Call, self).__init__('call', token, error_token, [expr] + args)
        self.expr = expr
        self.args = args

    def matches(self, tag, is_lvalue, expr, args):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.expr, expr) and
            matches(self.args, args)
        )


class Case(AST):
    def __init__(self, token, error_token, keyword, name, params, body):
        assert keyword.tag == 'case'
        assert name.tag == 'tag'
        assert type(params) is list
        assert all(p.tag == 'identifier' for p in params)
        assert body.tag == 'block'
        super(Case, self).__init__(
            'case', token, error_token, [keyword, name] + params + [body])
        self.keyword = keyword
        self.name = name
        self.params = params
        self.body = body

    def matches(self, tag, name, params, body):
        return (
            self.tag == tag and
            matches(self.name, name) and
            matches(self.params, params) and
            matches(self.body, body)
        )


class Cast(ExprAST):
    def __init__(self, token, error_token, expr, op, type_):
        assert is_expr(expr)
        assert op.tag == 'bin_op'
        assert op.s == ':'
        assert is_expr(type_)
        super(Cast, self).__init__(
            'cast',
            token,
            error_token,
            [expr, op, type_],
            is_lvalue=expr.is_lvalue,
        )
        self.expr = expr
        self.op = op
        self.type_ = type_

    def matches(self, tag, expr, type_):
        return (
            self.tag == tag and
            matches(self.expr, expr) and
            matches(self.type_, type_)
        )


class Character(ExprAST):
    '''
     - c - a unicode character of length 1.
    '''
    def __init__(self, token, c):
        assert token.tag == 'character'
        assert type(c) is str
        assert len(c) == 1
        super(Character, self).__init__('character', token, token, [])
        self.c = c

    def matches(self, tag, is_lvalue, c):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            self.c == c
        )


class Decl(AST):
    def __init__(self, token, error_token, constness, lhs, rhs):
        assert constness.tag in CONSTNESS_TAGS
        assert is_expr(lhs) and is_lvalue(lhs)
        assert is_expr(rhs)
        super(Decl, self).__init__(
            'decl', token, error_token, [constness, lhs, rhs])
        self.constness = constness
        self.lhs = lhs
        self.rhs = rhs

    def is_var(self):
        return self.constness.tag == 'var'

    def matches(self, tag, constness, lhs, rhs):
        return (
            self.tag == tag and
            matches(self.constness, constness) and
            matches(self.lhs, lhs) and
            matches(self.rhs, rhs)
        )


class Defer(ExprAST):
    def __init__(self, token, error_token, expr):
        assert is_expr(expr)
        super(Defer, self).__init__('defer', token, error_token, [expr])
        self.expr = expr

    def matches(self, tag, is_lvalue, expr):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.expr, expr)
        )


class Dereference(ExprAST):
    def __init__(self, token, error_token, op, expr):
        assert op.tag == 'un_op'
        assert op.s == '*'
        assert is_expr(expr)
        super(Dereference, self).__init__(
            'dereference', token, error_token, [op, expr], is_lvalue=True)
        self.op = op
        self.expr = expr

    def matches(self, tag, is_lvalue, expr):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.expr, expr)
        )


class Else(AST):
    def __init__(self, token, error_token, keyword, body):
        assert body.tag == 'block'
        super(Else, self).__init__('else', token, error_token, [])
        self.keyword = keyword
        self.body = body

    def matches(self, tag, body):
        return (
            self.tag == tag and
            matches(self.body, body)
        )


class Expr(AST):
    def __init__(self, token, error_token, expr):
        assert is_expr(expr)
        super(Expr, self).__init__('expr', token, error_token, [expr])
        self.expr = expr

    def matches(self, tag, expr):
        return (
            self.tag == tag and
            matches(self.expr, expr)
        )


class Field(ExprAST):
    def __init__(self, token, error_token, expr, name):
        assert is_expr(expr)
        assert name.tag == 'identifier'
        super(Field, self).__init__(
            'field',
            token,
            error_token,
            [expr, name],
            is_lvalue=expr.is_lvalue,
        )
        self.expr = expr
        self.name = name

    def matches(self, tag, is_lvalue, expr, name):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.expr, expr) and
            matches(self.name, name)
        )


class Float(ExprAST):
    def __init__(self, token, s):
        assert token.tag == 'float'
        assert type(s) is str
        super(Float, self).__init__('float', token, token, [])
        self.s = token.s

    def matches(self, tag, is_lvalue, s):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            self.s == s
        )


class For(AST):
    def __init__(self, token, error_token, identifier, expr, body, else_):
        assert identifier.tag == 'identifier'
        assert is_expr(expr)
        assert body.tag == 'block'
        assert else_ is None or else_.tag == 'block'
        super(For, self).__init__(
            'for', token, error_token, [identifier, expr, body, else_]
        )
        self.identifier = identifier
        self.expr = expr
        self.body = body
        self.else_ = else_

    def matches(self, tag, identifier, expr, body, else_):
        return (
            self.tag == tag and
            matches(self.identifier, identifier) and
            matches(self.expr, expr) and
            matches(self.body, body) and
            matches(self.else_, else_)
        )


class Fulfil(AST):
    def __init__(self, token, error_token, lhs, rhs):
        assert is_expr(lhs)
        assert is_expr(rhs)
        super(Fulfil, self).__init__('fulfil', token, error_token, [lhs, rhs])
        self.lhs = lhs
        self.rhs = rhs

    def matches(self, tag, lhs, rhs):
        return (
            self.tag == tag and
            matches(self.lhs, lhs) and
            matches(self.rhs, rhs)
        )


class FunctionDef(ExprAST):
    def __init__(self, token, error_token, name, params, return_type, body):
        assert name is None or name.tag == 'identifier'
        assert type(params) is list
        assert all(p.tag == 'param' for p in params)
        assert return_type is None or is_expr(return_type)
        assert body.tag == 'block'
        super(FunctionDef, self).__init__(
            'function_def',
            token,
            error_token,
            [name] + params + [return_type, body],
        )
        self.name = name
        self.params = params
        self.return_type = return_type
        self.body = body

    def matches(self, tag, is_lvalue, name, params, return_type, body):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.name, name) and
            matches(self.params, params) and
            matches(self.return_type, return_type) and
            matches(self.body, body)
        )


class HexInteger(ExprAST):
    def __init__(self, token, s):
        assert token.tag == 'hex_integer'
        assert type(s) is str
        super(HexInteger, self).__init__('hex_integer', token, token, [])
        self.s = token.s

    def matches(self, tag, is_lvalue, s):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            self.s == s
        )


class Identifier(ExprAST):
    '''
     - s - the name of the identifier (a str).
    '''
    def __init__(self, token, s):
        assert token.tag == 'identifier'
        assert type(s) is str
        super(Identifier, self).__init__(
            'identifier', token, token, [], is_lvalue=True)
        self.s = s

    def matches(self, tag, is_lvalue, s):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            self.s == s
        )

    def __repr__(self):
        return 'Identifier(%r)' % (self.s,)


class If(AST):
    def __init__(self, token, error_token, condition, then, else_):
        assert is_expr(condition)
        assert then.tag == 'block'
        assert else_ is None or else_.tag == 'block'
        super(If, self).__init__(
            'if', token, error_token, [condition, then, else_])
        self.condition = condition
        self.then = then
        self.else_ = else_

    def matches(self, tag, condition, then, else_):
        return (
            self.tag == tag and
            matches(self.condition, condition) and
            matches(self.then, then) and
            matches(self.else_, else_)
        )


class Integer(ExprAST):
    def __init__(self, token, s):
        assert token.tag == 'integer'
        assert type(s) is str
        super(Integer, self).__init__('integer', token, token, [])
        self.s = token.s

    def matches(self, tag, is_lvalue, s):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            self.s == s
        )


class New(ExprAST):
    def __init__(self, token, error_token, expr):
        assert is_expr(expr)
        super(New, self).__init__('new', token, error_token, [expr])
        self.expr = expr

    def matches(self, tag, is_lvalue, expr):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.expr, expr)
        )


class Param(AST):
    def __init__(self, token, error_token, name, type_):
        assert name is None or name.tag == 'identifier'
        assert is_expr(type_)
        super(Param, self).__init__(
            'param', token, error_token, [name, type_])
        self.name = name
        self.type_ = type_

    def matches(self, tag, name, type_):
        return (
            self.tag == tag and
            matches(self.name, name) and
            matches(self.type_, type_)
        )


class Return(AST):
    def __init__(self, token, error_token, expr):
        assert expr is None or is_expr(expr)
        super(Return, self).__init__('return', token, error_token, [expr])
        self.expr = expr

    def matches(self, tag, expr):
        return (
            self.tag == tag and
            matches(self.expr, expr)
        )


class String(ExprAST):
    def __init__(self, token, s):
        assert token.tag == 'string'
        assert type(s) is str
        super(String, self).__init__('string', token, token, [])
        self.s = s

    def matches(self, tag, is_lvalue, s):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            self.s == s
        )


class Struct(ExprAST):
    def __init__(self, token, error_token, exprs):
        assert type(exprs) is list
        assert all(is_expr(e) for e in exprs)
        all_lvalues = all(is_lvalue(e) for e in exprs)
        super(Struct, self).__init__(
            'struct', token, error_token, exprs, is_lvalue=all_lvalues)
        self.exprs = exprs

    def matches(self, tag, is_lvalue, exprs):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.exprs, exprs)
        )


class Subscript(ExprAST):
    def __init__(self, token, error_token, expr, index):
        assert is_expr(expr)
        assert is_expr(index)
        super(Subscript, self).__init__(
            'subscript',
            token,
            error_token,
            [expr, index],
            is_lvalue=True,
        )
        self.expr = expr
        self.index = index

    def matches(self, tag, is_lvalue, expr, index):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.expr, expr) and
            matches(self.index, index)
        )


class Switch(AST):
    def __init__(self, token, error_token, expr, cases, else_):
        assert is_expr(expr)
        assert type(cases) is list
        assert all(case.tag == 'case' for case in cases), cases
        assert else_ is None or else_.tag == 'else'
        super(Switch, self).__init__(
            'switch',
            token,
            error_token,
            [expr] + cases + [else_],
        )
        self.expr = expr
        self.cases = cases
        self.else_ = else_

    def matches(self, tag, expr, cases, else_):
        return (
            self.tag == tag and
            matches(self.expr, expr) and
            matches(self.cases, cases) and
            matches(self.else_, else_)
        )

    def __repr__(self):
        return 'Switch(%r, %r, %r)' % (self.expr, self.cases, self.else_)


class Tag(AST):
    '''
     - s - the name of the tag (a str).
    '''
    def __init__(self, token, s):
        assert token.tag == 'tag'
        assert type(s) is str
        super(Tag, self).__init__('tag', token, token, [])
        self.s = s

    def matches(self, tag, s):
        return (
            self.tag == tag and
            self.s == s
        )


class Ternary(ExprAST):
    def __init__(self, token, error_token, condition, left, right):
        assert is_expr(condition)
        assert is_expr(left)
        assert is_expr(right)
        super(Ternary, self).__init__(
            'ternary',
            token,
            error_token,
            [condition, left, right],
            is_lvalue=(left.is_lvalue and right.is_lvalue),
        )
        self.condition = condition
        self.left = left
        self.right = right

    def matches(self, tag, is_lvalue, condition, left, right):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.condition, condition) and
            matches(self.left, left) and
            matches(self.right, right)
        )


class TypeArray(ExprAST):
    def __init__(self, token, error_token, constness, type_):
        assert constness is None or constness.tag in CONSTNESS_TAGS
        assert is_expr(type_)
        super(TypeArray, self).__init__(
            'type_array', token, error_token, [constness, type_])
        self.constness = constness
        self.type_ = type_

    def matches(self, tag, is_lvalue, constness, type_):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.constness, constness) and
            matches(self.type_, type_)
        )


class TypeFunction(ExprAST):
    def __init__(self, token, error_token, params, return_type):
        assert type(params) is list
        assert all(p.tag == 'param' for p in params)
        assert is_expr(return_type)
        super(TypeFunction, self).__init__(
            'type_function',
            token,
            error_token,
            params + [return_type],
        )
        self.params = params
        self.return_type = return_type

    def matches(self, tag, is_lvalue, params, return_type):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.params, params) and
            matches(self.return_type, return_type)
        )


class TypeRef(ExprAST):
    def __init__(self, token, error_token, constness, type_):
        assert constness is None or constness.tag in CONSTNESS_TAGS
        assert is_expr(type_)
        super(TypeRef, self).__init__(
            'type_ref', token, error_token, [constness, type_])
        self.constness = constness
        self.type_ = type_

    def matches(self, tag, is_lvalue, constness, type_):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.constness, constness) and
            matches(self.type_, type_)
        )


class TypeStruct(ExprAST):
    def __init__(self, token, error_token, name, params):
        assert name is None or name.tag == 'identifier'
        assert type(params) is list
        assert all(p.tag == 'param' for p in params)
        super(TypeStruct, self).__init__(
            'type_struct', token, error_token, [name] + params)
        self.name = name
        self.params = params

    def matches(self, tag, is_lvalue, name, params):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.name, name) and
            matches(self.params, params)
        )


class TypeUnion(ExprAST):
    def __init__(self, token, error_token, name, cases):
        assert name is None or name.tag == 'identifier'
        assert type(cases) is list
        assert all(c.tag == 'union_case' for c in cases)
        super(TypeUnion, self).__init__(
            'type_union', token, error_token, [name] + cases)
        self.name = name
        self.cases = cases

    def matches(self, tag, is_lvalue, name, cases):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.name, name) and
            matches(self.cases, cases)
        )


class Typeof(ExprAST):
    def __init__(self, token, error_token, expr):
        assert is_expr(expr)
        super(Typeof, self).__init__('typeof', token, error_token, [expr])
        self.expr = expr

    def matches(self, tag, is_lvalue, expr):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.expr, expr)
        )


class Unary(ExprAST):
    def __init__(self, token, error_token, op, expr):
        assert op.tag == 'un_op'
        assert is_expr(expr)
        super(Unary, self).__init__('unary', token, error_token, [op, expr])
        self.op = op
        self.expr = expr

    def matches(self, tag, is_lvalue, op, expr):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.op, op) and
            matches(self.expr, expr)
        )


class Union(ExprAST):
    def __init__(self, token, error_token, name, exprs):
        assert name.tag == 'tag'
        assert type(exprs) is list
        assert all(is_expr(e) for e in exprs)
        super(Union, self).__init__(
            'union', token, error_token, [name] + exprs)
        self.name = name
        self.exprs = exprs

    def matches(self, tag, is_lvalue, name, exprs):
        return (
            self.tag == tag and
            self.is_lvalue == is_lvalue and
            matches(self.name, name) and
            matches(self.exprs, exprs)
        )


class UnionCase(AST):
    def __init__(self, token, error_token, name, types):
        assert name.tag == 'tag'
        assert type(types) is list
        assert all(is_expr(t) for t in types)
        super(UnionCase, self).__init__(
            'union_case', token, error_token, [name] + types)
        self.name = name
        self.types = types

    def matches(self, tag, name, types):
        return (
            self.tag == tag and
            matches(self.name, name) and
            matches(self.types, types)
        )


# TODO: Maybe delete this class now that Unary has an `error_token`?
class UnOp(AST):
    def __init__(self, token, s):
        assert token.tag in ('un_op', 'un_bin_op')
        assert type(s) is str
        super(UnOp, self).__init__('un_op', token, token, [])
        self.s = s

    def matches(self, tag, s):
        return (
            self.tag == tag and
            self.s == s
        )


class While(AST):
    def __init__(self, token, error_token, condition, body, else_):
        assert is_expr(condition)
        assert body.tag == 'block'
        assert else_ is None or else_.tag == 'block'
        super(While, self).__init__(
            'while', token, error_token, [condition, body, else_])
        self.condition = condition
        self.body = body
        self.else_ = else_

    def matches(self, tag, condition, body, else_):
        return (
            self.tag == tag and
            matches(self.condition, condition) and
            matches(self.body, body) and
            matches(self.else_, else_)
        )
