"""
Disambiguates tags based on top-down context, using recursive descent.
Tokens that are illegal in their top-down context are replaced with error
tokens. Removes unnecessary punctuation and keyword tokens (e.g. drops
commas). Finds at least one syntax error, or returns a valid abstract syntax
tree.

The most important tasks done by the AST builder are:

 - Disambiguate subscripts from list literals.

 - Disambiguate struct literals from brackets used to control operator
   precedence.

 - Disambiguate the three uses of ":": explicit type casts; argument labels;
   and ternary operators.

 - Disambiguate defining occurrences of identifiers from identifiers used as
   expressions.

 - Detect some expressions that should be types and obviously aren't.

 - Remove unnecessary nesting levels, e.g. parens.

 - Convert any remaining lower-level tokens into errors.

Disambiguation is achieved by changing the tags, e.g. replacing the
generic "()" tag with "struct" for struct literals.

The desired *output* data structure is that described in
"doc/grammar/source.txt", and is represented using the classes in "ast.py".
The input data structure is simpler and more general, and is represented using
the classes in "token.py".
"""

from .token import (
    Token, Error, NONFIX_TAGS, PREFIX_TAGS, POSTFIX_TAGS, INFIX_TAGS)
from .ast import *


class ASTError(Exception):
    '''
    Raised by `visit_XXX()` methods on finding one or more syntax errors.
    Fields:

     - errors - a non-empty list of Errors.
    '''
    def __init__(self, errors):
        assert type(errors) is list
        assert len(errors) > 0
        assert all(e.tag == 'error' for e in errors)
        self.errors = errors

    def __str__(self):
        return ', '.join(e.message for e in self.errors)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.errors)

    def __eq__(self, other):
        if isinstance(other, ASTError):
            return (
                [e.message for e in self.errors] ==
                [e.message for e in other.errors]
            )
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    __hash__ = None


def raise_token(error):
    raise ASTError([error])


def raise_message(message, token):
    raise_token(Error(message, False, [token]))


def visit_stmt(token):
    assert token.tag != ';'
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'assign':
        # `lvalue` could have e.g. a cast, so it is not just an identifier.
        return Assign(
            token,
            token.op,
            visit_lvalue(token.lhs),
            AssignOp(token.op, token.op.s),
            visit_expr(token.rhs),
        )
    elif token.tag == 'break':
        return Keyword('break', token)
    elif token.tag == 'case':
        raise_message(
            '"case" can only occur inside a "switch" statement', token)
    elif token.tag == 'continue':
        return Keyword('continue', token)
    elif token.tag == 'decl':
        if token.op.s != '=':
            raise_message('Expected "="', token.op)
        # `lvalue` could have e.g. a cast, so it is not just an identifier.
        return Decl(
            token,
            token.op,
            visit_constness(token.decl),
            visit_lvalue(token.lhs),
            visit_expr(token.rhs),
        )
    elif token.tag == 'else':
        raise_message(
            '"else" is only allowed at the end of an "if", "for" or "while" '
            'statement (on the same line) or inside a "switch" statement',
            token.keyword,
        )
    elif token.tag == 'expr':
        # Using `token` as `error_token` is not quite right but it's unlikely
        # to be visible to the user.
        return Expr(token, token, visit_expr(token))
    elif token.tag == 'for':
        return For(
            token,
            token.for_keyword,
            visit_identifier_expr(token.variable),
            visit_expr(token.sequence),
            visit_block(token.body),
            optional(visit_else, token.else_),
        )
    elif token.tag == 'fulfil':
        assert token.op.tag =='assign_op'
        if token.op.s != '=':
            raise_message('Expected "="', token.op)
        return Fulfil(
            token,
            token.keyword,
            visit_expr(token.lhs),
            visit_expr(token.rhs),
        )
    elif token.tag == 'if':
        return visit_stmt_if(token)
    elif token.tag == 'return':
        return Return(
            token,
            token.keyword,
            optional(visit_expr, token.expr),
        )
    elif token.tag == 'switch':
        return visit_stmt_switch(token)
    elif token.tag == 'throw':
        raise NotImplementedError('throw')
    elif token.tag == 'try':
        raise NotImplementedError('try')
    elif token.tag == 'while':
        return While(
            token,
            token.keyword,
            visit_expr(token.condition),
            visit_block(token.body),
            optional(visit_else, token.else_),
        )
    else:
        raise_message('Expected a statement', token)
    assert False


def visit_else(token):
    '''
    Returns a Block.
    '''
    assert token.tag == 'else'
    return visit_block(token.body)


def visit_else_if(token):
    '''
    Returns an If.
    '''
    assert token.tag == 'else_if'
    return visit_stmt_if(token.if_)


def visit_stmt_if(token):
    '''
    Returns an If.
    '''
    assert token.tag == 'if'
    condition_expr = visit_expr(token.condition)
    then_block = visit_block(token.then)
    if token.else_ is None:
        else_block = None
    elif token.else_.tag == 'else':
        else_block = visit_else(token.else_)
    elif token.else_.tag == 'else_if':
        else_block = Block(
            token.else_,
            token.else_.if_.keyword,
            [visit_else_if(token.else_)],
        )
    else:
        assert False
    return If(token, token.keyword, condition_expr, then_block, else_block)


def visit_stmt_switch(token):
    '''
    Returns a Switch.
    '''
    assert token.tag == 'switch'
    condition = visit_expr(token.condition)
    assert token.cases.tag == '{}'
    cases = collect_errors(visit_clause, token.cases.contents)
    # TODO: Rewrite to process the cases in top-to-bottom order.
    # Report an error if we find a 'case' or 'else' after an 'else'.
    if cases and cases[-1].tag == 'else':
        else_ = cases.pop()
    else:
        else_ = None
    for case in cases:
        if case.tag == 'else':
            if else_ is None:
                raise_message(
                    'The "else" clause of a "switch" statement must '
                    'come last, after all the "case"s',
                    case.keyword,
                )
            else:
                raise_message(
                    'Only one "else" clause is allowed in a "switch" '
                    'statement',
                    case.keyword,
                )
    assert all(case.tag == 'case' for case in cases)
    return Switch(token, token.keyword, condition, cases, else_)


def visit_clause(token):
    '''
    Returns a Case or an Else.
    '''
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'case':
        keyword = Keyword('case', token.keyword)
        name, params = visit_pattern(token.pattern)
        return Case(
            token,
            token.keyword,
            keyword,
            name,
            params,
            visit_block(token.body),
        )
    elif token.tag == 'else':
        keyword = Keyword('else', token.keyword)
        return Else(token, token.keyword, keyword, visit_block(token.body))
    else:
        raise_message('Expected "case" or "else"', token)


def visit_pattern(token):
    '''
    Given a `token` that is an 'expr', checks that it is of the form
    `tag(identifier, ...)` and returns `(tag, [identifier, ...])`.
    '''
    if token.tag == 'error':
        raise_token(token)
    assert token.tag == 'expr'
    # TODO: Allow "switch" on integers and others.
    # TODO: Allow arbitrary L-values in the payload, not just identifiers?
    if (
        token.left is not None or
        token.right is not None or
        token.op.tag != 'union'
    ):
        _ = visit_expr(token) # Just in case it raises.
        raise_message('Expected a pattern to match against', token)
    tag = Tag(token.op.name, token.op.name.s)
    ids = visit_payload(visit_identifier_expr, token.op.payload)
    return (tag, ids)


def visit_block(token):
    '''Returns a Block.'''
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == '{}':
        return Block(
            token,
            token.open_bracket,
            collect_errors(visit_stmt, token.contents),
        )
    assert False


def visit_lvalue(token, forbid_ternary=False):
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'expr':
        if token.left is None:
            if token.right is None:
                return visit_lvalue_nonfix(token.op)
            else:
                return visit_lvalue_prefix(token, token.op, token.right)
        else:
            if token.right is None:
                return visit_lvalue_postfix(token, token.left, token.op)
            else:
                return visit_lvalue_infix(
                    token,
                    token.left,
                    token.op,
                    token.right,
                    forbid_ternary,
                )
    else:
        raise_message('Expected an assignable expression', token)
    assert False


def visit_lvalue_nonfix(token):
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'identifier':
        return visit_identifier(token)
    elif token.tag == '()': # Expression in brackets. We remove the brackets.
        is_list, lvalues = visit_expr_list(visit_lvalue, token)
        if is_list:
            return Struct(token, token.open_bracket, lvalues)
        else:
            assert len(lvalues) == 1
            return lvalues[0]
    elif token.tag in NONFIX_TAGS:
        raise_message('The expression is not assignable', token)
    else:
        # This token was inside an 'expr'. It should be right.
        assert False, "Unrecognised nonfix token: %r" % (token,)
    assert False


def visit_lvalue_prefix(lvalue_token, token, right):
    # lvalue_prefix = "*"(expr) / ...
    if token.tag == 'error':
        raise_token(token)
    elif token.tag in ('un_op', 'un_bin_op'):
        if token.s == '*':
            return Dereference(
                lvalue_token,
                token,
                UnOp(token, token.s),
                visit_expr(right),
            )
        else:
            raise_message('The expression is not assignable', token)
    elif token.tag in PREFIX_TAGS:
        raise_message('The expression is not assignable', token)
    else:
        # This token was inside an 'expr'. It should be right.
        assert False, "Unrecognised prefix token: %r" % (token,)
    assert False


def visit_lvalue_postfix(lvalue_token, left, token):
    # lvalue_postfix = "field"(lvalue, identifier) / ...
    # lvalue_postfix = "subscript"(lvalue, expr*) / ...
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'field':
        return Field(
            lvalue_token,
            token.dot,
            visit_lvalue(left),
            visit_identifier(token.name),
        )
    elif token.tag == '[]':
        return Subscript(
            lvalue_token,
            token.open_bracket,
            visit_expr(left),
            visit_only_expr(
                visit_expr,
                token,
                'Expected an array index',
                'Expected "]"',
            ),
        )
    elif token.tag in POSTFIX_TAGS:
        raise_message('The expression is not assignable', token)
    # This token was inside an 'expr'. It should be right.
    assert False, "Unrecognised postfix token: %r" % (token,)


def visit_lvalue_infix(lvalue_token, left, token, right, forbid_ternary):
    if token.tag == 'error':
        # Can't happen? Check it anyway.
        raise_token(token)
    elif token.tag in ('bin_op', 'un_bin_op'):
        if token.s == ':':
            if forbid_ternary:
                raise_message(
                    'Ambiguous ":"; please add round brackets',
                    token,
                )
            return Cast(
                lvalue_token,
                token,
                visit_lvalue(left),
                BinOp(token, token.s),
                visit_expr(right, is_type=True, forbid_ternary=True),
            )
        elif token.s == '?':
            if forbid_ternary:
                raise_message(
                    'Ambiguous "?"; please add round brackets',
                    token,
                )
            return visit_lvalue_ternary(lvalue_token, token, left, right)
        else:
            raise_message('The expression is not assignable', token)
    else:
        assert token.tag not in INFIX_TAGS
        # This token was inside an 'expr'. It should be right.
        assert False, "Unrecognised infix token: %r" % (token,)
    assert False


def visit_lvalue_ternary(lvalue_token, error_token, condition, token):
    '''
    Returns the AST of a ternary expression 'condition ? if_true : if_false`.

     - lvalue_token - the Token that represents the entire expression.
     - error_token - the Token at which to report errors.
     - condition - the Token that represents the `condition` expression.
     - token - the Token that should be an Expr representing the
       `if_true : if_false` sub-expression.
    '''
    if token.tag == 'error':
        raise_token(token)
    assert token.tag == 'expr'
    if token.left is None or token.right is None or token.op.tag != 'bin_op':
        raise_message('Expected ":" after "?"', token)
    if token.op.s == '?':
        raise_message('Confusing "?"; please add round brackets', token.op)
    if token.op.s != ':':
        raise_message('Expected ":" after "?"', token.op)
    return Ternary(
        lvalue_token,
        error_token,
        visit_expr(condition),
        visit_lvalue(token.left),
        visit_lvalue(token.right, forbid_ternary=True),
    )


def visit_type(token):
    return visit_expr(token, is_type=True)


def visit_expr(token, is_type=False, forbid_ternary=False):
    '''
    Returns the AST of an 'expr' token.

     - token - the token that should be an 'expr'.
     - is_type - `True` to enable errors for expressions that are not types.
       This is used, for example, on the right-hand-side of a cast.
     - forbid_ternary - `True` to forbid expressions whose outermost operator
       is ':' or '?'. This is used, for example, to forbid ambiguous
       expressions such as "x:y:z" and "a?b:c?d:e".
    '''
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'expr':
        if token.left is None:
            if token.right is None:
                return visit_expr_nonfix(token.op, is_type)
            else:
                return visit_expr_prefix(
                    token,
                    token.op,
                    token.right,
                    is_type,
                )
        else:
            if token.right is None:
                return visit_expr_postfix(
                    token,
                    token.left,
                    token.op,
                    is_type,
                )
            else:
                return visit_expr_infix(
                    token,
                    token.left,
                    token.op,
                    token.right,
                    is_type,
                    forbid_ternary,
                )
    else:
        raise_message('Expected an expression', token)
    assert False


def visit_expr_nonfix(token, is_type):
    def raise_not_a_type():
        raise_message('The expression is not a type', token)
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'character':
        if is_type:
            raise_not_a_type()
        return Character(token, token.c)
    elif token.tag == 'identifier':
        return visit_identifier(token)
    elif token.tag == 'integer':
        if is_type:
            raise_not_a_type()
        return Integer(token, token.s)
    elif token.tag == 'hex_integer':
        if is_type:
            raise_not_a_type()
        return HexInteger(token, token.s)
    elif token.tag == 'float':
        if is_type:
            raise_not_a_type()
        return Float(token, token.s)
    elif token.tag == 'string':
        if is_type:
            raise_not_a_type()
        for part in token.parts:
            if isinstance(part, Token) and part.tag == 'error':
                raise_token(part)
        return String(token, ''.join(token.parts))
    elif token.tag == '()': # Expression in brackets. We remove the brackets.
        is_list, exprs = visit_expr_list(
            visit_type if is_type else visit_expr, # TODO: Abbreviate.
            token
        )
        if is_list:
            if is_type:
                raise_not_a_type()
            return Struct(token, token.open_bracket, exprs)
        else:
            assert len(exprs) == 1
            return exprs[0]
    elif token.tag == '[]': # Array literal.
        if is_type:
            raise_not_a_type()
        exprs = visit_expr_list(visit_expr, token)[1]
        return Array(token, token.open_bracket, exprs)
    elif token.tag == 'function':
        assert token.params.tag == '()'
        if token.body is None:
            # It's a function type.
            if token.name is not None:
                # Naming a function type could be useful for recursive
                # function types, but is unnecessary and would be confusing.
                raise_message(
                    'The function type has a name; please add a body if you '
                    'mean to define a function',
                    token,
                )
            return TypeFunction(
                token,
                token.keyword,
                visit_expr_list(visit_param_expr, token.params)[1],
                visit_return_type(token.return_type),
            )
        else:
            # It's a function definition.
            return FunctionDef(
                token,
                token.keyword,
                optional(visit_identifier, token.name),
                visit_expr_list(visit_named_param_expr, token.params)[1],
                optional(visit_return_type, token.return_type),
                visit_block(token.body),
            )
    elif token.tag == 'type_struct':
        assert token.params is not None
        return TypeStruct(
            token,
            token.keyword,
            optional(visit_identifier, token.name),
            visit_expr_list(visit_param_expr, token.params)[1],
        )
    elif token.tag == 'type_union':
        if token.cases.tag == 'error':
            # Can't happen? Check it anyway.
            raise_token(token.cases)
        assert token.cases.tag == '{}'
        return TypeUnion(
            token,
            token.keyword,
            optional(visit_identifier, token.name),
            collect_errors(visit_union_case, token.cases.contents),
        )
    elif token.tag == 'union':
        assert token.name.tag == 'tag'
        tag = Tag(token.name, token.name.s)
        ids = visit_payload(visit_expr, token.payload)
        return Union(token, token.name, tag, ids)
    else:
        assert token.tag not in NONFIX_TAGS
        # This token was inside an 'expr'. It should be right.
        assert False, "Unrecognised nonfix token: %r" % (token,)
    assert False


def visit_expr_prefix(expr_token, token, right, is_type):
    def raise_not_a_type():
        raise_message('The expression is not a type', token)
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'array':
        return TypeArray(
            expr_token,
            token,
            optional(visit_constness, token.constness),
            visit_expr(right, is_type=True),
        )
    elif token.tag == 'box':
        return Box(expr_token, token, visit_expr(right))
    elif token.tag == 'defer':
        return Defer(expr_token, token, visit_expr(right, is_type=True))
    elif token.tag == 'new':
        return New(expr_token, token, visit_expr(right))
    elif token.tag == 'ref':
        return TypeRef(
            expr_token,
            token,
            optional(visit_constness, token.constness),
            visit_expr(right, is_type=True),
        )
    elif token.tag == 'typeof':
        return Typeof(expr_token, token, visit_expr(right))
    elif token.tag in ('un_op', 'un_bin_op'):
        if token.s == '*':
            return Dereference(
                expr_token,
                token,
                UnOp(token, token.s),
                visit_expr(right),
            )
        elif is_type:
            # Other prefix operators can't return a type.
            raise_not_a_type()
        elif token.s == '&':
            return AddressOf(
                expr_token,
                token,
                UnOp(token, token.s),
                visit_expr(right),
            )
        else:
            return Unary(
                expr_token,
                token,
                UnOp(token, token.s),
                visit_expr(right),
            )
    else:
        assert token.tag not in PREFIX_TAGS
        # This token was inside an 'expr'. It should be right.
        assert False, "Unrecognised prefix token: %r" % (token,)
    assert False


def visit_expr_postfix(expr_token, left, token, is_type):
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'field':
        return Field(
            expr_token,
            token.dot,
            visit_expr(left),
            visit_identifier(token.name),
        )
    elif token.tag == '()':
        return Call(
            expr_token,
            token.open_bracket,
            visit_expr(left),
            visit_expr_list(visit_expr, token)[1],
        )
    elif token.tag == '[]':
        return Subscript(
            expr_token,
            token.open_bracket,
            visit_expr(left),
            visit_only_expr(
                visit_expr,
                token,
                'Expected an array index',
                'Expected "]"',
            ),
        )
    else:
        assert token.tag not in POSTFIX_TAGS
        # This token was inside an 'expr'. It should be right.
        assert False, "Unrecognised postfix token: %r" % (token,)
    assert False


def visit_expr_infix(expr_token, left, token, right, is_type, forbid_ternary):
    if token.tag == 'error':
        raise_token(token)
    elif token.tag in ('bin_op', 'un_bin_op'):
        if token.s == ':':
            if forbid_ternary:
                raise_message(
                    'Ambiguous ":"; please add round brackets',
                    token,
                )
            if is_type:
                raise_message(
                    'Operator ":" is forbidden where a type is expected',
                    token,
                )
            return Cast(
                expr_token,
                token,
                visit_expr(left, is_type=is_type),
                BinOp(token, token.s),
                visit_expr(right, is_type=True, forbid_ternary=True),
            )
        elif token.s == '?':
            if forbid_ternary:
                raise_message(
                    'Ambiguous "?"; please add round brackets',
                    token,
                )
            return visit_expr_ternary(expr_token, token, left, right, is_type)
        else:
            if is_type:
                # There are no other infix operators that return types.
                raise_message('The expression is not a type', token)
            return Binary(
                expr_token,
                token,
                visit_expr(left),
                BinOp(token, token.s),
                visit_expr(right),
            )
    else:
        assert token.tag not in INFIX_TAGS
        # This token was inside an 'expr'. It should be right.
        assert False, "Unrecognised infix token: %r" % (token,)
    assert False


def visit_expr_ternary(expr_token, error_token, condition, token, is_type):
    '''
    Returns the AST of a ternary expression 'condition ? if_true : if_false`.

     - expr_token - the Token that represents the entire expression.
     - error_token - the Token at which to report errors.
     - condition - the Token that represents the `condition` expression.
     - token - the Token that should be an Expr representing the
       `if_true : if_false` sub-expression.
     - is_type - True to enable errors for expressions that are not types.
    '''
    if token.tag == 'error':
        raise_token(token)
    assert token.tag == 'expr'
    if token.left is None or token.right is None or token.op.tag != 'bin_op':
        raise_message('Expected ":" after "?"', token)
    if token.op.s == '?':
        raise_message('Confusing "?"; please add round brackets', token.op)
    if token.op.s != ':':
        raise_message('Expected ":" after "?"', token.op)
    return Ternary(
        expr_token,
        error_token,
        visit_expr(condition),
        visit_expr(token.left, is_type=is_type),
        visit_expr(token.right, is_type=is_type, forbid_ternary=True),
    )


def visit_named_param_expr(token):
    return visit_param_expr(token, needs_name=True)


def visit_param_expr(token, needs_name=False):
    '''
    Deconstructs an 'expr' token that should be a parameter declaration.
    Returns `('param', token, identifier, type_expr)`. `identifier` is
    optional.
    '''
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'expr':
        if token.op.tag == 'bin_op' and token.op.s == ':':
            return Param(
                token,
                token.op,
                visit_identifier_expr(token.left),
                visit_type(token.right),
            )
        else:
            type_ = visit_type(token)
            if needs_name:
                raise_message(
                    'Function parameters must use the syntax "<name>: <type>"',
                    token,
                )
            return Param(token, token, None, type_)
    else:
        assert False, token


def visit_return_type(token):
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'identifier':
        return visit_identifier(token)
    elif token.tag == '()':
        # Expression in brackets. We remove the brackets.
        return visit_only_expr(
            visit_type,
            token,
            'Expected an expression inside "( ... )"',
            'Expected ")"',
        )
    else:
        assert False, token


def visit_union_case(token):
    '''
    Deconstructs a statement token that should be an 'expr' token that is a
    case of a union, i.e. of the form "<tag>(<expr>, ...)". Returns a
    UnionCase.
    '''
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'expr':
        if (
            token.left is None and
            token.right is None and
            token.op.tag == 'union'
        ):
            tag = Tag(token.op.name, token.op.name.s)
            ids = visit_payload(visit_type, token.op.payload)
            return UnionCase(token, token.op.name, tag, ids)
        else:
            _ = visit_expr(token) # Just in case it raises.
    raise_message('Expected a case declaration, e.g. "TAG(arg, ...)"', token)


def visit_identifier_expr(token):
    '''
    Checks that `token` is an 'expr' containing an 'identifier' containing
    a name, and returns an Identifier.
    '''
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'expr':
        if token.op.tag == 'error':
            # Can't happen? Check for it anyway.
            raise_token(token.op)
        elif token.op.tag == 'identifier':
            return visit_identifier(token.op)
        else:
            # Tempting to say 'Expected identifier'. However, this would be
            # confusing if the leftmost token is an identifier, as in e.g.
            # "x == 1" in "{x == 1: 1}". To avoid confusion, we first visit
            # `token` as an 'expr', and propagate any error found. If there is
            # no error, then we have enough context to write a sensible error
            # message.
            _ = visit_expr(token) # Just in case it raises.
            raise_message('The expression is not an identifier', token)
    else:
        raise_message('Expected an identifier', token)


def visit_identifier(token):
    '''
    Returns `None` if `token` is `None`, otherwise asserts that `token` is an
    'identifier', and returns an Identifier.
    '''
    assert token.tag == 'identifier'
    return Identifier(token, token.s)


def visit_constness(token):
    if token.tag == 'error':
        raise_token(token)
    elif token.tag == 'const':
        return Keyword('const', token)
    elif token.tag == 'var':
        return Keyword('var', token)
    else:
        assert False


def optional(visit, token):
    '''
    Utility function.
    Visits `token` unless it is None, in which case preserves it.
    '''
    if token is None:
        return token
    else:
        return visit(token)


def collect_errors(visit, tokens):
    '''
    Applies `visit()` to each of `tokens`. If none raises an ASTError,
    returns the list of results, otherwise collects all the errors and raises
    an ASTError containing them all.
    '''
    result = []
    errors = []
    for token in tokens:
        try:
            result.append(visit(token))
        except ASTError as e:
            errors.extend(e.errors)
    if errors:
        raise ASTError(errors)
    else:
        return result


def visit_payload(visit_item, token):
    '''
    If token is a token of type '()', visits the items it contains with
    `visit_item()` and returns a list of the results. Otherwise, token must be
    `None` and an empty list is returned.
    '''
    if token is None:
        return []
    return visit_expr_list(visit_item, token)[1]


def visit_expr_list(visit_item, token):
    '''
    Given a token of type '()' or '[]', returns a pair `(is_list, asts)`
    where:

     - `is_list` - `True` if `token.contents` is empty, or if `token.contents`
       is an 'expr_list' with at least one comma. `False` otherwise, i.e. if
       `token.contents` is non-empty and contains no comma.
     - `asts` - a list of the results of applying `visit_item` to each 'expr'
       Token in the brackets. In the `!is_list` case, the length of the list
       will be exactly one.
    '''
    assert token.tag in ('()', '[]'), token
    if len(token.contents) == 0:
        return (True, [])
    if token.contents[0].tag == 'error':
        raise_token(token.contents[0])
    assert len(token.contents) == 1, token
    (contents,) = token.contents
    assert contents.tag == 'expr_list', contents
    is_list = len(contents.commas) > 0
    exprs = contents.exprs
    return (is_list, [visit_item(e) for e in exprs])


def visit_only_expr(visit_item, token, message1, message2):
    '''
    Applies `visit_item()` to the unique 'expr' in `token`, and returns the
    result.

     - message1 - the error message to use if `token` is `None` or empty, e.g.
       'Expected an expression'.
     - message2 - the error message to use if `token` has a comma after the
       first expression, e.g. 'Expected ")"'.
    '''
    is_list, exprs = visit_expr_list(visit_item, token)
    if is_list:
        if len(exprs) == 0:
            raise_message(message1, token)
        else:
            raise_message(message2, token.contents[0].commas[0])
    else:
        assert len(exprs) == 1
        return exprs[0]
