"""
Defines a Grammar which recognises statements separated by semicolons or
newlines. Extra semicolons and newlines are permitted.

Statement tokens are preserved. Error tokens are preserved, except that
consequential errors are suppressed up to the next semicolon or newline.
Semicolons and newlines are removed. All other tags are replaced with
errors.
"""

from .token import Error
from .grammar import Grammar
from .util import discard, pass_on, pass_on_error, error, Unfinished
from .lexer import high_level_lexer_tags, keywords
from .brackets import bracket_tags
from .atom import atom_tags
from .stmt import stmt_tags


__all__ = ['block', 'block_tags']

# The set of tags this parser can output. It can also output 'error'.

block_tags = stmt_tags

# The set of all tags other than statements, semicolons, newlines and
# errors.

bad = set()
bad.update(high_level_lexer_tags)
bad.update(keywords)
bad.update(bracket_tags)
bad.update(atom_tags)
bad.difference_update(stmt_tags)
bad.difference_update([';', 'newline'])


# After a statement, errors and non-statements are replaced with a different
# error, which is eagerly emitted.

block_end_error = Grammar(
    pass_on_error('Expected ";" or a newline'), is_eager=True)

# Unseparated statements are collected and turned into an error.

block_unseparated = Grammar(
    error('Expected ";" or a newline'), name='block_unseparated')
block_unseparated.set_all(
    stmt_tags, block_unseparated, stream_name='discard')

# Statements must be followed by semicolon, newline or EOF.
# This is difficult because we can't match EOF.
# Instead, we make sure to handle everything else.

block_end = Grammar(pass_on, name='block_end')
block_end.set('error', block_end_error, stream_name='error')
block_end.set_all(bad, block_end_error, stream_name='discard')
block_end.set_all(stmt_tags, block_unseparated, stream_name='discard')

# Errors are eagerly passed on.
# Non-statements are turned into an error and eagerly passed on.

# TODO: Factor out commonality with `expr_list.bad_or_error()`.
def bad_or_error(multi_buffer):
    '''
    If there is a token in stream 'error', return it.
    Otherwise wrap up the contents of the default stream.
    '''
    ans = multi_buffer.get_optional('error')
    if ans is None:
        ans = Error(
            'Expected a statement',
            False,
            multi_buffer.get()
        )
    return ans

block_error = Grammar(bad_or_error, is_eager=True)

# The initial state of the block grammar.

block = Grammar(discard, name='block')
block.ignore_all([';', 'newline'])
block.set('error', block_error, stream_name='error')
block.set_all(bad, block_error)
block.set_all(stmt_tags, block_end)
