"""
Defines a Grammar which recognises expressions.

The post-process method resolves operator precedence and rewrites the
parse-tree accordingly.
"""

from .token import Concat, Expr
from .grammar import Grammar
from .util import discard, pass_on, Unfinished
from .lexer import high_level_lexer_tags, keywords, bin_ops
from .brackets import bracket_tags
from .atom import atom_tags
from .unary_expr import unary_expr_tags


__all__ = ['expr', 'expr_tags', 'parse_exprs', 'to_expr']

# The set of tags this parser can output, in addition to 'error' and
# those passed on from lower levels.

expr_tags = {'expr'}

# The initial state of the expression grammar.

expr = Grammar(discard, name='expr')

# Unless overridden later, all tags are passed on unchanged.

expr.set_all(high_level_lexer_tags, Grammar(pass_on))
expr.set_all(keywords, Grammar(pass_on))
expr.set_all(bracket_tags, Grammar(pass_on))
expr.set_all(atom_tags, Grammar(pass_on))
expr.set_all(unary_expr_tags, Grammar(pass_on))

# Errors, newlines and semicolons are passed on eagerly.

expr.set_all(['error', 'newline', ';'], Grammar(pass_on, is_eager=True))


# Keys are infix operators. Values are (left binding strength, right
# binding strength). To avoid ambiguity of binding, no strength is used both as
# a left strength and a right strength.
# No need to define the precedence of prefix and postfix operators, as they are
# all the same: 100 and 101 respectively.
strengths = {}
for i, (associativity, ops) in enumerate([
    ('right', ['?', ':']),
    ('left', ['||']),
    ('left', ['&&']),
    ('left', ['|']),
    ('left', ['^']),
    ('left', ['&']),
    ('left', ['==', '!=']),
    ('left', ['<', '>', '<=', '>=', '<>']),
    ('right', ['<<', '>>', '>>>']),
    ('left', ['+', '-']),
    ('left', ['*', '/', '%']),
    ('right', ['**']),
]):
    left_strength = 2*i + int(associativity == 'right')
    right_strength = 2*i + int(associativity == 'left')
    for op in ops:
        strengths[op] = (left_strength, right_strength)

assert set(strengths) == bin_ops | {'?', ':'}


def resolve_precedence(multi_buffer):
    stack = [] # List of (expr, op, right_strength).

    def flush(e, left_strength):
        '''
        We have seen <stack> <e> <op>. The left binding strength of 'op' is
        `left_strength`. Any operators on the stack that bind
        more tightly than <op> must be popped and applied to `e`. Modifies
        `stack` in place, and returns the new `e`.
        '''
        while stack and stack[-1][2] > left_strength:
            left, op, _ = stack.pop()
            e = Expr(left, op, e)
        return e

    for ue in multi_buffer.get():
        for prefix in ue.prefixes:
            # All prefix operators bind with right_strength `100`.
            stack.append((None, prefix, 100))
        e = Expr(None, ue.nonfix, None) # Nonfix, i.e. a leaf expression.
        for postfix in ue.postfixes:
            # All postfix operators bind with left_strength `101`.
            e = Expr(e, postfix, None)
        if ue.infix is None:
            return flush(e, -1)
        else:
            assert type(ue.infix) is Concat
            left_strength, right_strength = strengths.get(ue.infix.s)
            e = flush(e, left_strength)
            stack.append((e, ue.infix, right_strength))
    assert False, "Should not happen"

expr_end = Grammar(resolve_precedence, name='expr_end')
expr_unary_expr = Unfinished("Expected an expression")
expr_unary_expr.ignore('newline')
expr_unary_expr.set('unary_expr', expr_end)
expr_unary_expr.set('unary_expr_infix', expr_unary_expr)
expr.set('unary_expr', expr_end)
expr.set('unary_expr_infix', expr_unary_expr)
