"""
Defines a Grammar which recognises comma-separated lists of expressions.
This Grammar is only used to parse the contents of round and square brackets.

Trailing commas are permitted. Newlines are discarded. All other tags are
replaced with an error. The entire input stream (if non empty) is always
reduced to a single 'expr_list' or 'error' token.

Errors received from lower levels of the parser are passed on unless this
parser level detects an error first, in which case the earlier error takes
precedence.
"""

from .token import Error, ExprList
from .grammar import Grammar
from .util import discard, pass_on, pass_on_error
from .lexer import high_level_lexer_tags, keywords
from .brackets import bracket_tags
from .atom import atom_tags


# The set of tags this parser can output. It can also output 'error'.

expr_list_tags = ['expr_list']


# The set of all tags other than expressions, commas, newlines and
# errors.

bad = set()
bad.update(high_level_lexer_tags)
bad.update(keywords)
bad.update(bracket_tags)
bad.update(atom_tags)
bad.difference_update(['expr', ',', 'newline'])


# A post-process method that returns an 'expr_list' token.

def make_expr_list(multi_buffer):
    '''
    Returns an ExprList.
    '''
    return ExprList(
        multi_buffer.get(stream_name='exprs'),
        multi_buffer.get(stream_name='commas'),
    )


# Every grammar state defines transitions for every possible input token. To
# exhibit the exhaustivity, we always group the transitions in this order:
# 'newline', 'expr', ',', 'error', bad.

# If we're expecting a comma but we get something else, everything to
# EOF is collapsed into a single error token. The error is 'Expected "," or a
# closing bracket', unless we started with an error, in which case we use that
# error. We already have something in the buffer 'exprs' (and maybe also in
# 'commas'), so we report the error after it.

expr_list_comma_error = Grammar(
    pass_on_error('Expected "," or a closing bracket'),
    name='expr_list_comma_error',
)
expr_list_comma_error.ignore_all(['newline', 'expr', ',', 'error'])
expr_list_comma_error.ignore_all(bad)

# If we're expecting an expression but we get something else, everything to
# EOF is collapsed into a single error token. The error is 'Expected an
# expression or a closing bracket', unless we started with an error, in which
# case we use that error. If we don't already have an error, then we already
# have a bad token in the default buffer, so we report an error at that token.

def bad_or_error(multi_buffer):
    '''
    If there is a token in stream 'error', return it.
    Otherwise wrap up the contents of the default stream.
    '''
    ans = multi_buffer.get_optional('error')
    if ans is None:
        ans = Error(
            'Expected an expression or a closing bracket',
            False,
            multi_buffer.get()
        )
    return ans

expr_list_expr_error = Grammar(bad_or_error, name='expr_list_expr_error')
expr_list_expr_error.ignore_all(['newline', 'expr', ',', 'error'])
expr_list_expr_error.ignore_all(bad)

# A list is an alternating sequence of expressions and commas.

expr_list_comma = Grammar(make_expr_list, name='expr_list_comma')
expr_list_expr = Grammar(make_expr_list, name='expr_list_expr')

expr_list_comma.ignore('newline')
expr_list_comma.set('expr', expr_list_comma_error, stream_name='discard')
expr_list_comma.set(',', expr_list_expr, stream_name='commas')
expr_list_comma.set('error', expr_list_comma_error, stream_name='error')
expr_list_comma.set_all(bad, expr_list_comma_error, stream_name='discard')

expr_list_expr.ignore('newline')
expr_list_expr.set('expr', expr_list_comma, stream_name='exprs')
expr_list_expr.set(',', expr_list_expr_error)
expr_list_expr.set('error', expr_list_expr_error, stream_name='error')
expr_list_expr.set_all(bad, expr_list_expr_error)

# The initial state of the list grammar.

expr_list = Grammar(discard, name='expr_list')
expr_list.ignore('newline')
expr_list.set('expr', expr_list_comma, stream_name='exprs')
expr_list.set(',', expr_list_expr_error)
expr_list.set('error', expr_list_expr_error, stream_name='error')
expr_list.set_all(bad, expr_list_expr_error)
