"""A library for generating nested LL(1) parsers and lexers."""


from .token import Token


__all__ = ['MultiBuffer', 'Grammar', 'Listener', 'Buffer', 'Parser']


class MultiBuffer(object):
    '''
    Represents a collection of named buffers. The `append()` method appends a
    Token to a buffer, and the `get()` method returns the contents of a buffer.
    '''
    def __init__(self):
        self.buffers = {}

    def append(self, token, stream_name='default'):
        '''Appends `token` to the buffer named `stream_name`
           unless `stream_name` is 'discard'.'''
        if stream_name != 'discard':
            if stream_name not in self.buffers:
                self.buffers[stream_name] = []
            self.buffers[stream_name].append(token)

    def get(self, stream_name='default'):
        '''Returns the contents of the buffer named `stream_name`.'''
        return self.buffers.get(stream_name, [])[:]

    def get_all(self):
        '''
        Returns the Tokens in all the streams except 'discard'.
        '''
        ans = []
        for stream_name, tokens in self.buffers.items():
            if stream_name != 'discard':
                ans.extend(tokens)
        return ans

    def get_only(self, stream_name='default'):
        '''Returns the sole Token in the buffer named `stream_name`.'''
        buf = self.get(stream_name=stream_name)
        assert len(buf) == 1, (
            "Buffer '%s' does not contain exactly one Token: %r" %
            (stream_name, buf)
        )
        return buf[0]

    def get_optional(self, stream_name='default'):
        '''
        Returns the sole Token in the buffer named `stream_name` or `None` if
        that buffer is empty.
        '''
        buf = self.get(stream_name=stream_name)
        assert len(buf) <= 1, (
            "Buffer '%s' contains more than one Token" % stream_name
        )
        if buf:
            return buf[0]
        else:
            return None

    def is_empty(self):
        """Returns `True` if all buffers are empty."""
        return self.buffers == {}

    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            ", ".join('%s=%r' % (k, v) for k, v in self.buffers.iteritems())
        )


class Grammar(object):
    """
    Represents a state of a deterministic finite-state machine that can
    be used to parse streams of Tokens. Provides methods to help
    construct the machine.

    Instances of Grammar can be given a `post_process` function,
    which can mangle the parse tree in any way that is desired. The
    `post_process` function is also the recommended way to report
    parse errors, and possibly to recover a partial parse tree.

    The input and output of the parser is a stream of Tokens. Matching is done
    entirely on the `tag` of the Token, which is a str.
    """
    counter = 0 # debug aid

    def __init__(self, post_process, name=None, is_eager=False):
        """
        Constructs a Grammar. The Grammar initially accepts only the empty
        string.

        `post_process` - An arbitrary function that should be called when no
            further input can be parsed by this Grammar. The Tokens received
            in order to reach this Grammar are passed to the function as
            a MultiBuffer. The function should return a Token representing the
            parse tree, or None to indicate that the input should be discarded.

        `is_eager` - `True` to declare that this state has no transitions.
            This allows the Parser to call the `post_process` function as soon
            as it reaches this state, instead of waiting for it to reject the
            next Token.

        Note that the `post_process` function is responsible for handling
        parse errors properly. If e.g. the input terminates early, then
        whichever Grammar has been reached will be asked to process the
        input received so far.
        """
        assert post_process is not None
        self.post_process = post_process
        self.is_eager = is_eager
        Grammar.counter += 1
        if name is None:
            name = 'Grammar_%d' % Grammar.counter
        self.name = name
        self.number = Grammar.counter
        self.transitions = {} # Token type -> (Grammar, stream name).

    def clone(self, name=None):
        """
        Returns a fresh copy of this Grammar. It is a shallow copy; the
        Grammars reachable via `self.transitions` are not copied.

        This method can be used to obtain a Grammar that is unreachable.
        This is sometimes desirable prior to making modifications.
        """
        ans = Grammar(
            post_process=self.post_process,
            name=name if name else self.name,
            is_eager=self.is_eager
        )
        ans.transitions.update(self.transitions)
        return ans

    def accepts(self, tag):
        """
        Returns `True` if this Grammar state has a transition defined for
        `tag` or `False` otherwise.
        """
        return tag in self.transitions

    def set(self, tag, grammar, stream_name='default'):
        """
        Modifies this Grammar so that on receiving a Token of type `tag` it
        will go to state `grammar`.

        Does not copy this Grammar. Therefore, if there are multiple
        transitions reaching this Grammar, all of them will be affected.
        See `clone()` and `get_copy()` for ways to avoid this.

        `tag` - determines which transition is being written.
        `grammar` - the new destination for the transition.
        `stream_name` - the name of the stream that a Token of type `tag`
            should be fed to by a Parser in this state.
        """
        if self.is_eager:
            raise ValueError(
                'Eager state %r cannot have outgoing transitions' % self)
        self.transitions[tag] = (grammar, stream_name)

    def set_all(self, tag_list, grammar, stream_name='default'):
        """
        Utility method equivalent to::

            for tag in tag_list:
                self.set(tag, grammar, stream_name)
        """
        for tag in tag_list:
            self.set(tag, grammar, stream_name)

    def ignore(self, tag):
        """
        Utility method equivalent to:

            self.set(tag, self, stream_name='discard')
        """
        self.set(tag, self, stream_name='discard')

    def ignore_all(self, tag_list):
        """
        Utility method equivalent to::

            for tag in tag_list:
                self.ignore(tag)
        """
        for tag in tag_list:
            self.ignore(tag)

    def get_copy(self, tag, name=None):
        """
        Looks up the Grammar that would be reached from this Grammar on
        receiving a Token of type `tag`. Makes a fresh copy of it, and
        then modifies this Grammar so that it would go to the copy instead.
        Returns the copy.

        This method can be used to obtain a Grammar that is reachable in only
        one way. This is sometimes desirable before making modifications.
        """
        state, stream_name = self.transitions[tag]
        ans = state.clone(name=name)
        self.set(tag, ans, stream_name=stream_name)
        return ans

    def __repr__(self):
        if self.name is None:
            return '%s()' % (self.__class__.__name__,)
        else:
            return '%s(%r)' % (self.__class__.__name__, self.name)


class Listener(object):
    """
    Superclass of objects that receive a stream of Tokens.
    """

    def feed(self, token):
        """
        Feeds a Token to this Listener. The default implementation
        merely prints the Token.
        """
        print(token)

    def flush(self):
        """
        Forces this Listener to process all Tokens it has received so far,
        and prevents it from combining them with subsequent Tokens. This
        method is idempotent. The default implementation does nothing.
        """
        pass

    def is_flushed(self):
        """
        Returns `True` if calling `flush()` would be a no-op.
        """
        return True


class Buffer(Listener):
    """
    A Listener which records all Tokens it receives in a list. This is useful
    for many purposes, including debugging.

    Fields:

    contents - a list containing the Tokens fed to this Buffer.
    """
    def __init__(self):
        self.contents = []

    def feed(self, token):
        self.contents.append(token)


class Parser(Listener):
    """
    Represents the state of an algorithm that parses a stream of input
    Tokens and produces a stream of higher level Tokens. The output
    Tokens are fed to a Listener, which can either process them further
    or store them.

    To parse a complete sentence, first construct a Parser,
    then repeatedly feed it input Tokens, as follows::

        p = Parser(myGrammar, myListener)
        for t in tokens:
            p = p.feed(t)

    Parsers are themselves Listeners, so it is possible to chain
    Parsers together, such that the output of one Parser becomes the
    input stream for another. This facility is useful for parsing complex
    Grammars.
    """
    def __init__(self, grammar, listener=None):
        """
        Constructs a Parser for `grammar`. When no more input can be
        parsed, the input read so far will be collapsed to a single Token and
        passed to `listener`. The Parser will then be reset to its initial
        state, and parsing will continue.

        `grammar` - An instance of Grammar representing the type of
            sentence that is expected, and the processing that should be
            applied to it. After each output Token, the Parser returns to
            state `grammar`.
        `listener` - The Listener to which to stream the output of this Parser.
            If omitted, a default Listener is constructed which prints its
            output to sys.stdout.
        """
        self.home = grammar
        self.listener = listener or Listener()
        self.grammar = self.home
        self.multi_buffer = MultiBuffer()

    def feed(self, token):
        """
        Feeds an input Token to this Parser.

        `token` - the Token. The Token typically represents a partial parse
            tree, and its `tag` field is matched to determine what to do next.
        """
        assert isinstance(token, Token), token
#        print("feed(%r) to %r" % (token, self.grammar))
        # Look up the transition for this Token type.
        transition = self.grammar.transitions.get(token.tag)
        if transition is None:
            self._reduce()
            assert self.grammar is self.home
#            print("feed(%r) to %r instead" % (token, self.grammar))
            transition = self.grammar.transitions.get(token.tag)
            assert transition, (
                "No transition defined for Token type '%s' in state '%s'" %
                (token.tag, self.grammar.name)
            )
        # Accept the Token.
        self.grammar, stream_name = transition
        self.multi_buffer.append(token, stream_name)
#        print("    ...arrived at %r" % self.grammar)
        if self.grammar.is_eager:
            self._reduce()

    def flush(self):
        """
        Forces the Parser to end the output Token at this point, and to
        output any buffered data. Recursively calls `listener.flush()`.
        """
        self._reduce()
        self.listener.flush()

    def is_flushed(self):
        return (
            self.grammar is self.home and
            self.multi_buffer.is_empty() and
            self.listener.is_flushed()
        )

    def _reduce(self):
        """
        Calls the `post_process` method for the current Grammar state passing
        the contents of `self.buffer`. If the result is not None, feeds it to
        `self.listener`. Finally, switches to Grammar state `self.home`.
        """
#        print("%r.post_process(%r)" % (self.grammar, self.multi_buffer))
        token = self.grammar.post_process(self.multi_buffer)
        if token is not None:
            self.listener.feed(token)
        self.grammar = self.home
        self.multi_buffer = MultiBuffer()
