import pprint

from .ast import AST, matches
from .token import ASCII
from . import grammar, lexer, brackets, atom, unary_expr, expr, expr_list, \
    stmt, block, astbuilder

from .astbuilder import ASTError


def validate(source):
    '''Parses and validates `source` (a string) and returns the concrete
       parse tree (a list of ASTs).'''
    assert type(source) is str, type(source)
    def make_stmt_parser(p):
        p = grammar.Parser(block.block, p)
        p = grammar.Parser(stmt.stmt, p)
        p = grammar.Parser(expr.expr, p)
        p = grammar.Parser(unary_expr.unary_expr, p)
        p = grammar.Parser(atom.atom, p)
        return p
    def make_expr_parser(p):
        p = grammar.Parser(expr_list.expr_list, p)
        p = grammar.Parser(expr.expr, p)
        p = grammar.Parser(unary_expr.unary_expr, p)
        p = grammar.Parser(atom.atom, p)
        return p
    listener = grammar.Buffer()
    p = make_stmt_parser(listener)
    p = brackets.BracketMatcher(p, {
        '(': make_expr_parser,
        '[': make_expr_parser,
        '{': make_stmt_parser,
    })
    p = grammar.Parser(lexer.high_level_lexer, p)
    p = grammar.Parser(lexer.low_level_lexer, p)
    for i, c in enumerate(source):
        p.feed(ASCII(i, c))
    p.flush()
    return astbuilder.collect_errors(astbuilder.visit_stmt, listener.contents)


def check(source, pattern, is_safe_to_remove_newlines=True):
    '''
    Parses `source` (a string) and verifies that it matches `pattern`.
    The test is run thrice: first as-is, and then again after replacing
    any newlines with spaces, then again after doubling the newlines. Tests
    should be written so that this does not affect the expected result.
    '''
    replacements = ['\n', '\n\n']
    if is_safe_to_remove_newlines:
        replacements.append(' ')
    for replacement in replacements:
        source2 = source.replace('\n', replacement)
        try:
            result = validate(source2)
            for ast in result:
                # Check no Tokens or tuples have snuck through.
                assert isinstance(ast, AST), ast
        except ASTError as ast_error:
            result = [e.message for e in ast_error.errors]
        assert matches(result, pattern), (
            "Parsed:\n"
            "    %r\n"
            "as:\n"
            "    %s\n"
            "Expected:\n"
            "    %s\n" % (
                source2,
                pprint.pformat(result).replace('\n', '\n    '),
                pprint.pformat(pattern).replace('\n', '\n    ')
            )
        )


# Common error messages.
error_missing_comma_or_rbracket = 'Expected "," or a closing bracket'
error_bad_type = 'The expression is not a type'
error_confusing_cast = 'Operator ":" is forbidden where a type is expected'
error_illegal_char = 'Illegal character'
error_missing_expr = 'Expected an expression'
error_missing_expr_or_rbracket = 'Expected an expression or a closing bracket'
error_missing_id = 'Expected an identifier after "."'
error_not_identifier = 'The expression is not an identifier'
error_missing_equals = 'Expected "="'
error_missing_stmt = 'Expected a statement'
error_ambiguous_cast = 'Ambiguous ":"; please add round brackets'
error_not_assignable = 'The expression is not assignable'
error_missing_rsqbracket = 'Expected "]"'
error_missing_newline = 'Expected ";" or a newline'
error_missing_condition = 'Expected a condition after "if"'
error_unexpected_else = (
    '"else" is only allowed at the end of an "if", "for" or "while" '
    'statement (on the same line) or inside a "switch" statement')


# Useful abbreviations in parse trees.
int_1 = ('integer', False, "1")
int_2 = ('integer', False, "2")
int_3 = ('integer', False, "3")
int_4 = ('integer', False, "4")
int_5 = ('integer', False, "5")
id_f = ('identifier', True, 'f')
id_t = ('identifier', True, 't')
id_x = ('identifier', True, 'x')
id_y = ('identifier', True, 'y')
id_z = ('identifier', True, 'z')
id_term = ('identifier', True, 'term')
id_int = ('identifier', True, 'int')
id_str = ('identifier', True, 'str')

def mkblock(stmts):
    assert type(stmts) is list
    return ('block', stmts)

def assign(left, op='=', right=int_1):
    return ('assign', left, ('assign_op', op), right)

def mkexpr(x):
    '''
    Turns the parse tree of an expression into the parse tree of a statement.
    '''
    return ('expr', x)

def binary(left, op, right):
    return ('binary', False, left, ('bin_op', op), right)

def call(f, args):
    assert type(args) is list
    return ('call', False, f, args)

def unary(op, right):
    return ('unary', False, ('un_op', op), right)


def test_comments():
    check('// break', [])
    check('/* break */', [])


# stmts = (stmt? (";" / newline))* stmt?
def test_semicolons():
    check("", [])
    check(";", [])
    check(";;", [])
    # TODO: more tests here, inc. newlines


# stmt = expr / ...

# expr = expr_nonfix / ...

# expr_nonfix = "integer" / "hex_integer" / "float" / "string" /
#               "character" / "identifier" / ...
def test_expr_literal():
    error_bad_unicode = 'Bad unicode escape sequence'
    check("1", [mkexpr(int_1)])
    check("0x1", [mkexpr(('hex_integer', False, "0x1"))])
    check("0.1", [mkexpr(('float', False, "0.1"))])
    check("0 .1", [error_missing_newline])
    check("0. 1", [error_missing_newline])
    check("0 . 1", [error_missing_id])
    check('"a"', [mkexpr(('string', False, 'a'))])
    check(r'"\n"', [mkexpr(('string', False, '\n'))])
    check(r'"\u"', [error_bad_unicode])
    check(r'"a\na"', [mkexpr(('string', False, 'a\na'))])
    check(r'"a\ua"', [error_bad_unicode])
    check("'a'", [mkexpr(('character', False, 'a'))])
    check("f", [mkexpr(id_f)])


# expr_nonfix = "()"(expr) / ...
def test_expr_bracketed():
    check("()", [mkexpr(('struct', True, []))])
    check("(\n)", [mkexpr(('struct', True, []))])
    check("(@)", [error_illegal_char])
    check("(1)", [mkexpr(int_1)])
    check("(\n1)", [mkexpr(int_1)])
    check("(1\n)", [mkexpr(int_1)])
    check("(\n1\n)", [mkexpr(int_1)])
    check("(1,)", [mkexpr(('struct', False, [int_1]))])
    check("(1\n,\n)", [mkexpr(('struct', False, [int_1]))])
    check("(\n1\n,)", [mkexpr(('struct', False, [int_1]))])
    check("(\n1\n,\n)", [mkexpr(('struct', False, [int_1]))])
    check("(1,,)", [error_missing_expr_or_rbracket])
    check("(1\nx)", [error_missing_comma_or_rbracket])
    check("(\n1\nx)", [error_missing_comma_or_rbracket])
    check("(1\nx\n)", [error_missing_comma_or_rbracket])
    check("(\n1\nx\n)", [error_missing_comma_or_rbracket])
    check("(1,\nx)", [mkexpr(('struct', False, [int_1, id_x]))])
    check("(1,\nx,)", [mkexpr(('struct', False, [int_1, id_x]))])
    check("(1\n,\nx\n,\n)",[
        mkexpr(('struct', False, [int_1, id_x]))
    ])
    check("(\n1\n,\nx\n,)", [
        mkexpr(('struct', False, [int_1, id_x]))
    ])
    check("(\n1\n,\nx\n,\n)", [
        mkexpr(('struct', False, [int_1, id_x]))
    ])
    check("(1,\nx,,)", [error_missing_expr_or_rbracket])
    # TODO: test "(foo\n.bar)".


# expr_nonfix = "[]"(expr*) / ...
def test_expr_list():
    check("[]", [mkexpr(('array', False, []))])
    check("[\n]", [mkexpr(('array', False, []))])
    check("[1]", [mkexpr(('array', False, [int_1]))])
    check("[1\n]", [mkexpr(('array', False, [int_1]))])
    check("[\n1]", [mkexpr(('array', False, [int_1]))])
    check("[\n1\n]", [mkexpr(('array', False, [int_1]))])
    check("[1,]", [mkexpr(('array', False, [int_1]))])
    check("[1\n,\n]", [mkexpr(('array', False, [int_1]))])
    check("[\n1\n,]", [mkexpr(('array', False, [int_1]))])
    check("[\n1\n,\n]", [mkexpr(('array', False, [int_1]))])
    check("[,]", [error_missing_expr_or_rbracket])
    check("[,,]", [error_missing_expr_or_rbracket])
    check("[,1]", [error_missing_expr_or_rbracket])
    check("[,1,]", [error_missing_expr_or_rbracket])
    check("[1,,]", [error_missing_expr_or_rbracket])
    check("[1\nx]", [error_missing_comma_or_rbracket])
    check("[1\nx\ny]", [error_missing_comma_or_rbracket])
    check("[1\n,\nx\ny]", [error_missing_comma_or_rbracket])
    check("[1\nx\n,\ny]", [error_missing_comma_or_rbracket])
    check("[\n1\nx]", [error_missing_comma_or_rbracket])
    check("[1\nx\n]", [error_missing_comma_or_rbracket])
    check("[\n1\nx\n]", [error_missing_comma_or_rbracket])
    check("[1,\nx]", [mkexpr(('array', False, [int_1, id_x]))])
    check("[1,\nx,]", [mkexpr(('array', False, [int_1, id_x]))])
    check("[1\n,\nx\n,\n]", [mkexpr(('array', False, [int_1, id_x]))])
    check("[\n1\n,\nx\n,]", [mkexpr(('array', False, [int_1, id_x]))])
    check("[\n1\n,\nx\n,\n]", [mkexpr(('array', False, [int_1, id_x]))])
    check("[1,\nx,,]", [error_missing_expr_or_rbracket])


# expr_nonfix = "()"(expr*) / ...
def test_expr_struct():
    check("()", [mkexpr(('struct', True, []))])
    check("(\n)", [mkexpr(('struct', True, []))])
    check("(\n:\n)", [error_missing_expr_or_rbracket])
    check("(\n:\n:\n)", [error_missing_expr_or_rbracket])
    check("(\n1\n,)", [mkexpr(('struct', False, [int_1]))])
    check("(\n1:\n,)", [error_missing_expr])
    check("(\n1:\n:\n,)", [error_missing_expr])
    check("(\n:\n1\n,)", [error_missing_expr_or_rbracket])
    check("(\nx:\n1\n,)", [error_bad_type])
    check("(\n1:\nx\n,)",
        [mkexpr(('struct', False, [('cast', int_1, id_x)]))])
    check("(\n1:\nx\n,\n/\n)", [error_missing_expr_or_rbracket])
    check("(\n1:\nx\n,\n@\n)", [error_illegal_char])
    check("(\nx:\ny:\n1\n,)", [error_ambiguous_cast])
    check("(\nx:\ny:\nf\n,)", [error_ambiguous_cast])
    check("(\n(\nx:\ny\n) :\n1\n,)", [error_bad_type])

    # TODO: Make "()" tests analogous to these "[]" tests:
    #check("[1,,]", [mkexpr(None, ('array', False, [int_1, error_missing_expr_or_rbracket]), None)])
    #check("[1 x]", [mkexpr(None, ('array', False, [int_1, id_x]), None)])
    #check("[1, x]", [mkexpr(None, ('array', False, [int_1, id_x]), None)])
    #check("[1, x,]", [mkexpr(None, ('array', False, [int_1, id_x]), None)])
    #check("[1, x,,]", [
    #    mkexpr(None, ('array', False, [int_1, id_x, error_missing_expr_or_rbracket]), None)])


# expr_nonfix = def / ...
def test_expr_function():
    error_named_function_type = (
        'The function type has a name; please add a body if you mean '
        'to define a function')
    error_missing_return_type = (
        'Expected a function body, or ":" and a return type')
    error_missing_expr_in_bracket = 'Expected an expression inside "( ... )"'
    error_missing_bracket = 'Expected ")"'
    call_list_t = call(('identifier', True, 'list'), [id_t])

    check("fn\n(\nx\n:\nt\n)", [error_missing_return_type])
    check("fn\n(\nx\n,\ny\n)", [error_missing_return_type])
    check("fn\n(\nt\n)\n:\nt", [
        mkexpr(('type_function',
            False,
            [('param', None, id_t)],
            id_t,
        ))
    ])
    check("fn\n(\nx+\n1\n)\n:\nt", [error_bad_type])
    check("fn\n(\nx:\nt\n)\n:\nt", [
        mkexpr(('type_function',
            False,
            [('param', id_x, id_t)],
            id_t,
        ))
    ])
    # Newline terminates expression after "x".
    check("fn\n(\nx\n:\nt\n)\n:\nt",
        [error_missing_comma_or_rbracket],
        is_safe_to_remove_newlines=False)
    check("fn\nf\n(\nx:\nt\n)", [error_missing_return_type])
    check("fn\nf(\nx:\nt\n)\n:\nt", [error_named_function_type])
    check("fn\n(\nx:\nt\n) {\nreturn x\n}", [
        mkexpr(('function_def',
            False,
            None,
            [('param', id_x, id_t)],
            None,
            mkblock([('return', id_x)]),
        ))
    ])
    check("fn\n(\nx:\nt\n)\n:\nt {\nreturn x\n}", [
        mkexpr(('function_def',
            False,
            None,
            [('param', id_x, id_t)],
            id_t,
            mkblock([('return', id_x)]),
        ))
    ])
    check("fn\nf\n(\nx :\nlist(\nt\n)\n) {\nreturn x\n}", [
        mkexpr(('function_def',
            False,
            id_f,
            [('param', id_x, call_list_t)],
            None,
            mkblock([('return', id_x)]),
        ))
    ])
    check(
        "fn\nf\n(\nx :\nlist(\nt\n)\n)\n:\n(\nlist(\nt\n)\n) "
        "{\nreturn x\n}",
        [
            mkexpr(('function_def',
                False,
                id_f,
                [('param', id_x, call_list_t)],
                call_list_t,
                mkblock([('return', id_x)]),
            ))
        ]
    )
    check("fn\n(\n):(\n)", [error_missing_expr_in_bracket])
    check("fn\n(\n):(\nt\n,\n)", [error_missing_bracket])
    check("fn\n(\n):(\nt\n,\nt\n)", [error_missing_bracket])
    check("fn\n(\nx:\n1\n)\n:\nt", [error_bad_type])
    check("fn\n(\nx:\ny:\n1\n)\n:\nt", [error_confusing_cast])
    check("fn\n(\nx:\ny:\nf\n)\n:\nt", [error_confusing_cast])
    check("fn\n(\n(\nx :\ny\n) :\n1\n)\n:\nt",
        [error_not_identifier])
    check("fn\n(\nx :\n(\ny :\n1\n)\n)\n:\nt",
        [error_confusing_cast])
    # Newline before body should split the expression in two.
    check("fn ()\n: t\n{}",
        [error_missing_stmt],
        is_safe_to_remove_newlines=False
    )
    # Parameter declarations that use "var" or "const".
    check("fn\n(\nconst\nx:\nint\n)\n:\nint",
        [error_missing_expr_or_rbracket])
    check("fn\n(\nvar\nx:\nint\n)\n:\nint",
        [error_missing_expr_or_rbracket])
    # Erroneous parameter declarations that have an 'assign_op'.
    check("fn\n(\nx=\n1\n)\n:\nint",
        [error_missing_comma_or_rbracket])


# expr_nonfix = type_struct / ...
def test_expr_struct_types():
    error_incomplete_struct1 = 'Expected an identifier or "( ... )"'
    error_incomplete_struct2 = 'Expected "( ... )"'
    check("struct", [error_incomplete_struct1])
    check("struct f", [error_incomplete_struct2])
    check("struct ()", [mkexpr(('type_struct', False, None, []))])
    check("struct f()", [mkexpr(('type_struct', False, id_f, []))])
    check("struct (int)", [
        mkexpr(('type_struct', False, None, [('param', None, id_int)]))
    ])
    check("struct (x: int)", [
        mkexpr(('type_struct', False, None, [('param', id_x, id_int)]))
    ])
    check("struct (x: int,)", [
        mkexpr(('type_struct', False, None, [('param', id_x, id_int)]))
    ])
    check("struct (x: int, str)", [
        mkexpr(('type_struct', False, None, [
            ('param', id_x, id_int),
            ('param', None, id_str),
        ]))
    ])
    check("struct (x: int, y: str)", [
        mkexpr(('type_struct', False, None, [
            ('param', id_x, id_int),
            ('param', id_y, id_str),
        ]))
    ])


# expr_nonfix = type_union / ...
def test_expr_union_types():
    error_incomplete_union1 = 'Expected an identifier or "{ ... }"'
    error_incomplete_union2 = 'Expected "{ ... }"'
    error_incomplete_union3 = (
        'Expected a case declaration, e.g. "TAG(arg, ...)"')
    check("union", [error_incomplete_union1])
    check("union foo", [error_incomplete_union2])
    check("union {}", [mkexpr(('type_union', False, None, []))])
    check("union { break; }", [error_incomplete_union3])
    check("""\
        union term {
            VAR(str);
            FUN(str, term);
            APP(term, term);
            Y;
        }
    """, [
        mkexpr(('type_union', False, id_term, [
            ('union_case', ('tag', 'VAR'), [id_str]),
            ('union_case', ('tag', 'FUN'), [id_str, id_term]),
            ('union_case', ('tag', 'APP'), [id_term, id_term]),
            ('union_case', ('tag', 'Y'), []),
        ]))
    ])
    check("union {3}", [error_incomplete_union3])
    check("union {c}", [error_incomplete_union3])
    check("union {C}", [
        ('expr',
            ('type_union', False, None, [('union_case', ('tag', 'C'), [])])
        ),
    ])
    check("union {c[int]}", [error_incomplete_union3])
    check("union {C[int]}", [error_incomplete_union3])
    check("union {(int)}", [error_incomplete_union3])
    check("union {c()}", [error_incomplete_union3])
    check("""\
        union term {
            VAR(name: str);
            FUN(arg: str, body: term);
            APP(f: term, v: term);
        }
    """, [
        error_confusing_cast,
        error_confusing_cast,
        error_confusing_cast,
    ])


# expr_nonfix = union / ...
def test_expr_union():
    check('C', [mkexpr(('union', False, ('tag', 'C'), []))])
    check('C()', [mkexpr(('union', False, ('tag', 'C'), []))])
    check('C(1)', [mkexpr(('union', False, ('tag', 'C'), [int_1]))])
    check('C(1,)', [mkexpr(('union', False, ('tag', 'C'), [int_1]))])
    check('C(1, 1)', [
        mkexpr(('union', False, ('tag', 'C'), [int_1, int_1]))
    ])


# expr = expr_prefix / ...
def test_expr_prefix():
    check("!1", [mkexpr(unary('!', int_1))])
    check("\n!\n1\n", [mkexpr(unary('!', int_1))])
    check("!!1", [mkexpr(unary('!', unary('!', int_1)))])
    check("!\n!1", [mkexpr(unary('!', unary('!', int_1)))])
    check("\n!\n!\n1", [mkexpr(unary('!', unary('!', int_1)))])
    check("-1", [mkexpr(unary('-', int_1))])
    check("\n-\n1\n", [mkexpr(unary('-', int_1))])
    check("--1", [mkexpr(unary('--', int_1))])
    check("\n--\n1\n", [mkexpr(unary('--', int_1))])
    check("+1", [mkexpr(unary('+', int_1))])
    check("++1", [mkexpr(unary('++', int_1))])
    check("new 1", [mkexpr(('new', False, int_1))])
    check("defer 1", [error_bad_type])
    check("defer typeof 1", [
        mkexpr(('defer', False, ('typeof', False, int_1)))
    ])
    check("defer *&typeof 1", [
        mkexpr(
            ('defer',
                False,
                ('dereference',
                    True,
                    ('address_of', False, ('typeof', False, int_1))
                )
            )
        )
    ])


# expr = expr_postfix / ...

# expr_postfix = "field"(expr, identifier) / ...
def test_expr_field():
    check("f.x", [mkexpr(('field', True, id_f, id_x))])
    check("f.\nx", [mkexpr(('field', True, id_f, id_x))])
    check("f.x.y", [
        mkexpr(('field', True, ('field', True, id_f, id_x), id_y))
    ])
    check("f.\nx.\ny", [
        mkexpr(('field', True, ('field', True, id_f, id_x), id_y))
    ])
    check("f.(x)", [error_missing_id])
    check("f.\n(\nx\n)", [error_missing_id])


# expr_postfix = "call"(expr, expr*) / ...
def test_expr_call():
    check("f()", [mkexpr(call(id_f, []))])
    check("f(\n)", [mkexpr(call(id_f, []))])
    check("f(,)", [error_missing_expr_or_rbracket])
    check("f(\n,\n)", [error_missing_expr_or_rbracket])
    check("f(x)", [mkexpr(call(id_f, [id_x]))])
    check("f(\nx\n)", [mkexpr(call(id_f, [id_x]))])
    check("f(x,)", [mkexpr(call(id_f, [id_x]))])
    check("f(\nx\n,\n)", [mkexpr(call(id_f, [id_x]))])
    check("f(x,,)", [error_missing_expr_or_rbracket])
    check("f(\nx\n,\n,\n)", [error_missing_expr_or_rbracket])
    check("f(x,@)", [error_illegal_char])
    check("f(\nx\n,\n@\n)", [error_illegal_char])
    check("f(x,,@)", [error_missing_expr_or_rbracket])
    check("f(\nx\n,\n,\n@\n)", [error_missing_expr_or_rbracket])
    check("f(,x)", [error_missing_expr_or_rbracket])
    check("f(\n,\nx\n)", [error_missing_expr_or_rbracket])
    check("f(,,x)", [error_missing_expr_or_rbracket])
    check("f(x\ny)", [error_missing_comma_or_rbracket])
    check("f(\nx\ny\n)", [error_missing_comma_or_rbracket])
    check("f(x\ny,)", [error_missing_comma_or_rbracket])
    check("f(\nx\ny\n,\n)", [error_missing_comma_or_rbracket])
    check("f(x,\ny)", [mkexpr(call(id_f, [id_x, id_y]))])
    check("f(\nx\n,\ny\n)", [mkexpr(call(id_f, [id_x, id_y]))])
    check("f(x,\ny,)", [mkexpr(call(id_f, [id_x, id_y]))])
    check("f(\nx\n,\ny\n,\n)", [mkexpr(call(id_f, [id_x, id_y]))])
    check("f(x:\ny)", [mkexpr(call(id_f, [('cast', id_x, id_y)]))])
    check("f(x\n:\ny)",
        [error_missing_comma_or_rbracket],
        is_safe_to_remove_newlines=False)
    check("f(\nx :\ny\n)", [
        mkexpr(call(id_f, [('cast', id_x, id_y)]))
    ])
    check("f(x;\ny)", [error_missing_comma_or_rbracket])
    check("f(\nx\n;\ny\n)", [error_missing_comma_or_rbracket])
    check("f(x; @)", [error_missing_comma_or_rbracket])
    check("f(\nx\n; @)", [error_missing_comma_or_rbracket])


# expr_postfix = "subscript"(expr, expr*) / ...
def test_expr_subscript():
    error_missing_index = 'Expected an array index'
    check("f[]", [error_missing_index])
    check("f[\n]", [error_missing_index])
    check("f[,]", [error_missing_expr_or_rbracket])
    check("f[\n,\n]", [error_missing_expr_or_rbracket])
    check("f[x]", [mkexpr(('subscript', True, id_f, id_x))])
    check("f[\nx\n]", [mkexpr(('subscript', True, id_f, id_x))])
    check("f[x,]", [error_missing_rsqbracket])
    check("f[\nx\n,\n]", [error_missing_rsqbracket])
    check("f[x,,]", [error_missing_expr_or_rbracket])
    check("f[\nx\n,\n,\n]", [error_missing_expr_or_rbracket])
    check("f[x,@]", [error_illegal_char])
    check("f[\nx\n,\n@\n]", [error_illegal_char])
    check("f[x,,@]", [error_missing_expr_or_rbracket])
    check("f[\nx\n,\n,\n@\n]", [error_missing_expr_or_rbracket])
    check("f[,x]", [error_missing_expr_or_rbracket])
    check("f[\n,\nx\n]", [error_missing_expr_or_rbracket])
    check("f[,,x]", [error_missing_expr_or_rbracket])
    check("f[\n,\n,\nx\n]", [error_missing_expr_or_rbracket])
    check("f[x\ny]", [error_missing_comma_or_rbracket])
    check("f[\nx\ny\n]", [error_missing_comma_or_rbracket])
    check("f[x\ny,]", [error_missing_comma_or_rbracket])
    check("f[\nx\ny\n,\n]", [error_missing_comma_or_rbracket])
    check("f[x,\ny]", [error_missing_rsqbracket])
    check("f[\nx\n,\ny\n]", [error_missing_rsqbracket])
    check("f[x,\ny,]", [error_missing_rsqbracket])
    check("f[\nx\n,\ny\n,\n]", [error_missing_rsqbracket])
    check("f[x:\ny]", [
        mkexpr(('subscript', True, id_f, ('cast', id_x, id_y)))
    ])
    check("f[\nx:\ny\n]", [
        mkexpr(('subscript', True, id_f, ('cast', id_x, id_y)))
    ])
    check("f[x;\ny]", [error_missing_comma_or_rbracket])
    check("f[\nx\n;\ny\n]", [error_missing_comma_or_rbracket])
    check("f[x;\n@]", [error_missing_comma_or_rbracket])
    check("f[\nx\n;\n@\n]", [error_missing_comma_or_rbracket])


# expr = expr_infix / ...
def test_expr_infix():
    check("1+@", [error_illegal_char])
    check("1+x", [mkexpr(binary(int_1, '+', id_x))])
    check("\n1+\nx\n", [mkexpr(binary(int_1, '+', id_x))])
    check("\n1\n+\nx\n",
        [mkexpr(int_1), mkexpr(unary('+', id_x))],
        is_safe_to_remove_newlines=False)
    check("1^x", [mkexpr(binary(int_1, '^', id_x))])
    check("1+\n+x", [mkexpr(binary(int_1, '+', unary('+', id_x)))])
    check("\n1+\n+\nx\n", [mkexpr(binary(int_1, '+', unary('+', id_x)))])
    check("1^\n+x", [mkexpr(binary(int_1, '^', unary('+', id_x)))])
    check("\n1^\n+\nx\n", [mkexpr(binary(int_1, '^', unary('+', id_x)))])
    check("1+\n^x", [error_missing_expr, error_missing_stmt])
    check("\n1+\n^\nx\n", [error_missing_expr, error_missing_stmt])
    check("1^\n^x", [error_missing_expr, error_missing_stmt])
    check("\n1\n^\n^\nx\n",
        [error_missing_stmt, error_missing_stmt],
        is_safe_to_remove_newlines=False)
    check("1:@", [error_illegal_char])
    check("1:x", [mkexpr(('cast', int_1, id_x))])
    check("1:x:y", [error_ambiguous_cast])


# expr = expr_ternary / ...
def test_expr_ternary():
    error_missing_colon = 'Expected ":" after "?"'
    error_ambiguous_ternary = 'Ambiguous "?"; please add round brackets'
    error_confusing_ternary = 'Confusing "?"; please add round brackets'
    check("1?x", [error_missing_colon])
    check("1?@", [error_illegal_char])
    check("1?x:y", [mkexpr(('ternary', True, int_1, id_x, id_y))])
    check("1?x:y:t", [error_ambiguous_cast])
    check("1:t?x:y", [error_ambiguous_ternary])
    check("1?1?x:y:z", [error_confusing_ternary])
    check("1?x:1?x:y", [error_ambiguous_ternary])


def test_expr_type():
    # Test the extra error messages that apply to expressions used as types.
    check("1:1", [error_bad_type])
    check("1:(x:y)", [error_confusing_cast])
    check("1:defer t", [mkexpr(('cast', int_1, ('defer', False, id_t)))])

# stmt = assign / ...

# assign = lvalue assign_op expr

# lvalue = lvalue_nonfix / ...

# lvalue_nonfix = "identifier" / ...
def test_assign_identifier():
    check("x =\n1", [assign(id_x)])
    check("x +=\n1", [assign(id_x, '+=')])
    check("x ^=\n1", [assign(id_x, '^=')])
    check("x ==\n1", [mkexpr(binary(id_x, '==', int_1))])


# lvalue_nonfix = "()"(lvalue) / ...
def test_assign_bracketed():
    check("() =\n1", [assign(('struct', True, []))])
    check("(\n) =\n1", [assign(('struct', True, []))])
    check("(x) =\n1", [assign(id_x)])
    check("(\nx\n) =\n1", [assign(id_x)])


# lvalue_nonfix = "struct"(lvalue*) / ...
def test_assign_struct():
    error_named_field_in_lvalue = (
        'Named fields are not allowed in assignable expressions; '
        'please use round brackets if you mean "(expression: type)"')
    check("() =\n1", [assign(('struct', True, []))])
    check("(\n) =\n1", [assign(('struct', True, []))])
    check("(x,) =\n1", [assign(('struct', True, [id_x]))])
    check("(\nx\n,) =\n1", [assign(('struct', True, [id_x]))])
    check("(x,) =\n1", [assign(('struct', True, [id_x]))])
    check("(\nx\n,\n) =\n1", [assign(('struct', True, [id_x]))])
    check("(x,,) =\n1", [error_missing_expr_or_rbracket])
    check("(\nx\n,\n,\n) =\n1", [error_missing_expr_or_rbracket])
    check("(x,@) =\n1", [error_illegal_char])
    check("(\nx\n,\n@\n) =\n1", [error_illegal_char])
    check("(x,,@) =\n1", [error_missing_expr_or_rbracket])
    check("(\nx\n,\n,@) =\n1", [error_missing_expr_or_rbracket])
    check("(,x) =\n1", [error_missing_expr_or_rbracket])
    check("(\n,\nx\n) =\n1", [error_missing_expr_or_rbracket])
    check("(,,x) =\n1", [error_missing_expr_or_rbracket])
    check("(x\ny) =\n1", [error_missing_comma_or_rbracket])
    check("(\nx\ny\n) =\n1", [error_missing_comma_or_rbracket])
    check("(x\ny,) =\n1", [error_missing_comma_or_rbracket])
    check("(\nx\ny\n,\n) =\n1", [error_missing_comma_or_rbracket])
    check("(x,\ny) =\n1",
        [assign(('struct', True, [id_x, id_y]))])
    check("(\nx\n,\ny\n) =\n1",
        [assign(('struct', True, [id_x, id_y]))])
    check("(x,\ny,) =\n1",
        [assign(('struct', True, [id_x, id_y]))])
    check("(\nx\n,\ny\n,\n) =\n1",
        [assign(('struct', True, [id_x, id_y]))])
    check("(x:\ny,) =\n1",
        [assign(('struct', True, [('cast', id_x, id_y)]))])
    check("(\nx:\ny\n,) =\n1",
        [assign(('struct', True, [('cast', id_x, id_y)]))])
    check("(x;\ny) =\n1", [error_missing_comma_or_rbracket])
    check("(x;\n@) =\n1", [error_missing_comma_or_rbracket])
    check("(@, @) =\n1", [error_illegal_char])
    check("(x,\ny,) +=\n1", [
        assign(('struct', True, [id_x, id_y]), op='+=')
    ])


# lvalue = lvalue_prefix / ...

# lvalue_prefix = *expr / ...
def test_assign_prefix():
    check("*x =\n1", [('assign',
        ('dereference', True, id_x),
        ('assign_op', '='),
        ('integer', False, '1')
    )])
    check("!x =\n1", [error_not_assignable])
    check("&x =\n1", [error_not_assignable])


# lvalue = lvalue_postfix / ...

# lvalue_postfix = "field"(lvalue, identifier) / ...
def test_assign_field():
    check("f.x =\n1", [assign(('field', True, id_f, id_x))])
    check("f.\nx =\n1", [assign(('field', True, id_f, id_x))])
    check("1.x =\n1", [error_missing_newline])
    check("1. x =\n1", [error_missing_newline])
    check("1 .x =\n1", [error_not_assignable])
    check("1 .\nx =\n1", [error_not_assignable])


# lvalue_postfix = "subscript"(lvalue, expr*) / ...
def test_assign_subscript():
    check("f[1] =\n1", [assign(('subscript', True, id_f, int_1))])
    check("f[\n1\n] =\n1", [assign(('subscript', True, id_f, int_1))])
    check("f[x] =\n1", [assign(('subscript', True, id_f, id_x))])
    check("f[\nx\n] =\n1", [assign(('subscript', True, id_f, id_x))])
    check("f[\nx\n,\ny\n] =\n1", [error_missing_rsqbracket])
    check("1[x] =\n1", [assign(('subscript', True, int_1, id_x))])
    check("1[\nx\n] =\n1", [assign(('subscript', True, int_1, id_x))])


# lvalue = ":"(lvalue, expr)
def test_assign_infix():
    check("x:\ny =\n1", [assign(('cast', id_x, id_y))])
    check("x\n:\ny =\n1",
        [error_missing_stmt],
        is_safe_to_remove_newlines=False)
    check("x\n:\n:\ny =\n1",
        [error_missing_stmt, error_missing_stmt],
        is_safe_to_remove_newlines=False)
    check("x:\n1 =\n1", [error_bad_type])
    check("1:\nx =\n1", [error_not_assignable])
    check("x:\n=\n1", [error_missing_expr, error_missing_stmt])


# lvalue = lvalue_ternary / ...
def test_assign_ternary():
    error_missing_colon = 'Expected ":" after "?"'
    error_ambiguous_ternary = 'Ambiguous "?"; please add round brackets'
    error_confusing_ternary = 'Confusing "?"; please add round brackets'
    check("1?x = 1", [error_missing_colon])
    check("1?@ = 1", [error_illegal_char, error_missing_stmt])
    check("1?x:y = 1", [assign(('ternary', True, int_1, id_x, id_y))])
    check("1?x:y:t = 1", [error_ambiguous_cast])
    check("1:t?x:y = 1", [error_not_assignable])
    check("f:t?x:y = 1", [error_ambiguous_ternary])
    check("1?1?x:y:z = 1", [error_confusing_ternary])
    check("1?x:1?x:y = 1", [error_ambiguous_ternary])


# stmt = decl / ...

# decl = keyword lvalue expr
def test_decl():
    def decl(decl, left, right):
        return ('decl', (decl,), left, right)

    error_missing_rhs = (
        'Expected an expression for the right-hand side of the declaration')
    error_illegal_name = (
        'Named fields are not allowed in assignable expressions; please use '
        'round brackets if you mean "(expression: type)"')

    check("var\nx", [error_missing_equals])
    check("const\nx", [error_missing_equals])
    check("var\nx 1", [error_missing_equals])
    check("var\nx; 1", [error_missing_equals])
    check("var\nx--", [error_missing_equals])
    check("var\nx; --", [error_missing_equals, error_missing_expr])
    check("var\nx =", [error_missing_rhs])
    check("var\nx =\n1", [decl('var', id_x, int_1)])
    check("const\nx =\n1", [decl('const', id_x, int_1)])
    check("var\nx +=\n1", [error_missing_equals])
    check("var\nx ==\n1", [error_missing_equals])
    check("var\n(\n) =\n1", [decl('var', ('struct', True, []), int_1)])
    check("var\n(x) =\n1", [decl('var', id_x, int_1)])
    check("var\n(x\ny) =\n1", [error_missing_comma_or_rbracket])
    check("var\n((x)) =\n1", [decl('var', id_x, int_1)])
    check("var\n[] =\n1", [error_not_assignable])
    check("var\n(1,) =\n1", [error_not_assignable])
    check("var\n(x,) = 1", [decl('var', ('struct', True, [id_x]), int_1)])
    check("var\n(x,\ny) =\n1",
        [decl('var', ('struct', True, [id_x, id_y]), int_1)])
    check("var\n(x,\ny) +=\n1", [error_missing_equals])
    check("var\n(x:\nt,\ny) =\n1",
        [decl('var', ('struct', True, [('cast', id_x, id_t), id_y]), int_1)])


# stmt = break / ...
def test_break():
    check("break", [('break',)])


def test_case():
    case_error = '"case" can only occur inside a "switch" statement'
    check("case Var() {}", [case_error])


# stmt = continue / ...
def test_continue():
    check("continue", [('continue',)])


def test_else():
    check("else {}", [error_unexpected_else])


# stmt = for / ...
def test_for():
    error_missing_loop_variable = 'Expected a loop variable after "for"'
    error_missing_in = 'Expected "in" after the "for" loop variable'
    error_missing_in_comma_hint = (
        'Expected "in" after the "for" loop variable '
        '(structs, including tuples, must be enclosed in "( ... )")')
    error_missing_sequence = 'Expected a sequence after "for ... in"'
    error_missing_for_block = 'Expected "{ ... }" after "for ... in ..."'
    error_missing_else_block = 'Expected "{ ... }" after "for ... else"'
    check("for", [error_missing_loop_variable])
    check("for x", [error_missing_in])
    check("for x in", [error_missing_sequence])
    check("for x in 1", [error_missing_for_block])
    check("for x in 1 {}", [
        ('for', id_x, int_1, mkblock([]), None),
    ])
    check("for x in 1 {} else", [error_missing_else_block])
    check("for x in 1 {} else {}", [
        ('for', id_x, int_1, mkblock([]), mkblock([])),
    ])
    check("for (x, y) in 1 {} else {}", [error_not_identifier])
    check("for x, y in 1 {} else {}", [
        error_missing_in, error_missing_stmt, error_missing_newline,
        error_missing_newline, error_unexpected_else
    ])
    check("for x in 1 {} else if", [
        error_missing_else_block, error_missing_condition])
    check("for x =", [error_missing_in, error_missing_stmt])
    check("for x = 1 to 3", [
        error_missing_in, error_missing_stmt, error_missing_newline])
    check("for x + y", [error_missing_in])


# stmt = fulfil / ...
def test_fulfil():
    check("fulfil\nx =\n1", [('fulfil', id_x, int_1)])
    check("fulfil\nx +=\n1", [error_missing_equals])


# stmt = if / ...
def test_if():
    check("if\n1\n{\nf(\n)\n}", [('if',
        int_1,
        mkblock([mkexpr(call(id_f, []))]),
        None
    )])
    check(
        "if\n1\n{ x } else if 2 { y } else { z }",
        [('if',
            int_1,
            mkblock([mkexpr(id_x)]),
            mkblock([('if',
                int_2,
                mkblock([mkexpr(id_y)]),
                mkblock([mkexpr(id_z)]),
            )]),
        )]
    )
    # TODO: Test more cases, e.g. "if" with nothing else.


# stmt = return / ...
def test_return():
    check("return", [('return', None)])
    check("return 1", [('return', int_1)])


# stmt = switch / ...
def test_switch():
    # TODO: Error cases.
    check('switch t {}', [('switch', id_t, [], None)])
    check('''\
            switch t {
                case VAR(x) { return 1 }
                case FUN(y, z) { return 2 }
                case APP(x, y) { return 3 }
                case Y { return 4 }
                else { return 5 }
            }\
        ''',
        [('switch',
            id_t,
            [
                ('case',
                    ('tag', 'VAR'),
                    [id_x],
                    mkblock([('return', int_1)]),
                ),
                ('case',
                    ('tag', 'FUN'),
                    [id_y, id_z],
                    mkblock([('return', int_2)]),
                ),
                ('case',
                    ('tag', 'APP'),
                    [id_x, id_y],
                    mkblock([('return', int_3)]),
                ),
                ('case',
                    ('tag', 'Y'),
                    [],
                    mkblock([('return', int_4)]),
                ),
            ],
            ('else', mkblock([('return', int_5)])),
        )],
        is_safe_to_remove_newlines=False
    )


# stmt = while / ...
def test_while():
    error_missing_condition = 'Expected a condition after "while"'
    error_missing_body = 'Expected "{ ... }" after "while"'
    check('while', [error_missing_condition])
    check('while {}', [error_missing_condition, error_missing_stmt])
    check('while 1', [error_missing_body])
    check('while 1 {}', [('while', int_1, mkblock([]), None)])
    check('while 1 { 2 } else { 3 }', [
        ('while', int_1, mkblock([mkexpr(int_2)]), mkblock([mkexpr(int_3)]))
    ])
    check('while x <= 1 {}', [
        ('while', binary(id_x, '<=', int_1), mkblock([]), None)
    ])

# stmt = throw / ...
# stmt = try / ...
# TODO.
