import pprint

from .token import matches, Token, ASCII
from .grammar import Buffer, Parser
from .lexer import *
from .brackets import BracketMatcher
from .atom import atom
from .unary_expr import unary_expr
from .expr import expr
from .stmt import stmt, stmt_tags
from .block import block


allowed_tags = {'error'}.union(stmt_tags)


def check(source, expected_result):
    assert type(source) is str, type(src)
    listener = Buffer()
    p = Parser(block, listener)
    p = Parser(stmt, p)
    p = Parser(expr, p)
    p = Parser(unary_expr, p)
    p = Parser(atom, p)
    p = BracketMatcher(p, {
        '(': lambda x: x,
        '[': lambda x: x,
        '{': lambda x: x,
    })
    p = Parser(high_level_lexer, p)
    p = Parser(low_level_lexer, p)
    for i, c in enumerate(source):
        p.feed(ASCII(i, c))
    p.flush()
    result = listener.contents
    for token in result:
        assert isinstance(token, Token), token
        assert token.tag in allowed_tags, token
    assert matches(result, expected_result), (
        "Parsed:\n"
        "    %r\n"
        "as:\n"
        "    %s\n"
        "Expected:\n"
        "    %s\n" % (
            source,
            pprint.pformat(result).replace('\n', '\n    '),
            pprint.pformat(expected_result).replace('\n', '\n    ')
        )
    )


error_illegal = ('error', 'Illegal character')
error_missing = ('error', 'Expected a statement')
error_separator = ('error', 'Expected ";" or a newline')


def test_semicolons():
    check('', [])
    check(';', [])
    check('\n', [])
    check(';;', [])
    check(';\n', [])
    check('\n;', [])
    check('\n\n', [])


def test_statements():
    check('break', [('break',)])
    check('break;', [('break',)])
    check('break\n', [('break',)])
    check('break; break', [('break',), ('break',)])
    check('break\nbreak', [('break',), ('break',)])


def test_errors():
    check('@', [error_illegal])
    check('@ @', [error_illegal, error_illegal])
    check('@ break', [error_illegal, ('break',)])
    check('@; @', [error_illegal, error_illegal])
    check('@\n@', [error_illegal, error_illegal])
    check('@; break', [error_illegal, ('break',)])
    check('@\nbreak', [error_illegal, ('break',)])


def test_bad():
    check(',', [error_missing])
    check(', ,', [error_missing, error_missing])
    check(', @', [error_missing, error_illegal])
    check(', break', [error_missing, ('break',)])
    check(',; ,', [error_missing, error_missing])
    check(',\n,', [error_missing, error_missing])
    check(',; @', [error_missing, error_illegal])
    check(',\n@', [error_missing, error_illegal])
    check(',; break', [error_missing, ('break',)])
    check(',\nbreak', [error_missing, ('break',)])


def test_separators():
    check('break break', [error_separator])
    check('break @', [error_illegal])
    check('break break break', [error_separator])
    check('break break @', [error_separator, error_illegal])
    check('break @ break', [error_illegal, ('break',)])
    check('break @ @', [error_illegal, error_illegal])
    check('break break; break', [error_separator, ('break',)])
    check('break break\nbreak', [error_separator, ('break',)])
    check('break break; @', [error_separator, error_illegal])
    check('break break\n@', [error_separator, error_illegal])
    check('break @; break', [error_illegal, ('break',)])
    check('break @\nbreak', [error_illegal, ('break',)])
    check('break @; @', [error_illegal, error_illegal])
    check('break @\n@', [error_illegal, error_illegal])
