import random

from .token import matches, Token, ASCII, Concat
from .grammar import Buffer
from .brackets import BracketMatcher, bracket_tags


allowed_tags = {'error', ' '}.union(bracket_tags)


observed_tags = set()


def check(s, expected_result):
    '''
    s - a list of characters, each of which can be a bracket or a space.
    '''
    listener = Buffer()
    b = BracketMatcher(listener, {
        '(': lambda x: x,
        '[': lambda x: x,
        '{': lambda x: x,
    })
    for i, c in enumerate(s):
        a = ASCII(i, c)
        if c in '([{':
            b.feed(Concat('open', [a]))
        elif c in ')]}':
            b.feed(Concat('close', [a]))
        else:
            b.feed(a)
    b.flush()
    result = listener.contents
    for token in result:
        assert isinstance(token, Token), token
        assert token.tag in allowed_tags, token
        observed_tags.add(token.tag)
    assert matches(result, expected_result), (
        "%r parsed as:\n"
        "%r\n"
        "(expected:\n"
        "%r\n)" % (s, result, expected_result)
    )


def test_empty():
    check('', [])


def test_brackets():
    # Code coverage tests (mainly for token types and error messages).
    for s in '()', '[]', '{}':
        x, y = s
        check('%s%s' % (x, y), [(s, [])])
        check(' %s%s' % (x, y), [(' ',), (s, [])])
        check('%s %s' % (x, y), [(s, [(' ',)])])
        check('%s%s ' % (x, y), [(s, []), (' ',)])
        check(x, [(s, [('error', 'Unmatched "%s"'%x)])])
        check('%s ' % x, [(s, [('error', 'Unmatched "%s"'%x), (' ',)])])
        check(' %s' % x, [(' ',), (s, [('error', 'Unmatched "%s"'%x)])])
        check(y, [('error', 'Unmatched "%s"' % y)])
        check(' %s' % y, [(' ',), ('error', 'Unmatched "%s"' % y)])
        check('%s ' % y, [('error', 'Unmatched "%s"' % y), (' ',)])
        check('%s%s' % (x, x), [(s, [(s, [('error', 'Unmatched "%s"'%x)])])])
        check('%s%s' % (y, y), [
            ('error', 'Unmatched "%s"' % y), ('error', 'Unmatched "%s"' % y)])
        for z in ')', ']', '}':
            if z != y:
                check('%s%s' % (x, z), [
                    ('error', 'Unmatched "%s"' % x),
                    ('error', 'Unmatched "%s"' % z)
                ])
                check('%s %s' % (x, z), [
                    ('error', 'Unmatched "%s"' % x),
                    ('error', 'Unmatched "%s"' % z)
                ])
                check('%s%s%s' % (x, z, y), [
                    ('error', 'Unmatched "%s"' % x),
                    ('error', 'Unmatched "%s"' % z)
                ])


def test_nested():
    def generate_check(n):
        '''
        Returns a string of length `n` and its expected parse-tree.
        '''
        if n == 0:
            return '', []
        if n == 1:
            return ' ', [(' ',)]
        p = random.randrange(1, n)
        s1, r1 = generate_check(p-1) # From 1 to p.
        s2, r2 = generate_check((n-1)-p) # From p to n-1.
        b = random.choice(['()', '[]', '{}'])
        return b[0]+s1+s2+b[1], [(b, r1+r2)]

    for _ in range(100):
        s, r = generate_check(50)
        check(s, r)


# TODO: Test error propagation and suppression more fully.


def test_coverage():
    for t0 in bracket_tags:
        assert t0 in observed_tags, t0
