import random
import re

from .token import *
from .lexer import *
from .grammar import Parser, Buffer


allowed_tags = {'error'}.union(
    high_level_lexer_tags, keywords,
)


observed_tags = set()


# TODO: Test the lexer levels independently.
def check(source, expected_result):
    """Lexes the first token only of `source`. Checks result."""
    assert type(source) is str, type(src)
    listener = Buffer()
    p = Parser(low_level_lexer, Parser(high_level_lexer, listener))
    for i, c in enumerate(source):
        p.feed(ASCII(i, c))
    p.flush()
    result = listener.contents
    for token in result:
        assert isinstance(token, Token), token
        assert token.tag in allowed_tags, token
        observed_tags.add(token.tag)
    assert matches(result, expected_result), (
        "%r parsed as:\n"
        "%r\n"
        "(expected:\n"
        "%r\n)" % (source, result, expected_result)
    )


def test_illegal():
    for c in ['\x00', '\t', '\f', '\r', '@', '$']:
        check(c, [('error', "Illegal character")])

    check('@@', [
        ('error', "Illegal character"),
        ('error', "Illegal character")
    ])


def test_whitespace():
    for s in ['', ' ', '  ']:
        check(s, [])
        check(s+'@', [('error', "Illegal character")])
        check('@'+s, [('error', "Illegal character")])


def test_newline():
    check('\n', [('newline', '\n')])
    check('\r\n', [('newline', '\n')])
    check('\r\r\n', [('error', "Illegal character"), ('newline', '\n')])
    check('\n\r', [('newline', '\n'), ('error', "Illegal character")])
    check('\n\r\r', [
        ('newline', '\n'),
        ('error', "Illegal character"),
        ('error', "Illegal character")
    ])
    check('\\\n', [])


def test_punctuation():
    for c in ',.:;':
        check(c, [(c,)])

    check('::', [(':',), (':',)])


def test_brackets():
    for c in '([{':
        check(c, [('open', c)])

    for c in ')]}':
        check(c, [('close', c)])


def test_integers():
    for n in range(20):
        s = str(n)
        check(s, [('integer', s)])
        check(s+'l', [('integer', s), ('identifier', 'l')])
        check(s+'L', [('integer', s), ('tag', 'L')])


def test_floats():
    def check_float(s):
        # print s
        check(s, [('float', s)])

    patterns = [
        '',
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
        '10', '01'
    ]

    for i in patterns:
        for f in patterns:
            if i and f:
                check_float(i+'.'+f)
                for e in 'e', 'E':
                    for s in '', '+', '-':
                        for n in '0', '01', '10':
                            check_float(i+'.'+f+e+s+n)

    for i in patterns:
        if i:
            for e in 'e', 'E':
                for s in '', '+', '-':
                    for n in '0', '01', '10':
                        check_float(i+e+s+n)

    check('1. 0', [('float', '1.'), ('integer', '0')])
    check('1 .0', [('integer', '1'), ('float', '.0')])
    check('1 . 0', [('integer', '1'), ('.',), ('integer', '0')])
    check('/*1.0*/', [])
    check('"1.0"', [('string', ['1.0'])])
    check('1.e0g', [('float', '1.e0'), ('identifier', 'g')])
    check('1.0f', [('float', '1.0'), ('identifier', 'f')])
    check('1.0e', [('error', 'Missing exponent')])
    check('1.0e+', [('error', 'Missing exponent')])
    check('1.0e-', [('error', 'Missing exponent')])
    check('1.e', [('error', 'Missing exponent')])
    check('1.e+', [('error', 'Missing exponent')])
    check('1.e-', [('error', 'Missing exponent')])
    check('.0e', [('error', 'Missing exponent')])
    check('.0e+', [('error', 'Missing exponent')])
    check('.0e-', [('error', 'Missing exponent')])
    check('1e', [('error', 'Missing exponent')])
    check('1e+', [('error', 'Missing exponent')])
    check('1e-', [('error', 'Missing exponent')])
    check('.e', [('.',), ('identifier', 'e')])

    # TODO: Test error cases of 'float'.


def test_hex_integers():
    def check_hex_integer(s):
        # print s
        check(s, [('hex_integer', s)])

    for x in '0x', '0X':
        check(x, [('error', 'Bad hex integer')])
        check(x+'.', [('error', 'Bad hex integer'), ('.',)])
        for i in '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '01':
            check_hex_integer(x+i)

    check('0x0.', [('hex_integer', '0x0'), ('.',)])


def test_identifiers():
    def check_identifier(s):
        assert s
        if re.match(r'[A-Z][A-Z0-9_]*\Z', s):
            check(s, [('tag', s)])
        else:
            check(s, [('identifier', s)])

    for s1 in ['_', 'a', 'A']:
        for s2 in ['', 'a', 'A', '_', '0']:
            check_identifier(s1 + s2)
            check_identifier(s1 + s2 + s2)

    check('0A', [('integer', '0'), ('tag', 'A')])
    check('0a', [('integer', '0'), ('identifier', 'a')])
    check("0_", [('integer', '0'), ('identifier', '_')])

    check(' a ', [('identifier', 'a')])


def test_keywords():
    check('fo', [('identifier', 'fo')])
    check('for', [('for',)])
    check('forz', [('identifier', 'forz')])


def test_escapes():
    check('\\', [('error', 'Bad escape sequence')])
    check('\\n', [('error',
        'Unexpected escape sequence. Escape sequences are only allowed inside '
        'string and character literals and comments'
    )])


def test_operators():
    for s1 in [
        '**', '/', '%',
        '&&',
        '|', '||',
        '^',
        '?',
    ]:
        check(s1+'0', [('bin_op', s1), ('integer', '0')])
        check(s1+'!', [('bin_op', s1), ('un_op', '!')])
        for s2 in '+-':
            check(s1+s2, [('bin_op', s1), ('un_bin_op', s2)])
        for s2 in '<>?':
            check(s1+s2, [('bin_op', s1), ('bin_op', s2)])

    for s1 in ['!', '~', '++', '--']:
        check(s1+'0', [('un_op', s1), ('integer', '0')])
        for s2 in '<>/^?':
            check(s1+s2, [('un_op', s1), ('bin_op', s2)])

    for s1 in ['+', '-', '&', '*']:
        check(s1+'0', [('un_bin_op', s1), ('integer', '0')])
        check(s1+'!', [('un_bin_op', s1), ('un_op', '!')])
        for s2 in '<>?':
            check(s1+s2, [('un_bin_op', s1), ('bin_op', s2)])

    for s1 in [
        '==', '!=',
        '<', '>', '<=', '>=',
        '<>', '<<', '>>>', '>>',
        '?',
    ]:
        check(s1+'0', [('bin_op', s1), ('integer', '0')])
        check(s1+'!', [('bin_op', s1), ('un_op', '!')])
        for s2 in '+-&':
            check(s1+s2, [('bin_op', s1), ('un_bin_op', s2)])
        for s2 in '^/%?':
            check(s1+s2, [('bin_op', s1), ('bin_op', s2)])

    for s1 in [
        '*=', '**=', '/=', '%=',
        '&&=',
        '|=', '||=',
        '^=',
        '+=', '-=', '&=',
        '='
    ]:
        check(s1+'0', [('assign_op', s1), ('integer', '0')])
        check(s1+'!', [('assign_op', s1), ('un_op', '!')])
        for s2 in '+-&':
            check(s1+s2, [('assign_op', s1), ('un_bin_op', s2)])
        for s2 in '^/%<>?':
            check(s1+s2, [('assign_op', s1), ('bin_op', s2)])

    for s1 in ['<=', '<>', '>=']:
        check(s1+'=', [('error', 'Ambiguous operator')])

    check('*/', [('un_bin_op', '*'), ('bin_op', '/')])
    check('/*', [('error', "Unclosed comment")])
    check('//', [])

    check(':=', [('error',
        'Unknown operator ":=" (assignments use "=" and declarations start '
        'with "const" or "var")')
    ])

    # Regression test. Once we parsed ",=" as ":=".
    check(',=', [(',',), ('assign_op', '=')])


def test_oneline_comments():
    check('/', [('bin_op', '/')])
    check('//', [])
    check('///', [])
    check('//\n', [('newline', '\n')])
    check('//\n\n', [('newline', '\n'), ('newline', '\n')])
    check('//\\', [])
    check('//1', [])
    check('//0x1', [])
    check('//.', [])
    check('//.1', [])
    check('//a', [])
    check('//@', [])
    check('//\r', [])
    check('//break', [])


def test_multiline_comments():
    check('/', [('bin_op', '/')])
    check('/*', [('error', "Unclosed comment")])
    check('/*\n', [('error', "Unclosed comment")])
    check('/*/', [('error', "Unclosed comment")])
    check('/**', [('error', "Unclosed comment")])
    check('/*@', [('error', "Unclosed comment")])
    check('/*\r', [('error', "Unclosed comment")])
    check('/**/', [])
    check('/*/**', [('error', "Unclosed comment")])
    check('/*/*/', [])
    check('/***/', [])
    check('/*\n*/', [])
    check('/*\\*/', [])
    check('/*/**/', [])
    check('/**/*/', [('un_bin_op', '*'), ('bin_op', '/')])
    check('/*/*/*', [('un_bin_op', '*')])
    check('/*/**/*', [('un_bin_op', '*')])
    check('/*1*/', [])
    check('/*0x1*/', [])
    check('/*.*/', [])
    check('/*.1*/', [])
    check('/*a*/', [])


def test_strings():
    check(r'"', [('error', "Unclosed string literal")])
    check(r'""', [('string', [])])
    check(r'"a"', [('string', ['a'])])
    check(r'"aa"', [('string', ['aa'])])
    check(r'"\a', [('error', "Unclosed string literal")])
    check(r'"\a"', [('string', [('error', "Bad escape sequence"), 'a'])])
    check(r'"\"', [('error', "Unclosed string literal")])
    check(r'"\""', [('string', ['"'])])
    check(r'"\\""', [('string', ['\\']), ('error', "Unclosed string literal")])
    check(r'"\\\""', [('string', [r'\"'])])
    check(r'"\"a"', [('string', ['"a'])])
    check(r'"a\""', [('string', ['a"'])])
    check(r'"0"', [('string', ["0"])])
    check(r'"0."', [('string', ["0."])])
    check(r'"0.0"', [('string', ["0.0"])])
    check(r'"."', [('string', ["."])])
    check(r'"0x0"', [('string', ["0x0"])])
    check(r'"for"', [('string', ["for"])])
    # WONTFIX: The division between the high- and low-level lexers is wrong.
    # We should not be looking for floats inside strings. This affects cases
    # such as "0.0e", "0.0e+" and "0x"'. This requires a substantial rewrite
    # of the lexer. We won't try to fix these cases before then.


def test_characters():
    check(r"'", [('error', "Unclosed character literal")])
    check(r"''", [('error', "Empty character literal")])
    check(r"'a'", [('character', 'a')])
    check(r"'aa'",
        [('error', "Character literals may only contain one character")]
    )
    check(r"'\a", [('error', "Unclosed character literal")])
    check(r"'\a'", [('error', "Bad escape sequence")])
    check(r"'\'", [('error', "Unclosed character literal")])
    check(r"'\''", [('character', "'")])
    check(r"'\\''",
        [('character', '\\'), ('error', "Unclosed character literal")]
    )
    check(r"'\\\'", [('error', "Unclosed character literal")])
    check(r"'\'a'",
        [('error', "Character literals may only contain one character")]
    )
    check(r"'a\'", [('error', "Unclosed character literal")])
    check(r"'for'",
        [('error', "Character literals may only contain one character")]
    )


def test_unicode_escapes():
    check(r'"\u0041"', [('string', ['A'])])
    check(r"'\u0041'", [('character', 'A')])
    check(r"'\u041'", [('error', 'Bad unicode escape sequence')])
    check(r"'\u00041'",
        [('error', "Character literals may only contain one character")]
    )


def test_coverage():
    for t0 in high_level_lexer_tags:
        assert t0 in observed_tags, t0

# Fuzzer.

alphabet = list("""\x00\x07\n !=<>///***."'"'"'\\012ABCbfnrtu_@""")
alphabet.append('\u0333')


def fuzz_once():
    s = alphabet
    for _ in range(random.randrange(1, 20)):
        s = [random.choice(s) for _ in range(60)]
    s = ''.join(s)
#    print()
#    print("Trying", s)
    listener = Buffer()
    p = Parser(low_level_lexer, Parser(high_level_lexer, listener))
    for i, c in enumerate(s):
        if ord(c) < 128:
            p.feed(ASCII(i, c))
        else:
            p.feed(NonASCII(i, i+1, c)) # WONTFIX: byte offsets are wrong.
    p.flush()
    result = listener.contents
#    print(len(result), "tokens.")
#    for token in result:
#        print(token)


def test_fuzz():
    for _ in range(100):
        fuzz_once()
