from .token import Error, Concat
from .grammar import Grammar


# Character ranges.

def char_range(first, last):
    return [chr(x) for x in range(ord(first), ord(last)+1)]


# Some common `post_process` routines. These are designed to be passed to the
# constructor of grammar.Grammar.

def impossible(multi_buffer):
    assert False, "Should not happen"


def discard(multi_buffer):
    return None


def pass_on(multi_buffer):
    return multi_buffer.get_only()


def concat(tag):
    '''
    A `post_process` function that concatenates the ASCIIs in the default
    stream and wraps them as a Concat of type `tag`.
    '''
    return lambda multi_buffer: Concat(tag, multi_buffer.get())


def error(message):
    '''
    Returns a `post_process' function that returns an Error.
    '''
    return lambda multi_buffer: Error(message, True, multi_buffer.get_all())


def pass_on_error(message):
    '''
    Returns a `post_process' function that returns an Error unless there is a
    token in stream 'error', in which case it returns that token instead.
    '''
    def choose_error(multi_buffer):
        ans = multi_buffer.get_optional('error')
        if ans is None:
            ans = Error(message, True, multi_buffer.get_all())
        return ans
    return choose_error


# A subclass of Grammar for internal states.

class Unfinished(Grammar):
    '''
    A Grammar representing a state in which more input is needed to finish the
    current token. Its `post_process` method generates an error token. If it
    receives an error token from a lower level, it throws away the input
    buffered so far and passes on the lower-level error token.
    '''
    def __init__(self, message, name=None):
        if name is None:
            name = "Unfinished(%r)" % message
        post_process = pass_on_error(message)
        super(Unfinished, self).__init__(post_process, name=name)
        self.set('error', Grammar(post_process), stream_name='error')

    def __repr__(self):
        return self.name
