'''
A library for writing type-specific pretty-printing logic, and a helper for
wrapping common Python types as pretty-print-aware values.

Basic usage:

    print(pretty(value))

Customisation:

    print(pretty(value, width=135, height=40, indent='==> ', tab='   .'))

If you pass a callable to `pretty()`, it treats it as a pretty-print-aware
value, and calls it to obtain a 4-tuple as described in `help(wrapper)`. For
example:

    def my_printable(value):
        assert isinstance(value, MyType)
        return lambda: (
            'MyType(',
            ', ',
            ')',
            [
                # We could use customised wrappers if necessary.
                auto_wrap(value.my_field1),
                auto_wrap(value.my_field2),
                auto_wrap(value.my_field3)
            ]
        )

    print(pretty(my_printable(MyType(...))))

Alternatively, if you are the author of MyType, you can implement
`MyType.__pretty__()` such that instances do not need a wrapper:

    class MyType:
        ...

        def __pretty__(self):
            return (
                'MyType(',
                ', ',
                ')',
                [
                    # We could use customised wrappers if necessary.
                    auto_wrap(value.my_field1),
                    auto_wrap(value.my_field2),
                    auto_wrap(value.my_field3)
                ]
            )

    print(pretty(MyType(...)))

As in this example, if a pretty-print-aware value has any nested values, it
must wrap them. The `auto_wrap()` function is useful for simple values (it is
the same wrapper that `pretty()` uses automatically when you pass it a
non-callable value).

A special wrapper `key_value()` is provided for (key, value) pairs.
'''


from .wrapper import auto_wrap, key_value
from .formatter import pretty


__all__ = ['auto_wrap', 'key_value', 'pretty']
