import heapq

from .wrapper import auto_wrap


def _omitted_text(remaining, omitted):
    '''
    Returns a unicode string suitable for replacing items omitted from a list.
    Returns the empty string if `omitted` is zero.

     - remaining - the number of items remaining in the list.
     - omitted - the number of items omitted from the list.
    '''
    if omitted == 0:
        return ''
    elif omitted == 1:
        return '...'
#    elif omitted == 2:
#        return '..., ...'
    else:
        if remaining == 0:
            return '%d items' % omitted
        else:
            return '%d more' % omitted


class Node(object):
    '''
    A mutable object representing a decision between various ways of printing
    a value.

    The lifecycle is as follows. A Node starts completely collapsed. Call the
    `make_children()` method to create its immediate children, and recursively
    make their children down to any desired depth. Then call
    `calculate_sizes()` on the root, to initialise the internal data
    structures. Then repeatedly call the `shrink()` method on the root to
    shrink its largest child, until the desired `height()` is achieved.

    Public fields:
     - value - the pretty-print-aware value.
     - children - a list of Nodes, initially `None`.
     - omitted - a bool indicating whether any children are missing, initially
       `None`.
    '''
    def __init__(self, value, tie_breaker, available_width, tab):
        '''
         - tie_breaker - an integer priority used to decide which child to
           `shrink()` if they are all the same size.
         - available_width - a number of characters, used to decide whether to
           format the value on a single line.
         - tab - a unicode string to be used to indent nested values.
           Typically this is a string of spaces.
        '''
        assert hasattr(value, '__call__'), (
            '%r is not callable; did you forget to put it in a '
            'pretty-print-aware wrapper?'
        )
        self.value = value
        self.tie_breaker = tie_breaker
        self.available_width = available_width
        self.tab = tab
        self.head = None
        self.children = None
        self.separator = None
        self.tail = None
        # A flag indicating if any children have been removed.
        self.omitted = None
        # A cache of sum(x.width() for x in self.children)
        self.total_width = None
        # A cache of sum(x.height() for x in self.children)
        self.total_height = None
        # A heapq of Nodes that prioritises the largest children and
        # omits those which cannot be shrunk further.
        self.queue = None

    def _cmp(self, other):
        '''
        Returns a negative number if `self` is bigger than `other`. `height()`
        is compared first, then `width()`, and finally `tie_breaker`.
        '''
        if type(other) is not Node:
            raise ValueError
        ret = other.height() - self.height()
        if ret == 0:
            ret = other.width() - self.width()
            if ret == 0:
                ret = other.tie_breaker - self.tie_breaker
        return ret

    def __lt__(self, other):
        return self._cmp(other) < 0

    def __le__(self, other):
        return self._cmp(other) <= 0

    def __eq__(self, other):
        return self._cmp(other) == 0

    def __ne__(self, other):
        return self._cmp(other) != 0

    def __gt__(self, other):
        return self._cmp(other) > 0

    def __ge__(self, other):
        return self._cmp(other) >= 0

    def __repr__(self):
        '''
        Returns this value as a unicode string as if printed on one line.
        The length of the returned string should equal `width()`.
        '''
        buf = []
        self._horizontal(buf)
        return ''.join(buf)

    def __str__(self):
        '''
        Returns `format()`.
        '''
        return self.format()

    def format(self, indent='\n'):
        '''
        Returns this value as a unicode string in the format we've decided to
        use. The number of lines of the returned string should equal
        `height()`.

         - indent - a unicode string that will be used to begin new lines.
           Typically this will be a string of spaces.
        '''
        buf = [indent]
        self._vertical(buf, '\n' + indent)
        return ''.join(buf)

    def _horizontal(self, buf):
        '''
        Appends to `buf` unicode strings representing this Node as a single
        line. The total length of the strings should equal `width()`.
        '''
        if self.children is None:
            buf.append('...')
            return
        buf.append(self.head)
        sep = ''
        for x in self.children:
            buf.append(sep)
            x._horizontal(buf)
            sep = self.separator
        if self.omitted > 0:
            buf.append(sep)
            buf.append(_omitted_text(len(self.children), self.omitted))
        buf.append(self.tail)

    def width(self):
        '''
        Returns the size of this Node if it were printed as one long line.
        The size is the head length, plus the sum of the widths of the
        children (and the ellipsis if any), plus a separator length for each
        child (or ellipsis) except the last, plus a separator length plus
        `len('...')` if any children are missing, plus the tail length.
        '''
        if self.queue is None:
            raise ValueError('calculate_sizes() not called yet')
        if self.children is None:
            return len('...')
        num_separators = max(0,
            len(self.children) - (1 if self.omitted > 0 else 0))
        return (
            len(self.head) +
            self.total_width +
            len(self.separator) * num_separators +
            (len(_omitted_text(len(self.children), self.omitted))) +
            len(self.tail)
        )

    def _vertical(self, buf, newline):
        '''
        Appends to `buf` unicode strings representing this Node in the format
        we've decided to use. The number of lines of the returned string
        should equal `height()`.

         - buf - a list.
         - newline - a unicode string that will be used to begin new lines.
        '''
        if self.children is None or self.width() <= self.available_width:
            self._horizontal(buf)
            return
        new_newline = newline + self.tab
        if self.head != '':
            buf.append(self.head)
            sep = new_newline
        else:
            sep = ''
        for x in self.children:
            buf.append(sep)
            x._vertical(buf, new_newline)
            sep = self.separator + new_newline
        if self.omitted > 0:
            buf.append(sep)
            buf.append(_omitted_text(len(self.children), self.omitted))
        if self.tail != '':
            buf.append(newline)
            buf.append(self.tail)

    def height(self):
        '''
        height - the number of lines occupied by this Node. If `width()`
        is less than `available_width`, returns 1, otherwise returns 1
        (for the head if non-empty) plus the sum of the heights of the
        children, plus 1 if there is an ellipsis, plus 1 if the tail is
        non-empty.
        '''
        if self.queue is None:
            raise ValueError('calculate_sizes() not called yet')
        if self.children is None or self.width() <= self.available_width:
            return 1
        return (
            int(self.head != '') +
            self.total_height +
            (1 if self.omitted > 0 else 0) +
            int(self.tail != '')
        )

    def make_children(self):
        '''
        Sets `self.children` to a list of Nodes, one for each of
        `self.value.children()`. Returns `self.children`.
        '''
        self.head, self.separator, self.tail, children = self.value()
        assert type(self.head) is str
        assert type(self.separator) is str
        assert type(self.tail) is str
        assert type(children) is list
        self.children = [Node(
                x,
                i,
                self.available_width - len(self.tab) - len(self.separator),
                self.tab
            ) for i, x in enumerate(children)
        ]
        self.omitted = 0
        return self.children

    def trim_boring(self):
        '''
        Removes trailing children whose `make_children()` method has not been
        called. The only thing we know about such children is how many of them
        there are; there's no point in trying to display them individually.
        '''
        if self.queue is None:
            raise ValueError('calculate_sizes() not called yet')
        while len(self.children) > 0 and self.children[-1].children is None:
            x = self.children.pop()
            self.omitted += 1
            self.total_width -= x.width()
            self.total_height -= x.height()

    def calculate_sizes(self):
        if self.children is None:
            self.total_width = 0
            self.total_height = 0
            self.queue = []
        else:
            for x in self.children:
                x.calculate_sizes()
            self.total_width = sum(x.width() for x in self.children)
            self.total_height = sum(x.height() for x in self.children)
            self.queue = [x for x in self.children if x.children is not None]
            heapq.heapify(self.queue)
            self.trim_boring()

    def shrink(self):
        '''
        Recursively calls `shrink()` on the largest child. If it cannot
        shrink, tries the next largest, and so on. If no child can shrink,
        removes a child. If there isn't one, returns `False`.

        Returns `True` if the call had any effect.
        '''
        if self.queue is None:
            raise ValueError('calculate_sizes() not called yet')
        if self.children is None:
            # `make_children()` not called on this Node.
            return False
        # Try to shrink the largest child.
        while len(self.queue) > 0:
            x = heapq.heappop(self.queue)
            w = x.width()
            h = x.height()
            if x.shrink():
                self.total_width += x.width() - w
                self.total_height += x.height() - h
                heapq.heappush(self.queue, x)
                return True
        # Try to remove a child.
        if len(self.children) > 0:
            x = self.children.pop()
            self.omitted += 1
            self.total_width -= x.width()
            self.total_height -= x.height()
            self.trim_boring()
            return True
        return False


def pretty(value, width=80, height=25, indent='', tab='  ', max_depth=10):
    '''
    Returns a nicely formatted unicode string representation of `value` that
    fits inside the specified screen area, if possible.

    As a convenience, if `value` is not callable, it will be passed through
    `auto_wrap()` in an attempt to make it pretty-print-aware. This copes with
    common cases. However, if you wish to pretty-print a value that might be
    callable but not pretty-print-aware, you must wrap it yourself, e.g. by
    passing it through `auto_wrap()`.

     - width - (optional) the width in characters of the desired output.
     - height - (optional) the height in lines of the desired output.
     - indent - (optional) a unicode string to prepend to every line:
       typically a string of spaces.
     - tab - (optional) a unicode string to use for further indentation:
       typically a string of spaces.
     - max_depth - (optional) the maximum nesting depth to show.
    '''
    def expand(root):
        '''
        Make more than enough descendant Nodes, subject to area and depth
        limits.
        '''
        nodes = [root]
        i = 0
        for depth in range(max_depth):
            end = len(nodes)
            while i < end:
                if len(nodes)*10 >= width*height:
                    # Heuristic estimate of the screen area the nodes will
                    # occupy exceeds the area available.
                    return
                nodes.extend(nodes[i].make_children())
                i += 1

    if not hasattr(value, '__call__'):
        value = auto_wrap(value)
    root = Node(value, 0, width-len(indent), tab)
    expand(root)
    # Shrink to fit.
    root.calculate_sizes()
    while root.height() > height:
        if not root.shrink():
            break
    return root.format(indent)
