'''
Pretty-print-aware values expose internal structure that is useful for
pretty-printing. Such values are typically wrappers around
non-pretty-print-aware values. The internal structure allows the pretty
printer to abbreviate values, and to split them across lines and indent them
in a sensible way.

Pretty-print-aware values are represented by a callable that returns a tuple
of four values:
 - head - a unicode string that the pretty printer outputs first, e.g. an
   open bracket.
 - separator - a unicode string that the pretty printer outputs after
   every child value, e.g. a comma. The pretty printer may omit the
   separator after the last child.
 - tail - a unicode string that the pretty printer outputs last, e.g. a
   close bracket.
 - children - a finite number of nested values, which must be
   pretty-print-aware.

The list of children must be evaluated lazily, i.e. it must be possible to
construct a value in a finite amount of time even if it has an infinite number
of descendents.

The pretty printer has freedom to output the value is several ways. It
will choose how to output the value depending on the amount of space
available, and possibly on other configuration options. Example output
formats for a value with three children include:

 - head ... tail

 - head child sep ... tail

 - head child sep child sep child tail

 - head
     child sep child sep child sep
   tail

 - head
     child sep
     ...
   tail

 - head
     child sep
     child sep
     child sep
   tail
'''


def key_value(k, v):
    return lambda: ('', ': ', '', [auto_wrap(k), auto_wrap(v)])


def auto_wrap(value):
    '''
    Converts an ordinary Python value into a pretty-print-aware value.
    If `value` has a `__pretty__` attribute, returns it.
    If `value` is an instance of a common Python container type, returns a
    suitable pretty-print-aware wrapper around it which will recursively wrap
    its contents. Otherwise, returns a wrapper that calls `repr()` on the
    value.
    '''
    if hasattr(value, '__pretty__'):
        return value.__pretty__
    if isinstance(value, tuple):
        return lambda: ('(', ', ', ')', [auto_wrap(x) for x in value])
    elif isinstance(value, list):
        return lambda: ('[', ', ', ']', [auto_wrap(x) for x in value])
    elif isinstance(value, set):
        return lambda: ('{', ', ', '}', [auto_wrap(x) for x in value])
    elif isinstance(value, frozenset):
        return lambda: (
            'frozenset([', ', ', '])', [auto_wrap(x) for x in value])
    elif isinstance(value, dict):
        return lambda: (
            '{', ', ', '}', [key_value(k, v) for k, v in value.items()])
    else:
        return lambda: ('%r' % value, '', '', [])
