from ..parser import WellyParser
from ..printer import auto_wrap
from ..model import defer, pack, wtypes
from ..compiler import operators, syscalls, linker, builtins

from ..model.errors import *
from ..model.memory import Ref, Array
from ..model.ops import Dynamic
from ..compiler.flattener import LexicalScope, Repl, from_id


int_size = wtypes.sizeof(wtypes.int_)
dynamic_size = wtypes.sizeof(wtypes.dynamic)


class Stack(object):
    def __init__(self):
        self.items = []

    @staticmethod
    def from_list(items):
        '''
         - items - an iterable yielding Refs.
        '''
        ret = Stack()
        for x in items:
            ret.push(x)
        return ret

    def pop_many(self, n):
        if n < 0:
            raise IndexError('Negative index')
        new_length = len(self.items) - n
        if new_length < 0:
            raise IndexError('Stack underflow')
        result = self.items[new_length:]
        self.items[new_length:] = []
        return result

    def pop(self):
        return self.pop_many(1)[0]

    def push(self, x):
        assert type(x) is Ref
        self.items.append(x)

    def push_many(self, xs):
        assert type(xs) is list
        assert all(type(x) is Ref for x in xs)
        self.items.extend(xs)

    def __pretty__(self):
        return (
            'Stack([',
            ', ',
            '])',
            [x.__pretty__ for x in self.items]
        )


# TODO: Move?
class WellyRunner(WellyParser):
    def __init__(self, globals_, on_output):
        '''
        Constructs a WellyParser that runs the statements it parses.

         - globals_ - a dict from str to Dynamic representing the current
           global variables. This dict will be modified when code runs.
         - on_output - `None` or a Dynamic representing a Welly function to
           be used to print out expressions used as statements. It takes:
            - is_static - bool.
            - value - dynamic.
           and returns void.
        '''
        super(WellyRunner, self).__init__()
        self.globals_ = globals_
        self.on_output = on_output

    def do_stmt(self, ast):
        '''
        Attempts to run `ast`, and updates `globals_` if successful.

         - ast - the abstract syntax tree of a statement.

        Raises:
         - ASTError - for any kind of syntax error.
         - FlattenerError - for syntax usage errors.
         - StaticVMError - for errors detected using static checking.
         - RuntimeVMError - for errors detected at runtime.
         - NotImplementedError - for unimplemented features.
         - MemoryError, OverflowError - in extremis.
        Other Exception types are bugs.
        '''
        # Flatten.
        scope = LexicalScope(Repl())
        scope.flatten_stmt(ast, 'the top-level scope')
        # Append code to push newly defined values. Also, record their names.
        new_names = []
        for token in scope.locals_.as_list():
            name = from_id(token)
            if name is not None:
                scope.flatten_identifier(token)
                new_names.append(name)
        # Construct captures array.
        capture_ids = scope.top_level.captures.as_list()
        captures = Array.malloc(len(capture_ids), dynamic_size)
        for identifier, ref in zip(capture_ids, captures):
            name = from_id(identifier)
            if name not in self.globals_:
                raise UndefinedNameVMError(name, identifier.error_loc)
            ref.write(self.globals_[name].to_welly())
        # Run.
        result = run_statement(
            scope.block.get_code(is_break_allowed=False),
            captures,
            self.on_output,
        )
        # Update `globals`.
        assert len(new_names) == len(result), (new_names, result)
        self.globals_.update(zip(new_names, result))


def run_statement(flat_code, captures_array, on_output):
    '''
    Runs a top-level statement.

    Params:
     - flat_code - a list of flat code instructions to run.
     - captures_array - an Array of dynamic values, providing values for
       use by 'capture' instructions. These are used to represent values
       mentioned in the code but defined outside it.
     - on_output - `None` or a Dynamic representing a Welly function to be
       used to print out expressions used as statements. It takes:
        - is_static - bool.
        - value - dynamic.
       and returns void.

    Returns a list of Dynamic, being the values that `flat_code` leaves on
    the stack.
    '''
    captures = [Dynamic.from_welly(x) for x in captures_array]
    typed_code, cts = linker.link_statement(flat_code, captures, on_output)
    reason = run(typed_code, [], [])
    r0 = reason[0]
    if r0 == 'fallthrough':
        _, values = reason
        assert len(values) == len(cts)
        return [Dynamic(t, ref, constness=c)
            for ref, (c, t) in zip(values, cts)]
    elif r0 == 'throw':
        raise NotImplementedError('throw')
    else:
        # 'break', 'continue', 'return' are impossible.
        assert False, reason


def run_function_call(
    typed_code, # From (typed) 'closure' instruction.
    captures, # From closure value.
    arguments, # From caller.
):
    '''
    Call a function.

    Params:
     - typed_code - a list of typed code instructions representing the
       function body (as included in the 'closure' instruction).
     - captures - a list of Refs, providing values for use by 'capture'
       instructions. These are used to represent run-time values mentioned in
       the code but defined outside it.
     - arguments - a const Ref to a struct containing the values passed in by
       the caller of the function. The struct completely fills a heap block.

    Returns one of the following, as per "doc/python_model.txt":
     - `('return', result)` where `result` is const Ref to the return value.
     - `('throw', exception)` where `exception` is TODO.
    '''
    assert arguments.offset == 0
    assert arguments.size == len(arguments.block)
    # Run the function body.
    reason = run(typed_code, captures, [], stack=[arguments])
    if reason[0] == 'fallthrough':
        raise IllegalInstructionVMError('Missing "return"')
    elif reason[0] == 'break':
        raise IllegalInstructionVMError('Cannot "break" outside a loop')
    elif reason[0] == 'continue':
        raise IllegalInstructionVMError('Cannot "continue" outside a loop')
    elif reason[0] == 'return':
        _, result = reason
        return ('return', result.copy())
    elif reason[0] == 'throw':
        _, exception = reason
        raise NotImplementedError('throw')
    assert False, "Unknown reason for exiting"


# All cases call this.
def run(typed_code, captures, locals_, stack=[]):
    '''
    Runs `typed_code`.

     - typed_code - a list of typed code instructions. You are encouraged to
       generate this using welly.compiler.typed_code.Block, which performs
       type-checking and gives decent error messages.
     - captures - list of Ref.
     - locals_ - list of Ref.
     - stack - list of Ref.

    Returns one of:
     - ('fallthrough', values) - Indicates that execution reached the end
       of the block. `values` is a list of Refs to the values
       remaining on the block's stack when it terminates. The list will be
       non-empty e.g. for a call to `eval()`.
     - ('break',) - Indicates that execution ended because of a 'break'
       instruction.
     - ('continue',) - Indicates that execution ended because of a
       'continue' instruction.
     - ('return', value) - Indicates that execution ended because of a
       'return' instruction. `value` is a Ref to the returned value. The value
       has not yet been copied.
     - ('throw', exception) - Indicates that execution ended because of an
       exception.
    '''
    for x in captures:
        assert type(x) is Ref
    for x in locals_:
        assert type(x) is Ref
    stack = Stack.from_list(stack)
    num_locals = len(locals_)
    for instr in typed_code:
        i0 = instr[0]
        if i0 == 'address_of':
            (_,) = instr
            x = stack.pop()
            result = Ref.new((x,))
            stack.push(result)
        elif i0 == 'array0':
            (_,) = instr
            result = Ref.new(Array.malloc(0, 0).to_welly())
            stack.push(result)
        elif i0 == 'array':
            _, size, n = instr
            xs = stack.pop_many(n)
            array = Array.malloc(n, size)
            for x, y in zip(xs, array):
                y.write(x.read())
            result = Ref.new(array.to_welly())
            stack.push(result)
        elif i0 == 'array_concat':
            _, elem_size = instr
            right_array = Array.from_welly(stack.pop().read(), elem_size)
            left_array = Array.from_welly(stack.pop().read(), elem_size)
            result = Array.malloc(
                len(left_array) + len(right_array), elem_size)
            out = iter(result)
            for elem in left_array.read():
                next(out).write(elem)
            for elem in right_array.read():
                next(out).write(elem)
            for ref in out:
                assert False
            stack.push(Ref.new(result.to_welly()))
        elif i0 == 'array_length':
            _, elem_size = instr
            array = Array.from_welly(stack.pop().read(), elem_size)
            result = len(array)
            stack.push(Ref.new((result,)))
        elif i0 == 'array_repeat':
            _, elem_size = instr
            (count,) = stack.pop().read()
            array = Array.from_welly(stack.pop().read(), elem_size)
            if count < 0:
                raise ArrayRepeatVMError(count)
            result = Array.malloc(len(array) * count, elem_size)
            out = iter(result)
            for i in range(count):
                for elem in array.read():
                    next(out).write(elem)
            for ref in out:
                assert False
            stack.push(Ref.new(result.to_welly()))
        elif i0 == 'array_slice':
            _, elem_size = instr
            (end,) = stack.pop().read()
            (start,) = stack.pop().read()
            array = Array.from_welly(stack.pop().read(), elem_size)
            if 0 > start or start > end or end > len(array):
                raise ArraySliceVMError(start, end, len(array))
            result = array.slice(start, end)
            stack.push(Ref.new(result.to_welly()))
        elif i0 == 'assign':
            _, size = instr
            rvalue, lvalue = stack.pop_many(2)
            assert lvalue.size == size
            assert rvalue.size == size
            lvalue.write(rvalue.read())
        elif i0 == 'binary':
            _, opcode = instr
            (y,) = stack.pop().read()
            (x,) = stack.pop().read()
            result = operators.index[opcode].implementation(x, y)
            stack.push(Ref.new((result,)))
        elif i0 == 'box':
            _, constness, type_ = instr
            ref = stack.pop()
            result = Dynamic(type_, ref, constness=constness)
            stack.push(Ref.new(result.to_welly()))
        elif i0 == 'break':
            (_,) = instr
            locals_[num_locals:] = [] # Because we're returning.
            return ('break',)
        elif i0 == 'bury':
            _, n = instr
            x = stack.pop()
            items = stack.pop_many(n)
            stack.push(x)
            stack.push_many(items)
        elif i0 == 'call_function':
            _, sizes_ref, sizes_len = instr
            sizes = Array(sizes_ref, sizes_len, int_size)
            xs = stack.pop_many(len(sizes))
            func = stack.pop()
            func_code, func_captures = defer.read(
                func, noun='the function value')
            # Construct the arg struct.
            args = Ref.malloc(sum(size for (size,) in sizes.read()))
            offset = 0
            for x, (size,) in zip(xs, sizes.read()):
                args.field(offset, size).write(x.read())
                offset += size
            assert offset == args.size
            # Run `func_code`.
            reason = run_function_call(func_code, func_captures, args)
            # Switch on the return value.
            if reason[0] == 'return':
                _, result = reason
                stack.push(result)
            elif reason[0] == 'throw':
                raise NotImplementedError('throw')
            else:
                assert False, reason
        elif i0 == 'capture':
            _, index = instr
            if index < 0 or index >= len(captures):
                raise IllegalInstructionVMError('"capture" index out of range')
            stack.push(captures[index])
        elif i0 == 'cast':
            _, size = instr
            x = stack.pop()
            assert size == x.size
            result = x.copy()
            stack.push(result)
        elif i0 == 'closure':
            _, body_code, n = instr
            func_captures = stack.pop_many(n)
            result = Ref.new((body_code, func_captures))
            stack.push(result)
        elif i0 == 'continue':
            (_,) = instr
            locals_[num_locals:] = [] # Because we're returning.
            return ('continue',)
        elif i0 == 'defer':
            (_,) = instr
            result = defer.defer()
            stack.push(result)
        elif i0 == 'dereference':
            _, size = instr
            x = stack.pop()
            (result,) = x.read()
            assert result.size == size
            stack.push(result)
        elif i0 == 'drop':
            (_,) = instr
            _ = stack.pop()
        elif i0 == 'dynamic_constness':
            (_,) = instr
            x = stack.pop()
            result = Dynamic.from_welly(x).constness
            stack.push(result)
        elif i0 == 'dynamic_type':
            (_,) = instr
            x = stack.pop()
            result = Dynamic.from_welly(x).type_
            stack.push(result)
        elif i0 == 'fulfil':
            (_,) = instr
            rvalue, lvalue = stack.pop_many(2)
            assert lvalue.size == 2
            assert rvalue.size == 2
            defer.fulfil(lvalue, rvalue)
        elif i0 == 'if':
            _, n, then_code, else_code = instr
            (condition,) = stack.pop().read()
            if condition:
                reason = run(then_code, captures, locals_)
            else:
                reason = run(else_code, captures, locals_)
            if reason[0] == 'fallthrough':
                _, values = reason
                assert len(values) == n
                for x in values:
                    stack.push(x)
            else:
                locals_[num_locals:] = [] # Because we're returning.
                return reason
        elif i0 == 'init':
            _, size = instr
            x = stack.pop()
            assert x.size == size
            locals_.append(x.copy())
        elif i0 == 'inline':
            _, body_code = instr
            reason = run(body_code, captures, locals_)
            if reason[0] == 'fallthrough':
                _, values = reason
                for x in values:
                    stack.push(x)
            else:
                locals_[num_locals:] = [] # Because we're returning.
                return reason
        elif i0 == 'local':
            _, index = instr
            if index < 0 or index >= len(locals_):
                raise IllegalInstructionVMError('"local" index out of range')
            stack.push(locals_[index])
        elif i0 == 'loop':
            _, body_code = instr
            while True:
                reason = run(body_code, captures, locals_)
                r0 = reason[0]
                if r0 == 'fallthrough':
                    pass
                elif r0 == 'break':
                    break
                elif r0 == 'continue':
                    pass
                else: # 'return' or 'throw'.
                    locals_[num_locals:] = [] # Because we're returning.
                    return reason
        elif i0 == 'new':
            _, size = instr
            x = stack.pop()
            assert x.size == size
            result = Ref.new((x.copy(),))
            stack.push(result)
        elif i0 == 'pick':
            _, n = instr
            items = stack.pop_many(n)
            x = stack.pop()
            stack.push(x)
            stack.push_many(items)
            stack.push(x)
        elif i0 == 'return':
            _, size = instr
            x = stack.pop()
            locals_[num_locals:] = [] # Because we're returning.
            return ('return', x)
        elif i0 == 'reverse':
            _, n = instr
            xs = stack.pop_many(n)
            stack.push_many(xs[::-1])
        elif i0 == 'static':
            _, ref, n = instr
            items = stack.pop_many(n)
            stack.push(ref)
            stack.push_many(items)
        elif i0 == 'str_length':
            (_,) = instr
            (str_,) = stack.pop().read()
            result = len(str_)
            stack.push(Ref.new((result,)))
        elif i0 == 'struct':
            _, sizes_ref, sizes_len = instr
            sizes = Array(sizes_ref, sizes_len, int_size)
            xs = stack.pop_many(len(sizes))
            result = Ref.malloc(sum(size for (size,) in sizes.read()))
            offset = 0
            for x, (size,) in zip(xs, sizes.read()):
                result.field(offset, size).write(x.read())
                offset += size
            assert offset == result.size
            stack.push(result)
        elif i0 == 'subscript_array':
            _, size = instr
            (i,) = stack.pop().read()
            a = Array.from_welly(stack.pop().read(), size)
            if i < 0 or i >= a.length:
                raise ArrayIndexVMError(i, a.length)
            result = a[i]
            stack.push(result)
        elif i0 == 'subscript_struct':
            _, offset, size = instr
            x = stack.pop()
            result = x.field(offset, size)
            stack.push(result)
        elif i0 == 'switch':
            _, case_triples, else_code = instr
            x = stack.pop()
            tag, payload = defer.read(x)
            assert type(tag) is str, x
            payload = Ref.block(payload)
            # Search for the case to execute.
            for tag2, sizes_ref, sizes_len, case_code in case_triples:
                if tag2 == tag:
                    # It's the case we're looking for. Unpack the payload.
                    sizes = Array(sizes_ref, sizes_len, int_size)
                    before = len(locals_)
                    offset = 0
                    for (size,) in sizes.read():
                        field = payload.field(offset, size)
                        locals_.append(field)
                        offset += size
                    assert offset == payload.size
                    reason = run(case_code, captures, locals_)
                    locals_[before:] = []
                    break
            else:
                # Execute the "else" block.
                reason = run(else_code, captures, locals_)
            # Check/propagate the exit reason.
            if reason[0] == 'fallthrough':
                _, values = reason
                assert len(values) == 0
            else:
                # Note: Unlike C/Java, "break" propagates out of "switch".
                locals_[num_locals:] = [] # Because we're returning.
                return reason
        elif i0 == 'syscall':
            _, name = instr
            x = stack.pop().read()
            result = syscalls.index[name].implementation(x)
            stack.push(Ref.new(result))
        elif i0 == 'unary':
            _, opcode = instr
            (x,) = stack.pop().read()
            result = operators.index[opcode].implementation(x)
            stack.push(Ref.new((result,)))
        elif i0 == 'union':
            _, tag, sizes_ref, sizes_len = instr
            sizes = Array(sizes_ref, sizes_len, int_size)
            xs = stack.pop_many(len(sizes))
            payload = Ref.malloc(sum(size for (size,) in sizes.read()))
            offset = 0
            for x, (size,) in zip(xs, sizes.read()):
                payload.field(offset, size).write(x.read())
                offset += size
            assert offset == payload.size
            result = Ref.new((tag, payload.block))
            stack.push(result)
        elif i0 == 'unpack_struct':
            _, sizes_ref, sizes_len = instr
            sizes = Array(sizes_ref, sizes_len, int_size)
            x = stack.pop()
            offset = 0
            for (size,) in sizes.read():
                result = x.field(offset, size)
                stack.push(result)
                offset += size
            assert offset == x.size
        else:
            assert False, 'Unknown typed code instruction: %r' % i0
    locals_[num_locals:] = []
    return ('fallthrough', stack.items)
