from ..model import defer, pack, const, wtypes
from ..model.memory import Array


'''
The definitions in this file are roughly equivalent to the following Welly
code.

    const register: type = int
    array_of_register = array const register

    fn new_register(): register { ... }
    macro map_from_register(t: type) { ... }

    union intrinsic {} // TODO.

    fn input_types(i: intrinsic): array const type { ... }
    fn output_types(i: intrinsic): array const type { ... }

    const op = defer type
    const block = array const op
    fulfil op = struct(
        /* Registers to receive results from the previous instruction. */
        array_of_register,
        /* The instruction and its operands. */
        union instruction {
            /* Op-code, operands. */
            ARITHMETIC(intrinsic, array_of_register);
            /* Target, arguments. */
            GOTO(register, array_of_register);
            /* Argument types, body. */
            CONTINUATION(array const type, block);
            /* Argument types, body.
             * The return continuation is not included in the argument types. */
            FUNCTION(array const type, block);
            /* Selector, case_name, case_body, else_body. */
            SWITCH(register, str, block, block);
        },
    )
'''

register = wtypes.int_
array_of_register = wtypes.array(const.CONST, register)

intrinsic = wtypes.union([])

op = defer.defer()
block = wtypes.array(const.CONST, op)
instruction = wtypes.union([
    ('ARITHMETIC', [intrinsic, array_of_register]),
    ('GOTO', [register, array_of_register]),
    ('CONTINUATION', [wtypes.array_of_type, block]),
    ('FUNCTION', [wtypes.array_of_type, block]),
    ('SWITCH', [register, wtypes.str_, block, block]),
])
defer.fulfil(op, wtypes.struct([array_of_register, instruction]))


_SWITCH_DICT = {
    'ARITHMETIC': 'r2a1',
    'GOTO': 'w1a1',
    'CONTINUATION': 'a2a4',
    'FUNCTION': 'a2a4',
    'SWITCH': 'w1w1a4a4',
}

def switch(instruction):
    '''
     - instruction - a Ref to an `instruction`.
    '''
    return pack.switch(instruction, _SWITCH_DICT)


_next_register = 100

class Register(object):
    def __init__(self, r):
        self.r = r

    @staticmethod
    def new():
        global _next_register
        ret = Register(_next_register)
        _next_register += 1
        return ret

    def to_welly(self):
        return (self.r,)

    def from_welly(self, ref):
        '''
         - ref - a Ref to a Welly int.
        '''
        (r,) = ref.read()
        return Register(r)

    def __eq__(self, other):
        return self.r == other.r

    def __hash__(self):
        return hash(self.r)

    def __repr__(self):
        return 'R%d' % (self.r,)


R0 = Register(0)
R1 = Register(1)
R2 = Register(2)
R3 = Register(3)
R4 = Register(4)
R5 = Register(5)
R6 = Register(6)
R7 = Register(7)
R8 = Register(8)
R9 = Register(9)

