from .code_types import Register


class UntypedConsumer(object):
    '''
    The abstract superclass of all things that consume a program.

    Methods with names matching the cases in the grammar (`arithmetic`,
    `goto`, `continuation`, `function` and `switch`) are called by a Producer
    to pass instructions to this class. They mutate this UntypedConsumer.
    They return `None`.
    '''
    def arithmetic(self, intrinsic, operands, results):
        '''
        Processes an ARITHMETIC instruction.
         - intrinsic - an Intrinsic.
         - operands - a list of Registers.
         - results - a list of Registers.
        '''
        raise NotImplementedError('Abstract method')

    def goto(self, target, arguments):
        '''
        Processes a GOTO instruction.
         - target - a Register.
         - arguments - a list of Registers.
        '''
        raise NotImplementedError('Abstract method')

    def continuation(self, arg_types, body, result):
        '''
        Processes a continuation instruction.
         - arg_types - a list of Refs to Welly types.
         - body - an Array representing an `array const op`.
         - result - a Register.
        '''
        raise NotImplementedError('Abstract method')

    def function(self, arg_types, body, result):
        '''
        Processes a FUNCTION instruction.
         - arg_types - a list of Refs to Welly types.
         - body - an Array representing an `array const op`.
         - result - a Register.
        '''
        raise NotImplementedError('Abstract method')

    def switch(self, selector, case_name, case_body, else_body):
        '''
        Processes one case of a SWITCH instruction.
         - selector - a Register.
         - case_name - the tag we're looking for.
         - case_body - an Array representing an `array const op`.
         - else_body - an Array representing an `array const op`.
        '''
        raise NotImplementedError('Abstract method')

    def get_value(self, register):
        '''
        Returns a Ref representing the current value of `register`, if known,
        otherwise `None`.
        '''
        raise NotImplementedError('Abstract method')


class Consumer(UntypedConsumer):
    def get_type(self, register):
        '''
        Returns the type of `register`.
        '''
        raise NotImplementedError('Abstract method')


# TODO: Can we remove this?
class Producer(object):
    '''
    The abstract superclass of all things that compute a program.

    Public fields:
     - bindings - a list of Registers, representing the inputs to the program.
    '''
    def __init__(self, bindings):
        for r in bindings:
            assert isinstance(r, Register)
        self.bindings = bindings

    def produce(self, consumer):
        '''
        Computes all the instructions to the end of the basic block (ending
        with a non-falling-through instruction such as a 'GOTO' or a
        'SWITCH'). It passes each instruction to `consumer` by calling its
        method corresponding to the instruction type. It then returns `None`.

         - consumer - an UntypedConsumer (which may be a Consumer) to which to
           feed instructions.
        '''
        raise NotImplementedError('Abstract method')
