import core


def _registers_from_welly(array):
    return [Register.from_welly(ref) for ref in array]


def _heap_block_from_array_const_op(array):
    '''
    Given an Array representing an `array const op` encompassing an entire
    heap block, returns the heap block.
    '''
    assert isinstance(array, Array)
    assert array.elem_size = wtypes.sizeof(op)
    ref = array.ref
    assert ref.offset = 0
    assert ref.size = len(ref.block)
    return ref.block


class Interpreter(core.UntypedConsumer):
    '''
    Public fields:
     - env - a dict from Register to Ref.
     - producer - a core.Producer.
    '''
    def __init__(self, env, producer):
        self.env = env
        self.producer = producer

    def _update_pairs(self, registers, refs):
        assert len(registers) == len(refs)
        self.env.update(zip(registers, refs))

    def step(self):
        self.producer.produce(self)

    def arithmetic(self, intrinsic, operands, results):
        operand_refs = [self.env[r] for r in operands]
        result_refs = intrinsic.execute(operands)
        self.env.update_pairs(results, result_refs)

    def goto(self, target, arguments):
        target_ref = self.get_value(target)
        argument_refs = [self.get_value(r) for r in arguments]
        code, captures = defer.read(target_ref)
        code = Array.from_list(code, wtypes.sizeof(op))
        self.producer = Producer(code)
        self.env = captures.copy()
        self.env.update_pairs(arguments, argument_refs)

    def continuation(self, arg_types, body, result):
        body = _heap_block_from_array_const_op(body)
        captures = self.env.copy() # TODO: Convert to a Welly value.
        result_ref = Ref.new((body, captures))
        self.env[result] = result_ref

    def function(self, arg_types, body, result):
        body = _heap_block_from_array_const_op(body)
        captures = self.env.copy() # TODO: Convert to a Welly value.
        result_ref = Ref.new((body, captures))
        self.env[result] = result_ref

    def switch(self, selector, case_name, case_body, else_body):
        selector_ref = self.get_value[selector]
        tag, payload_ref = defer.read(selector_ref)
        if tag == case_name:
            self.producer = Producer(case_body)
            self.env.update_pairs(self.producer.bindings, [payload_ref])
        else:
            self.producer = Producer(else_body)
            self.env.update_pairs(self.producer.bindings, [selector_ref])

    def get_value(self, register):
        return self.env[register]


class Producer(core.Producer):
    '''
    A core.Producer that reads from a Welly `array const op`.
    '''
    def __init__(self, ops):
        '''
         - ops - an Array representing the `array const op`.
        '''
        assert isinstance(ops, Array)
        self.ops = iter(ops)
        bindings = self._next()
        super(Producer, self).__init__(bindings)

    def _next(self):
        '''
        Unpacks the next `op`. Sets `self._instruction` to a Ref to the
        Welly `instruction`. Returns the bindings as a list of Registers.
        '''
        bindings, self._instruction = pack.unpack(next(self.ops), 'a1r2')
        return [Register.from_welly(ref) for ref in bindings]

    def _finish(self):
        for _ in self.ops:
            assert False, "Unreachable code"
        self._instruction = None

    def _produce_one(self, consumer):
        tag, payload = core_types.switch(self._instruction)
        if tag == 'ARITHMETIC':
            instrinsic, operands = payload
            # TODO: Convert `intrinsic` into an Intrinsic.
            operands = [Register.from_welly(ref) for ref in operands]
            results = self._next()
            consumer.arithmetic(intrinsic, operands, results)
        elif tag == 'GOTO':
            target, arguments = payload
            arguments = [Register.from_welly(ref) for ref in arguments]
            self._finish()
            consumer.goto(target, arguments)
        elif tag == 'CONTINUATION':
            arg_types, body = payload
            arg_types = list(arg_types)
            (result,) = self._next()
            consumer.continuation(arg_types, body, result)
        elif tag == 'FUNCTION':
            arg_types, body = payload
            arg_types = list(arg_types)
            (result,) = self._next()
            consumer.function(arg_types, body, result)
        elif tag == 'SWITCH':
            selector, case_name, case_body, else_body = payload
            self._finish()
            consumer.switch(selector, case_name, case_body, else_body)
        else:
            assert False, 'Unknown instruction: %r' % tag

    def produce(self, consumer):
        while self._instruction is not None:
            self.produce_one(consumer)
