'''
Interfaces the pretty printer to the Welly data model.
'''


import re

from . import defer, const, wtypes, subtypes, pretty_type

from .errors import UnfulfilledValueVMError
from .memory import Ref, Array
from .ops import Dynamic, struct_iter


__all__ = ['wrap']


def _constant(value):
    assert type(value) is str
    return lambda: (value, '', '', [])


# To convert a string value into a Welly string literal, we have to make the
# following replacements.
_esc = {
    '\t': r'\t',
    '\n': r'\n',
    '\r': r'\r',
    '\"': r'\"',
    '\\': r'\\',
}


# Exploit the fact that Welly escapes are a subset of Python escapes.
_regexp = re.compile(r'[%s]' % ''.join(_esc.values()))


def _str(value):
    assert type(value) is str, type(value)
    value = _regexp.sub(lambda m: _esc[m.group()], value)
    return lambda: ('"%s"' % value, '', '', [])


def _struct(value, tag=''):
    assert type(value) is Dynamic
    return lambda: (
        '%s(' % tag,
        ', ',
        ')',
        [wrap(x) for x in struct_iter(value)]
    )


def wrap(x):
    '''
    Given a Welly value, returns a pretty-print-aware wrapper around it.

     - x - a Dynamic.
    '''
    tag, payload = wtypes.switch(x.type_)
    if tag in ('FLOAT', 'INT'):
        # Use a Python value with the same syntax as the Welly value.
        (value,) = x.read()
        return _constant(str(value))
    elif tag == 'BOOL':
        (value,) = x.read()
        return _constant('true' if value else 'false')
    elif tag == 'STR':
        (value,) = x.read()
        return _str(value)
    elif tag == 'REF':
        (constness, referee_type) = payload
        (value,) = x.read()
        return lambda: ('&', '', '', [
            wrap(Dynamic(referee_type, value, constness=constness))
        ])
    elif tag == 'ARRAY0':
        # Use a Python value with the same syntax as the Welly value.
        return _constant('[]')
    elif tag == 'ARRAY':
        (constness, elem_type) = payload
        values = Array.from_welly(x.read(), wtypes.sizeof(elem_type))
        return lambda: ('[', ', ', ']', [
            wrap(Dynamic(elem_type, ref, constness=constness))
            for ref in values
        ])
    elif tag == 'STRUCT':
        (types,) = payload
        return _struct(x)
    elif tag == 'UNION':
        (cases,) = payload
        if subtypes.eq(x.type_, wtypes.type_):
            # Use Welly syntactic sugar.
            return pretty_type.wrap(x.ref)
        try:
            tag, payload = defer.read(x.ref)
        except UnfulfilledValueVMError:
            return _constant('unfulfilled')
        assert type(tag) is str, x
        payload = Ref.block(payload)
        types = wtypes.union_index(cases)[tag]
        payload_t = wtypes.struct(types)
        payload_struct = Dynamic(payload_t, payload)
        return _struct(payload_struct, tag=tag)
    elif tag == 'FUNCTION':
        (param_types, return_type) = payload
        try:
            tag, payload = defer.read(x.ref)
        except UnfulfilledValueVMError:
            return _constant('unfulfilled')
        # We have no way of displaying any details.
        return _constant('<fn>')
    elif tag == 'DYNAMIC':
        value = Dynamic.from_welly(x.ref)
        return lambda: ('box ', '', '', [wrap(value)])
    assert False, tag
